import os
import re

from si_tests import settings

from si_tests.deployments.utils import (file_utils, wait_utils,
                                        commons, kubectl_utils, helpers)
from si_tests.deployments.utils.namespace import NAMESPACE


def deploy_dns(os_manager, timeout):
    kubectl = kubectl_utils.Kubectl()
    commons.LOG.info("Deploy coredns")
    kubectl.apply(file_utils.get_release("ci/30-coredns.yaml"))
    commons.LOG.info("Wait for coredns service")
    core_dns_service = os_manager.api.services.get(
        "coredns-coredns", NAMESPACE.coredns)

    wait = wait_utils.Waiter(os_manager, timeout)

    wait.wait_pass(wait.read_checker, k8s_object=core_dns_service,
                   expected=Exception)

    ext_dns_ip = helpers.get_external_dns_ip(os_manager)

    commons.LOG.info("Check is coredns or kubedns")
    is_core_dns = helpers.is_core_dns(os_manager)

    if is_core_dns:
        commons.LOG.info("Coredns successfully deployed")
        return

    commons.LOG.info("Configure kubedns")
    kube_dns_content = """
    apiVersion: v1
    kind: ConfigMap
    metadata:
      name: kube-dns
      namespace: kube-system
    data:
      stubDomains: |
        {"mirantis.net": ["172.18.208.44"],
         "it.just.works": ["%s"]
        }
    """ % ext_dns_ip
    artifacts_dir = settings.ARTIFACTS_DIR

    kube_dns_yaml = os.path.join(artifacts_dir, "kube-dns.yml")
    with open(kube_dns_yaml, "w") as f:
        f.write(kube_dns_content)

    commons.LOG.info("Deploy kubedns")
    kubectl.apply(kube_dns_yaml, cwd=artifacts_dir)

    commons.LOG.info("Kubedns successfully deployed")


def configure_dns(os_manager, timeout):

    configure_dns_kubesys(os_manager=os_manager,
                          timeout=timeout)

    if settings.OPENSTACK_DEPLOY_FAKE_DEPLOYMENT:
        commons.LOG.info(
            "Wont configure dns, since fake openstack deployment")
        return

    commons.LOG.info("Get ingress ip")
    ingress_service = os_manager.api.services.get(
        "ingress", "openstack")

    wait = wait_utils.Waiter(os_manager, timeout)

    wait.wait_pass(wait.read_checker, k8s_object=ingress_service,
                   expected=Exception)

    ingress_service = ingress_service.read()
    ingress_ip = ingress_service.status.load_balancer.ingress[0].ip
    ingress_ip = ingress_ip or ingress_service.spec.cluster_ip

    commons.LOG.info("Update coredns yaml")
    core_dns_yaml = file_utils.get_release("ci/30-coredns.yaml")
    dns_data = file_utils.get_yaml_content(core_dns_yaml)
    zone_files = dns_data["spec"]["releases"][0]["values"]["zoneFiles"]
    dns_content = zone_files[0]["contents"].replace("1.2.3.4", ingress_ip)
    kubectl = kubectl_utils.Kubectl(namespace=NAMESPACE.kube_system)

    if settings.OPENSTACK_DEPLOY_IAM or settings.OPENSTACK_USE_MGMT_IAM:
        keystone_client_name = os_manager.api.pods.list_starts_with(
            "keystone-client", NAMESPACE.openstack
        )[0].read().metadata.name

        kubectl.exec(
            keystone_client_name,
            "bash -c 'openstack group create m:os@admin"
            " && openstack role add --project admin --group "
            "m:os@admin member'",
            namespace=NAMESPACE.openstack)

    if settings.OPENSTACK_DEPLOY_IAM:
        iam_service = os_manager.api.services.get(
            "openstack-iam-keyclo-http", "iam").read()

        iam_ext_ip = iam_service.status.load_balancer.ingress[0].ip
        iam_domain = "\n          keycloak.it.just.works.   " \
                     " IN      A       {}".format(iam_ext_ip)
        if iam_domain not in dns_content:
            dns_content = dns_content + iam_domain

    zone_files[0]["contents"] = dns_content

    commons.LOG.info("Save new coredns yaml")
    new_dns_yaml = file_utils.get_new_name(core_dns_yaml)
    file_utils.save_to_yaml(dns_data, new_dns_yaml)

    commons.LOG.info("Deploy coredns")
    kubectl.apply(new_dns_yaml)

    commons.LOG.info("Wait for coredns service")
    core_dns_service = os_manager.api.services.get(
        "coredns-coredns", NAMESPACE.coredns)

    wait.wait_pass(wait.read_checker, k8s_object=core_dns_service,
                   expected=Exception)

    commons.LOG.info("coredns successfully configured")


def configure_dns_kubesys(os_manager, timeout):
    commons.LOG.info("Check is coredns or kubedns")
    is_core_dns = helpers.is_core_dns(os_manager)
    kubectl = kubectl_utils.Kubectl(namespace=NAMESPACE.kube_system,
                                    kubeconfig=os_manager.kubeconfig)

    if is_core_dns:
        coredns_config = os_manager.api.configmaps.get(
            "coredns", NAMESPACE.kube_system)
        dns_content = coredns_config.read().data["Corefile"]
        os_dns_section = """
it.just.works:53 {
    errors
    cache 30
    forward . %s
}""" % helpers.get_external_dns_ip(os_manager)
        if os_dns_section not in dns_content:
            dns_content_re = re.sub(r"it.just.works:53 \{([^\$]+)\}",
                                    os_dns_section, dns_content, 0)
            if dns_content_re == dns_content:
                dns_content = dns_content + os_dns_section
            else:
                dns_content = dns_content_re
            coredns_config.patch(
                {"data": {"Corefile": dns_content}})

        coredns_pods = os_manager.api.pods.list_starts_with(
            "coredns", NAMESPACE.kube_system)
        for pod in coredns_pods:
            kubectl.delete("pod", pod.name)

        wait = wait_utils.Waiter(os_manager, timeout, NAMESPACE.kube_system)
        wait.pods("coredns", replicas=1)

    commons.LOG.info("coredns successfully configured")
