import json
import yaml
import pytest
import unittest

from si_tests import settings
from si_tests import logger

from si_tests.managers import introspect_manager

LOG = logger.logger


def _introspect_MKE_7914(openstack_client):
    """Check for UCP installer failures evidences"""
    JIRA = "MKE-7914"

    introspect = introspect_manager.IntrospectBootstrap(openstack_client, JIRA)
    introspect.init()
    bastion_ip = introspect.bastion_ip
    private_key_path = introspect.private_key_path
    PREF = introspect.PREF
    remote = introspect.remote

    collected_data = {}

    for mgmt_ip in introspect.mgmt_ips:
        ssh_cmd = introspect.get_ssh_cmd(mgmt_ip, private_key_path, bastion_ip)
        details = introspect.get_openstack_instance_details(ssh_cmd, mgmt_ip)
        if not details:
            continue
        hostname = details['hostname']
        collected_data[mgmt_ip] = details

        LOG.debug(f"{PREF} Show docker containers on the mgmt node")
        try:
            res = remote.check_call(f"{ssh_cmd} 'docker ps --all || true'")
            collected_data[mgmt_ip]['containers'] = res.stdout_str.splitlines()
        except Exception as e:
            LOG.error(f"{PREF} Unable to get docker containers list "
                      f"from {mgmt_ip}: {e}")
            continue

        LOG.debug(f"{PREF} Check for the bug evidence")
        try:
            res = remote.execute(
                f"{ssh_cmd} \"sudo fgrep -R -B3 "
                f"'expected response status code 200 but got 500' "
                f"/var/log/lcm/runners/\"")

            res1 = remote.execute(
                f"{ssh_cmd} \"sudo fgrep 'OCI runtime create failed' "
                f"/var/log/syslog\"")
            if not res.stdout and not res1.stdout:
                continue
            collected_data[mgmt_ip]["/var/log/lcm/runners/"] = res.stdout_str
            collected_data[mgmt_ip]["/var/log/syslog"] = res1.stdout_str
            # Something happened
            if res.stdout_str:
                LOG.info(f"{PREF} Found evidence of the bug on the node "
                         f"{hostname} in LCM logs:\n{res.stdout_str}")
            if res1.stdout_str:
                LOG.info(f"{PREF} Found evidence of the bug on the node "
                         f"{hostname} in /var/log/syslog:\n{res1.stdout_str}")

                res2 = remote.execute(
                    f"{ssh_cmd} docker ps --all | fgrep 'Exited'")
                if res2.stdout_str:
                    collected_data[mgmt_ip][
                        "docker_ps_exited"] = res2.stdout_str
                    LOG.info(f"{PREF} Containers in 'Exited' status on "
                             f"the node {hostname}:\n{res2.stdout_str}")

        except Exception as e:
            LOG.error(f"{PREF} Unable to check the evidence of the bug "
                      f"on the node {hostname}: {e}")
            continue

    LOG.debug(f"{PREF} Save the collected data to artifacts")
    LOG.debug(yaml.dump(collected_data))
    with open(f'{settings.ARTIFACTS_DIR}/introspect_{JIRA}.yaml',
              mode='w') as f:
        f.write(yaml.dump(collected_data))


def _introspect_PRODX_9238(openstack_client):
    """Measure disk speed"""
    JIRA = "PRODX-9238"

    introspect = introspect_manager.IntrospectBootstrap(openstack_client, JIRA)
    introspect.init()
    bastion_ip = introspect.bastion_ip
    private_key_path = introspect.private_key_path
    PREF = introspect.PREF
    remote = introspect.remote

    collected_data = {}

    for mgmt_ip in introspect.mgmt_ips:
        ssh_cmd = introspect.get_ssh_cmd(mgmt_ip, private_key_path, bastion_ip)
        details = introspect.get_openstack_instance_details(ssh_cmd, mgmt_ip)
        if not details:
            continue
        hostname = details['hostname']
        hypervisor_hostname = details.get(
            'os', {}).get('hypervisor_hostname', '-')
        collected_data[mgmt_ip] = details

        collected_data[mgmt_ip]['disk_speed'] = {}

        LOG.debug(f"{PREF} Perform the local disk speed measurement "
                  f"on the mgmt node ({JIRA})")
        try:
            # Default write, as most of applications do
            res = remote.check_call(
                f"{ssh_cmd} 'sudo dd if=/dev/urandom of=/.dd_speedtest_file "
                f"bs=1024 count=100000 2>&1|tail -n1'")
            write_default = res.stdout_str
            collected_data[mgmt_ip]['disk_speed'][
                'write_default'] = res.stdout_str

            # Write with direct sync, as a database server may be configured
            res = remote.check_call(
                f"{ssh_cmd} 'sudo dd if=/dev/urandom of=/.dd_speedtest_file "
                f"bs=1024 count=1000 oflag=dsync 2>&1|tail -n1'")
            write_direct_sync = res.stdout_str
            collected_data[mgmt_ip]['disk_speed'][
                'write_direct_sync'] = res.stdout_str

            w_default = write_default.split(",")[-1]
            w_dsync = write_direct_sync.split(",")[-1]

            LOG.info(f"{PREF} "
                     f"{hypervisor_hostname}/{hostname}: "
                     f"{w_dsync} (dsync) / {w_default} (default)")

            remote.check_call(f"{ssh_cmd} 'sudo rm /.dd_speedtest_file'")
        except Exception as e:
            LOG.error(f"{PREF} Unable to collect the disk speed info "
                      f"from {mgmt_ip}: {e}")
            continue

    LOG.debug(f"{PREF} Save the collected data to artifacts")
    LOG.debug(yaml.dump(collected_data))
    with open(f'{settings.ARTIFACTS_DIR}/introspect_{JIRA}.yaml',
              mode='w') as f:
        f.write(yaml.dump(collected_data))


def _get_ipam_data(kaas_manager):
    clusters = kaas_manager.get_clusters()

    data = {}  # resulting data

    for cluster in clusters:
        if cluster.provider != 'BaremetalClusterProviderSpec':
            LOG.warning(f"Skip IPAM check for provider "
                        f"{cluster.provider} in {cluster.name}")
            continue

        data[cluster.name] = {}
        LOG.info(f"Collecting IPAM data from cluster {cluster.name}")

        ns = kaas_manager.get_namespace(cluster.namespace)
        ipamhosts = [host.read() for host in ns.get_ipam_hosts()]
        # ipaddrs = [addr.read() for addr in ns.get_ipam_ipaddrs()]
        # subnets = ns.get_ipam_subnets()

        for ipamhost in ipamhosts:
            nics_data = {
                'gateway': '',
                'ifaces': {},
                'netconfig': []
            }
            nics_data['netconfig'] = ipamhost.status.get('netconfigV2', [])
            nics = ipamhost.status['nicMACmap'] or []

            for nic in nics:
                if 'mac' not in nic:
                    LOG.error(f"The following NIC on the host {ipamhost.name} "
                              f"doesn't have MAC: {nic}")
                    continue
                if nic['mac'] in nics_data['ifaces']:
                    raise Exception(f"Two NICs with the same MAC found on the "
                                    f"host {ipamhost.name}:\n{nic}\n"
                                    f"{nics_data['ifaces'][nic['mac']]}")

                if 'ip' in nic:
                    # remember only ifaces that has IPs
                    nics_data['ifaces'][nic['mac']] = {
                        'name': nic['name'],
                        'ip': nic['ip'],
                    }
                if 'gateway' in nic:
                    if 'primary' not in nic:
                        raise Exception(f"The following non-primary NIC has "
                                        f"gateway for the host "
                                        f"{ipamhost.name}: {nic}")
                    if nics_data['gateway']:
                        raise Exception(
                            f"The following primary NIC for the host "
                            f"{ipamhost.name} with the gateway "
                            f"{nics_data['gateway']} has another "
                            f"extra gateway: {nic}")
                    nics_data['gateway'] = nic['gateway']

            ipamhost_name = ipamhost.metadata.name
            data[cluster.name][ipamhost_name] = nics_data

    LOG.debug(yaml.dump(data))
    return data


def _get_hosts_data(kaas_manager):
    clusters = kaas_manager.get_clusters()
    data = {}  # resulting data
    for cluster in clusters:
        if cluster.provider != 'BaremetalClusterProviderSpec':
            LOG.warning(f"Skip network config check for provider "
                        f"{cluster.provider} in {cluster.name}")
            continue

        data[cluster.name] = {
            'machines': {},
            'loadbalancer': '',
        }
        provider_status = cluster.data['status']['providerStatus']
        loadbalancer = provider_status['loadBalancerHost']
        data[cluster.name]['loadbalancer'] = loadbalancer
        LOG.info(f"Cluster: {cluster.name} with loadbalancer {loadbalancer}")

        machines = cluster.get_machines()
        for machine in machines:
            if machine.machine_status != 'Ready':
                LOG.error(f"Skipping machine {machine.name} because of wrong "
                          f"status: {machine.machine_status}")
                continue

            nics_data = {
                'machine_types': machine.machine_types,
                'gateway': '',
                'vip': '',
                'ifaces': {},
            }

            # Get the addresses from interfaces
            # TODO(ddmitriev): consider to use a daemonset, or job, or a pod
            #     with enabled hostnetwork to get network configuration from
            #     the cluster without ssh access
            #     PRODX-11650
            private_key = settings.KAAS_CHILD_CLUSTER_PRIVATE_KEY_FILE
            username = settings.KAAS_CHILD_CLUSTER_SSH_LOGIN
            res = machine._run_cmd("ip -j a", verbose=False,
                                   ssh_key=private_key, ssh_login=username)
            ip_a = res.stdout_str
            ips = json.loads(ip_a)

            vip = ''
            for nic in ips:
                addresses = []
                nic_name = nic.get('ifname')
                mac = nic['address']

                if nic.get("qdisc", "noqueue") == "noqueue" and \
                        nic_name != "lo":
                    # Skip virtual interfaces with 'noqueue' class
                    # except local interface
                    continue
                if 'addr_info' not in nic:
                    continue
                if len(nic['addr_info']) == 0:
                    continue
                if nic_name is None:
                    LOG.error(f"Interface with MAC {mac} on the machine "
                              f"{cluster.name}/{machine.name} don't have "
                              f"the 'ifname' attribute while has "
                              f"some addresses assigned to it, skipping")
                    continue

                for addr in nic['addr_info']:
                    # VIP properties: no broadcast, /32, unique
                    if addr['local'] == loadbalancer:
                        LOG.info(f"Found VIP {addr['local']} on the node "
                                 f"{cluster.name}/{machine.name} "
                                 f"iface {nic_name}")
                        if vip:
                            raise Exception(f"VIP {addr['local']} is found on "
                                            f"multiple interfaces on the node "
                                            f"{cluster.name}/{machine.name}")
                        vip = addr['local']
                    else:
                        address = f"{addr['local']}/{addr['prefixlen']}"
                        addresses.append(address)

                # assume that there is no two interfaces with the same name on
                # the same node
                nics_data['ifaces'][nic_name] = {
                    'addresses': sorted(addresses),
                    'mac': mac,
                }

            nics_data['vip'] = vip

            # Get the default gateway address
            res = machine._run_cmd(
                "route -n | grep ' UG ' | grep -e '^0.0.0.0'",
                verbose=False, ssh_key=private_key, ssh_login=username)
            ip_r = res.stdout_str
            default_gws = [route for route in ip_r.splitlines()]
            if len(default_gws) > 1:
                raise Exception(f"More than one default gateways found "
                                f"on the node {machine.name}: {default_gws}")
            elif len(default_gws) == 1:
                default_gw = default_gws[0].split()[1]
                nics_data['gateway'] = default_gw

            data[cluster.name]['machines'][machine.name] = nics_data

    LOG.debug(yaml.dump(data))
    return data


@pytest.fixture(scope='session')
def introspect_MKE_7914(openstack_client):
    """Check the result of management cluster bootstrap test

    Collect the data specific to the bug:
    https://mirantis.jira.com/browse/MKE-7914
    """
    # pass the bootstrap test first
    yield

    try:
        _introspect_MKE_7914(openstack_client)
    except Exception as e:
        LOG.error(f"Exception in introspect_MKE_7914:\n{e}")
        # Do not fail the testcase because of fixture errors
        pass


@pytest.fixture(scope='session')
def introspect_PRODX_9238(openstack_client):
    """Check the result of management cluster bootstrap test

    Collect the data specific to the bug:
    https://mirantis.jira.com/browse/PRODX-9238
    """
    # pass the bootstrap test first
    yield

    try:
        _introspect_PRODX_9238(openstack_client)
    except Exception as e:
        LOG.error(f"Exception in introspect_PRODX_9238:\n{e}")
        # Do not fail the testcase because of fixture errors
        pass


@pytest.fixture(scope='session')
def introspect_PRODX_9696(kaas_manager):
    """Check that the network configuration is not changed during the test"""

    # ------------------
    # 1. Get info from ipamhosts: MAC addresses, interfaces and IPs
    # 2. TODO(ddmitriev, PRODX-11595): Check that ipaddrs contain the same
    #    address as in imaphosts
    # 3. Check that BM hosts has the same IP address on the specified
    #    interfaces from ipamhosts
    # 4. Check that BM host has the 'default' route which match the gateway
    #    from ipamhosts
    # 5. Check that a BM host has the IP address for keepalived VIP:
    #    cluster.data['status']['providerStatus']['loadBalancerHost']
    #    which is unique and is *ONLY* on 'control' node types

    # Collect data before the test
    ipam_data1 = _get_ipam_data(kaas_manager)

    if not ipam_data1.keys():
        LOG.warning("[PRODX-9696] Skip network config check "
                    "because no clusters with suitable provider found")
        yield
        return

    if settings.DISABLE_IP_CONSISTENCY_CHECK:
        LOG.warning("PRODX_9696: Check for IP addresses on nodes is skipped")
        hosts_data1 = {}
    else:
        hosts_data1 = _get_hosts_data(kaas_manager)

    # Run the test
    yield

    # Collect data after the test
    ipam_data2 = _get_ipam_data(kaas_manager)
    if settings.DISABLE_IP_CONSISTENCY_CHECK:
        LOG.warning("PRODX_9696: Check for IP addresses on nodes is skipped")
        hosts_data2 = {}
    else:
        hosts_data2 = _get_hosts_data(kaas_manager)

    errors = []
    # 1. Compare IPAM data
    for cluster_name1 in ipam_data1.keys():
        if cluster_name1 not in ipam_data2:
            LOG.error(f"Cluster {cluster_name1} is disappeared after test,"
                      " skipping check for the cluster")
            continue
        ipam_hosts1 = ipam_data1[cluster_name1]
        ipam_hosts2 = ipam_data2[cluster_name1]
        for ipam_host1 in ipam_hosts1.keys():
            if ipam_host1 not in ipam_hosts2:
                errors.append(f"Ipamhost {cluster_name1}/{ipam_host1} "
                              f"disappeared from ipamhosts after test")
                continue
            try:
                unittest.TestCase().assertMultiLineEqual(
                    yaml.dump(ipam_hosts1[ipam_host1]['netconfig']),
                    yaml.dump(ipam_hosts2[ipam_host1]['netconfig'])
                )
            except AssertionError as e:
                errors.append(
                    f"Netconfig on the Ipamhost {cluster_name1}/"
                    f"{ipam_host1} has been changed: {e}")

            gw1 = ipam_hosts1[ipam_host1]['gateway']
            gw2 = ipam_hosts2[ipam_host1]['gateway']
            if gw1 != gw2:
                errors.append(
                    f"Default gateway on the Ipamhost {cluster_name1}/"
                    f"{ipam_host1} has been changed in ipamhosts: was "
                    f"'{gw1}', became '{gw2}'")
            ifaces1 = ipam_hosts1[ipam_host1]['ifaces']
            ifaces2 = ipam_hosts2[ipam_host1]['ifaces']
            for mac1 in ifaces1.keys():
                if mac1 not in ifaces2:
                    errors.append(
                        f"Interface with MAC {mac1} became "
                        f"unconfigured on the ipamhost {cluster_name1}/"
                        f"{ipam_host1} after test, was: {ifaces1[mac1]}")
                    continue
                iface1 = ifaces1[mac1]
                iface2 = ifaces2[mac1]
                if iface1['name'] != iface2['name']:
                    errors.append(
                        f"Interface with MAC {mac1} change the name on the "
                        f"ipamhost {cluster_name1}/{ipam_host1} after test, "
                        f"was {iface1['name']}, became {iface2['name']}")
                if iface1['ip'] != iface2['ip']:
                    errors.append(
                        f"Interface with MAC {mac1} change the IP address on "
                        f"the ipamhost {cluster_name1}/{ipam_host1} after "
                        f"test, was {iface1['ip']}, became {iface2['ip']}")

    # 2. Compare hosts data
    for cluster_name1 in hosts_data1.keys():
        if cluster_name1 not in hosts_data2:
            # do not double the error message about missing cluster
            continue
        cluster_data1 = hosts_data1[cluster_name1]
        cluster_data2 = hosts_data2[cluster_name1]
        if cluster_data1['loadbalancer'] != cluster_data2['loadbalancer']:
            errors.append(f"Loadbalancer is changed on the cluster "
                          f"{cluster_name1} after test, was "
                          f"{cluster_data1['loadbalancer']}, "
                          f"became {cluster_data2['loadbalancer']}")

        machines_data1 = cluster_data1['machines']
        machines_data2 = cluster_data2['machines']
        vip1 = ''
        vip1_machine_name = ''
        vip2 = ''
        vip2_machine_name = ''
        for machine_name1 in machines_data1.keys():
            if machine_name1 not in machines_data2:
                errors.append(f"Machine {machine_name1} disappeared after "
                              f"test from the cluster {cluster_name1}")
                continue
            machine1 = machines_data1[machine_name1]
            machine2 = machines_data2[machine_name1]
            if machine1['gateway'] != machine2['gateway']:
                errors.append(
                    f"Default gateway on the machine {cluster_name1}/"
                    f"{machine_name1} has been changed from "
                    f"{machine1['gateway']} to {machine2['gateway']}")

            # Check VIP on machines before test
            if machine1['vip']:
                if vip1 != '':
                    errors.append(
                        f"Found VIP assigned to multiple machines in the "
                        f"cluster {cluster_name1} before test: {vip1} on the "
                        f"machine {vip1_machine_name} and {machine1['vip']} "
                        f"on the machine {machine_name1}")
                else:
                    vip1 = machine1['vip']
                    vip1_machine_name = machine_name1

                if 'control' not in machine1['machine_types']:
                    errors.append(
                        f"Found VIP assigned to the non-controller machine "
                        f"in the cluster {cluster_name1} before test: "
                        f"{machine1['vip']} on the machine {machine_name1} "
                        f"({machine1['machine_types']})")

            # Check VIP on machines after test
            if machine2['vip']:
                if vip2 != '':
                    errors.append(
                        f"Found VIP assigned to multiple machines in the "
                        f"cluster {cluster_name1} after test: {vip2} on the "
                        f"machine {vip2_machine_name} and {machine2['vip']} "
                        f"on the machine {machine_name1}")
                else:
                    vip2 = machine2['vip']
                    vip2_machine_name = machine_name1

                if 'control' not in machine2['machine_types']:
                    errors.append(
                        f"Found VIP assigned to the non-controller machine in "
                        f"the cluster {cluster_name1} after test: "
                        f"{machine2['vip']} on the machine {machine_name1} "
                        f"({machine2['machine_types']})")

            ifaces1 = machine1['ifaces']
            ifaces2 = machine2['ifaces']
            for iface1_name in ifaces1.keys():
                if iface1_name not in ifaces2:
                    errors.append(
                        f"Interface {iface1_name} on the machine "
                        f"{cluster_name1}/{machine_name1} became unconfigured "
                        f"after test, was: {ifaces1[iface1_name]}")
                    continue
                ifmac1 = ifaces1[iface1_name]['mac']
                ifmac2 = ifaces2[iface1_name]['mac']
                if ifmac1 != ifmac2:
                    errors.append(
                        f"Interface {iface1_name} on the machine "
                        f"{cluster_name1}/{machine_name1} changed "
                        f"the MAC: was {ifmac1}, became {ifmac2}")
                addresses1 = ifaces1[iface1_name]['addresses']
                addresses2 = ifaces2[iface1_name]['addresses']
                if addresses1 != addresses2:
                    errors.append(
                        f"Interface {iface1_name} on the machine "
                        f"{cluster_name1}/{machine_name1} changed the IP "
                        f"address configuration: was {addresses1}, "
                        f"became {addresses2}")

        if vip1 != vip2:
            errors.append(
                f"VIP address is changed on the cluster {cluster_name1}: "
                f"was {vip1} on the machine {vip1_machine_name}, "
                f"became {vip2} on the machine {vip2_machine_name}")

    # 3. TODO(ddmitriev, PRODX-11595): Compare IPAM and machines data

    # 4. Raise an exception if any error has been found
    if errors:
        msg = '\n'.join(errors)
        raise Exception(f"Found the following errors:\n{msg}")
    else:
        LOG.info("PRODX-9696: PASSED. Network settings in IPAM and machines "
                 "were not changed")
