import yaml
import time
import json
import kubernetes
import socket


from retry import retry
from packaging import version
from urllib3.exceptions import MaxRetryError
from si_tests import settings
from si_tests import logger
from si_tests.utils import utils, waiters, exceptions, templates, helpers
from kubernetes.client.rest import ApiException
from websocket._exceptions import WebSocketBadStatusException
from OpenSSL.SSL import Connection, Context, SSLv3_METHOD, TLSv1_2_METHOD

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from si_tests.managers.kaas_manager import Cluster

LOG = logger.logger

# PRODX-143
# fluentd-notifications and telegraf-openstack Deployments' pods are in
# the init state till the corresponding Secrets aren't created by Openstack.
# It is expected in any MOS ClusterRelease till the OpenStackDeployment
# creation.


class ClusterCheckManager(object):
    """Cluster check manager"""

    EXCLUDED_PODS = [
        'fluentd-notifications',
        'telegraf-openstack',
        'ui-e2e-test',
        'stacklight/fluentd-notifications',
        'stacklight/telegraf-openstack'
    ]
    EXCLUDED_CONDITIONS = {
        'Kubernetes': 'stacklight/telegraf-openstack got 0/',
        'StackLight': 'stacklight/telegraf-openstack got 0/',
    }
    EXCLUDED_JOBS = [
        'elasticsearch-curator',
    ]

    def __init__(self, cluster: "Cluster"):
        self._cluster: "Cluster" = cluster
        self.deploys_generation = {}
        self.deploys_next_generation = {}

    @property
    def cluster(self) -> "Cluster":
        return self._cluster

    @property
    def k8sclient(self):
        return self._cluster.k8sclient

    def get_deploys_generation(self, namespace=None, deploys=None):
        deploys_generation = {}
        if not deploys and not namespace:
            deploys = self.cluster.k8sclient.deployments.list_all()
        if namespace and not deploys:
            deploys = [d for d in self.cluster.k8sclient.deployments.list(
                namespace=namespace)]
        for deploy in deploys:
            generation = deploy.read().to_dict().get('status').get(
                'observed_generation')
            deploys_generation[deploy.name] = int(generation)
        return deploys_generation

    def store_deploys_generation(self, namespace=None, deploys=None):
        generations = self.get_deploys_generation(
            namespace=namespace, deploys=deploys)
        self.deploys_generation = generations
        self.deploys_next_generation = {k: v+1 for k, v in generations.items()}

    def wait_deploy_generation_changed(self, namespace=None, deploys=None,
                                       timeout=600, interval=60):
        LOG.info("Waiting for generations are changed")
        for k, v in self.deploys_generation.items():
            LOG.info("{}: {} -> {}".format(
                k, v, self.deploys_next_generation[k]))

        waiters.wait(lambda: self.get_deploys_generation(
            namespace=namespace,
            deploys=deploys) == self.deploys_next_generation,
            timeout=timeout, interval=interval,
            timeout_msg="Some generations are not changed. Current "
                        "generations info: \n{} but should be: \n{}".format(
                yaml.dump(self.get_deploys_generation(namespace=namespace,
                                                      deploys=deploys)),
                yaml.dump(self.deploys_next_generation)))

    def get_ceph_health_status(self):
        cmd_status = ['/bin/sh', '-c', 'ceph -s -f json']
        ceph_tools_pod = self.cluster.get_ceph_tool_pod()
        return yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health', {}).get('status')

    def wait_ceph_health_status(self, expected_status='HEALTH_OK', timeout=600,
                                interval=30):
        waiters.wait(
            lambda: self.get_ceph_health_status() == expected_status,
            timeout=timeout, interval=interval,
            timeout_msg=f"Ceph status failed to became "
                        f"{expected_status} in {timeout} sec")

    def check_ceph_pvc(self):
        """
        Simple test for ceph pvc in child-cluster
        Scenario:
            1. Spawn pod with attached pv
            2. Write some data to file in pv
            3. Detach and delete pod
            4. Spawn new pod and attach to old pv
            5. Check data content
            6. Delete pod; delete pvc; Delete ns
        """
        pvc_name = 'test-pv-claim'
        pod_name = 'test-pod'
        mount_path = '/data/test'
        file_path = f'{mount_path}/text.txt'
        ns = f'test-ns-{utils.gen_random_string(6)}'
        expected_string = utils.gen_random_string(10)
        storage_classes = \
            self.cluster.k8sclient.api_storage.list_storage_class().to_dict()[
                'items']
        # Search storage class based on name
        storage_class = next(x for x in storage_classes if
                             x['metadata']['name'].startswith('kubernetes-'))
        storage_class_name = storage_class['metadata']['name']
        pvc_template = templates.render_template(
            settings.CEPH_PVC_YAML, {
                'PVC_NAME': pvc_name,
                'PVC_SIE': '1Gi',
                'STORAGE_CLASS_NAME': storage_class_name})
        pvc_json_body = json.dumps(
            yaml.load(pvc_template, Loader=yaml.SafeLoader))
        pod_template = templates.render_template(
            settings.CEPH_PV_CHECK_POD_YAML,
            {'POD_NAME': 'test-pod',
             'PVC_NAME': pvc_name,
             'MOUNT_PATH': mount_path})
        pod_json_body = json.dumps(
            yaml.load(pod_template, Loader=yaml.SafeLoader))
        LOG.info('Check Ceph pvc')
        LOG.info(f'Creating ns "{ns}"')
        namespace = self.cluster.k8sclient.namespaces.create(
            name=ns,
            namespace='default',
            body={'metadata': {'name': ns}})
        LOG.info('Creating pvc')
        pvc = self.cluster.k8sclient.pvolumeclaims.create(name=pvc_name,
                                                          namespace=ns,
                                                          body=json.loads(
                                                              pvc_json_body))
        LOG.info('Create pod')
        pod = self.cluster.k8sclient.pods.create(name=pod_name,
                                                 namespace=ns,
                                                 body=json.loads(
                                                     pod_json_body))
        pod.wait_phase('Running')

        LOG.info('Create new file in attached pv')
        pod.exec(
            ['/bin/sh', '-c',
             f'echo "{expected_string}" > {file_path}'])
        pod.check_file_exists(file_path)

        LOG.info('Delete pod')
        pod.delete()
        assert not pod.exists()

        LOG.info('Re-create pod')
        pod = self.cluster.k8sclient.pods.create(name=pod_name,
                                                 namespace=ns,
                                                 body=json.loads(
                                                     pod_json_body))
        pod.wait_phase('Running')

        LOG.info('Check file content')
        pod.check_file_exists(file_path)
        res = pod.exec(['/bin/sh', '-c', f'cat {file_path}'])
        assert res.rstrip() == expected_string

        LOG.info('Delete pod')
        pod.delete()
        LOG.info('Delete pvc')
        pvc.delete()
        LOG.info('Delete namespace')
        namespace.delete()
        LOG.info('Check Ceph pvc done')

    def check_k8s_nodes(self, timeout=360, interval=10):
        LOG.info("Checking k8s nodes status")
        try:
            waiters.wait(
                lambda: self.k8sclient.nodes.all_ready(),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            nodes = {key: value for key, value in
                     self.k8sclient.nodes.list_statuses().items()
                     if value != 'True'}
            err = f"Timeout waiting for nodes to be Ready. " \
                  f"After {timeout} sec next nodes are not Ready: {nodes}"
            raise exceptions.TimeoutError(err)
        LOG.info("All k8s nodes are Ready")

    def wait_k8s_node_status(self, nodename='', expected_status='',
                             timeout=360, interval=10):
        LOG.info(f"Wait for k8s node stay in {expected_status} status.")
        try:
            waiters.wait(
                lambda: self.k8sclient.nodes.node_status(nodename,
                                                         expected_status),
                timeout=timeout, interval=interval,
                timeout_msg=f"Node not in status {expected_status}")
        except exceptions.TimeoutError:
            err = f"Timeout waiting for node {nodename} to be "\
                  f"{expected_status} After {timeout} sec."
            raise exceptions.TimeoutError(err)
        LOG.info(f"k8s node {nodename} is in status <{expected_status}>.")

    # note: check_machines_status will do similar thing, but we still need this
    def check_cluster_nodes(self, timeout=1800, interval=10,
                            expected_state='Ready'):
        LOG.info("Checking cluster nodes status")
        try:
            waiters.wait(lambda: self.cluster.cluster_status.lower() ==
                         expected_state.lower(),
                         timeout=timeout, interval=interval)
            LOG.info(f"All cluster nodes are {expected_state}")
        except exceptions.TimeoutError:
            status = self.cluster.data['status']
            if not status:
                raise Exception("Cluster is in Pending status")
            nodes = status.get('providerStatus', {}).get('nodes', {})
            if not nodes:
                raise Exception(f"Nodes for cluster are not found {status}")
            err = f"Timeout waiting for cluster " \
                  f"nodes to be {expected_state} " \
                  f"after {timeout} sec: {nodes}"
            raise exceptions.TimeoutError(err)

    def check_pods_number(self, pod_name, ns, pods):
        """Wait till replica number restores"""
        def correct_number_of_pods(pods):
            # let's take number of pods we had
            number = len(pods)
            # extract old names (we expect to see the same pods)
            pods_names = [x.name for x in pods]
            new_list = self.k8sclient.pods.list(
                ns, field_selector="status.phase=Running"
            )
            # check if new list has same number of pods (with same names)
            new_pods_len = len([x for x in new_list if x.name in pods_names])
            return True if number == new_pods_len else False
        try:
            waiters.wait(
                lambda: correct_number_of_pods(pods),
                timeout=360, interval=10)
        except exceptions.TimeoutError:
            actual_num = len(self.k8sclient.pods.list_starts_with(
                pod_name, ns, field_selector="status.phase=Running"))
            LOG.error(f'Timeout waiting for correct number. '
                      f'Actual number: {actual_num}')
            raise TimeoutError

    def check_k8s_pods(self, phases=('Running', 'Succeeded'),
                       target_namespaces=None,
                       timeout=settings.WAIT_PODS_READY_TIMEOUT,
                       interval=30, pods_prefix=''):
        """Wait till all expected pods for cluster are in specified
           phase and have Ready=True for all containers
        Args:
            phases: list of expected pod phases
            target_namespaces: namespace (str) or namespaces (list)
                               where pods should be checked
            timeout: timeout to wait
            interval: time between checks
        """

        def wait_for_running_pods(pause=60):
            """Get the pods statuses and compare the restart counts

            :param pause: int, seconds, pause between restart checks
            """
            LOG.info("Check k8s pods status")
            pods_info = self.k8sclient.pods.check_pods_statuses(
                            target_namespaces=target_namespaces, phases=phases,
                            excluded_pods=self.EXCLUDED_PODS,
                            pods_prefix=pods_prefix)
            if not pods_info:
                return False

            LOG.info(f"Wait for {pause} seconds before the second check")
            time.sleep(pause)

            LOG.info("Check k8s pods status one more time")
            pods_info_new = self.k8sclient.pods.check_pods_statuses(
                            target_namespaces=target_namespaces, phases=phases,
                            excluded_pods=self.EXCLUDED_PODS,
                            pods_prefix=pods_prefix)
            if not pods_info_new:
                return False

            result = True
            for pod, info in pods_info.items():
                if pod not in pods_info_new:
                    LOG.warning(f"Pod {pod} disappeared after {pause} seconds")
                    result = False
                    continue
                cont_restarts_new = pods_info_new[pod]['containers_restarts']
                for container, restarts in info['containers_restarts'].items():
                    if container not in cont_restarts_new:
                        LOG.warning(f"Container {container} from pod {pod} "
                                    f"disappeared after {pause} seconds")
                        result = False
                        continue
                    restarts_new = cont_restarts_new[container]
                    if restarts_new > restarts:
                        LOG.warning(f"Container {container} from pod {pod} "
                                    f"has been restarted after {pause} seconds"
                                    f", current restarts: {restarts_new}")
                        result = False

            if result:
                LOG.info("All pods are in correct state")
            return result

        LOG.info("Checking k8s pods phase and containers status")
        if not target_namespaces:
            target_namespaces = self.cluster.expected_pods.keys()
        try:
            # Wait for pods status
            waiters.wait(wait_for_running_pods,
                         timeout=timeout, interval=interval)

            # Wait for jobs
            waiters.wait(
                lambda: self.k8sclient.jobs.check_jobs_completed(
                    target_namespaces=target_namespaces,
                    excluded_jobs=self.EXCLUDED_JOBS,
                    jobs_prefix=pods_prefix),
                timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            pods = self.k8sclient.pods.list_raw().to_dict()['items']
            if target_namespaces:
                if isinstance(target_namespaces, str):
                    target_namespaces = [target_namespaces]
                pods = [pod for pod in pods
                        if pod['metadata']['namespace'] in target_namespaces]
            utils.print_pods_status(pods)
            failed_pods = {}
            ready_text = " and Ready=True for " \
                         "each container"
            for pod, info in \
                    self.k8sclient.pods.get_pods_statuses(pods).items():
                # Check for the expected pod phase
                if info['phase'] not in phases:
                    failed_pods[pod] = info
                # Check for containers status inside the non-completed pod
                if (False in info['containers'].values()
                        and info['phase'] not in 'Succeeded'):
                    failed_pods[pod] = info
            err = f"Timeout waiting for pods statuses. " \
                  f"After {timeout} sec next pods are not in {phases} phase" \
                  f"{ready_text}: {failed_pods}"

            raise TimeoutError(err)
        LOG.info("All pods and jobs are in correct state")

    def check_ha_kill_proc_pods(self, pod_group, command):
        """Kill processed in every container in all cluster pods
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every namespace and
        expected pod entry in get_expected_pods
        Note: 'kill -- - 1' will kill all processes except pid 1

        Scenario:
            1. Iterate by each pod in a group
            2. Iterate by each container in a pod and
               kill processes in the container
            3. Wait till number of replicas will be restored
            4. Check pods and container statuses in this group
               (Running and Ready)
            5. Check all pods phases (must be Running or Completed)
               (in fixture)

        Expected result - pods are recreated, number of replicas is restored.
        """
        if settings.RUN_HA_TESTS:
            pod_name, pod_num, pods = pod_group

            ns = [pod.namespace for pod in pods][0]
            LOG.info("Iterate by each pod in a group")
            for pod in pods:
                # let's reread every pod to make sure it is healthy
                # before killing containers
                pod_data = pod.read()
                assert False not in [
                    x.ready for x in pod_data.status.container_statuses], \
                    f"Some containers are not ready for {pod.name} " \
                    f"pod: {pod_data.status.container_statuses}"
                LOG.info(f"Killing services in "
                         f"all containers of the pod: {pod.name}")
                for cont in pod_data.status.container_statuses:
                    restarts = cont.restart_count
                    LOG.info(f"Execute command {command} "
                             f"in {cont.name} container. "
                             f"Restarts: {restarts}")
                    try:
                        result = pod.exec(['/bin/sh', '-c', command],
                                          container=cont.name)
                        if result:
                            LOG.info(f"Killed with result: {result}")
                    except (ApiException, WebSocketBadStatusException) as e:
                        LOG.error(e)
                    continue
            LOG.info("Check number of pods")
            self.check_pods_number(pod_name, ns, pods)
            LOG.info(f"Check status for {pod_name} pods")
            # PRODX-12786
            if 'kibana' in pod_name:
                time.sleep(30)
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=600,
                                    interval=30)
            else:
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=300,
                                    interval=30)
        else:
            LOG.info("\nHA tests for services in pod containers were "
                     "skipped as RUN_HA_TESTS was set to False\n")

    def check_ha_delete_pods(self, pod_group):
        """Delete all pods in any cluster one by one
         Precondition - all expected pods and their replicas must be presented
         The following scenario is executed for every namespace and
         expected pod entry in get_expected_pods

         Scenario:
             1. Compare actual number of replicas for pod with expected
             2. Iterate by each replica
             3. Delete pod (1 replica)
             4. Wait till number of replicas will be restored
             5. Check pods statuses in this group (Running and Ready)

         Expected result - pods are recreated, number of replicas is restored.
         """
        if settings.RUN_HA_TESTS:
            pod_name, pod_num, pods = pod_group
            ns = [pod.namespace for pod in pods][0]
            LOG.info("Deleting pods")
            for pod in pods:
                LOG.info(f"Deleting pod: {pod.name}")
                pod.delete(timeout=180)
                # special handling of mariadb-server StatefulSets
                if 'mariadb-server' in pod.name:
                    LOG.info(f"Checking readiness "
                             f"for {pod.name} (StatefulSet)")
                    pod.wait_ready()
                LOG.info("Check number of pods")
                self.check_pods_number(pod_name, ns, pods)
                LOG.info(f"Check status for {pod_name} pods")
                self.check_k8s_pods(pods_prefix=pod_name,
                                    timeout=300,
                                    interval=30)
        else:
            LOG.info("\nHA tests for pods were skipped "
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_containerd_shim(self):
        """Kill all containerd-shim processes on every cluster node
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node and run kill command for containerd-shim processes
            2. Wait
            3. Check that all pods are Running and Ready

        Expected result - pods and their containers are Running and Ready.
        Number of replicas are the same.
        """
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            LOG.info("Checking precondition: all expected pods and "
                     "their replicas must be presented")
            self.check_actual_expected_pods()
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                node._run_cmd("for i in $(ps uax | grep containerd-shim | "
                              "grep -v grep | awk '{print $2}'); "
                              "do sudo kill -9 $i; done",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                time.sleep(60)
                LOG.info("Waiting for k8s to be ready")
                waiters.wait_pass(
                    lambda: self.k8sclient.pods.list_all(), timeout=120)
                LOG.info("Waiting for pods")
                self.check_k8s_pods(timeout=1200, interval=30)
                # TODO: maybe check pod restarts also?
                # additonal timeout ?
        else:
            LOG.info("\nHA tests to kill containerd-shim were skipped "
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_system_proc_in_cont(self):
        """Kill main process in mke system containers
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to each cluster node 1 by 1
            2. Collect list of container id's
            3. Kill containers 1 by 1
            4. Wait till all containers are running again
            5. Check that all pods are in a correct state

        Expected result - pods and their containers are Running and Ready.
        """
        # see PRODX-15408 for details
        cont_to_skip = ['ucp-kubectl',
                        'ucp-cluster-agent']
        if settings.RUN_HA_TESTS:
            self.check_actual_expected_pods()
            nodes = self.cluster.get_machines()
            LOG.debug("Check precondition: all expected pods and "
                      "their replicas must be presented.")
            KEY_FILE = settings.HA_TEST_PRIVATE_KEY_FILE
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont_list = node._run_cmd("docker ps --format '{{.Names}}' |"
                                          " grep -v k8s",
                                          ssh_key=KEY_FILE).stdout_str.\
                    split("\n")
                LOG.debug(f"Containers: {cont_list}")
                LOG.info("The state of the containers before the start of the "
                         "test")
                containers_before_ha = node._run_cmd("docker ps | grep -v k8s",
                                                     ssh_key=KEY_FILE).\
                    stdout_str
                LOG.debug(f"{containers_before_ha}")
                for container in cont_list:
                    _container = container.split(".")[0]
                    if _container not in cont_to_skip:
                        node._run_cmd(f"id=`docker ps | grep {_container} |"
                                      f"awk {{'print $1'}}`;"
                                      f"docker kill --signal=9 $id",
                                      ssh_key=KEY_FILE)
                        cmd = f"docker ps | grep {_container} |  wc -l"
                        waiters.wait(lambda:
                                     node._run_cmd(cmd, ssh_key=KEY_FILE).
                                     stdout_str == '1', timeout=60, interval=5)
                    else:
                        LOG.info(f"Container {_container} skipped because "
                                 f"it is in a skip list")
                LOG.info("The condition of the containers that passed the "
                         "test")
                containers_after_ha = node._run_cmd("docker ps | grep -v k8s",
                                                    ssh_key=KEY_FILE).\
                    stdout_str
                LOG.debug(f"{containers_after_ha}")
                LOG.info("Waiting for pods")
                self.check_k8s_pods(timeout=1200, interval=30)
        else:
            LOG.info("\nHA tests Kill processes in containers were skipped"
                     "as RUN_HA_TESTS was set to False\n")

    def check_ha_kill_ucp_cluster_agent(self):
        """Kill ucp-cluster-agent container
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to each cluster node 1 by 1
            2. Search for ucp-cluster-agent container
            3. Kill ucp-cluster-agent container
            4. Wait till container IS running
            5. Check that all pods are in a correct state
        Expected result - pods and container are Running and Ready.
        """

        def check_if_cluster_agent_is_running():
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont_count = \
                    node._run_cmd(f"docker ps | grep {container} | "
                                  f"wc -l", ssh_key=KEY_FILE).stdout_str
                if int(cont_count) == 1:
                    return True
                elif int(cont_count) > 1:
                    LOG.warning(
                        f"Found {cont_count} ucp-cluster-agent containers"
                        f"on {node.name} - only 1 was expected."
                    )
        if settings.RUN_HA_TESTS:
            LOG.debug("Check precondition: all expected pods and "
                      "their replicas must be presented.")
            self.check_actual_expected_pods()
            nodes = self.cluster.get_machines()
            container = 'ucp-cluster-agent'
            KEY_FILE = settings.HA_TEST_PRIVATE_KEY_FILE
            for node in nodes:
                LOG.info(f"Access {node.name}")
                cont = node._run_cmd(f"docker ps | grep {container} | wc -l",
                                     ssh_key=KEY_FILE).stdout_str
                if int(cont) != 0:
                    LOG.info(f"Container {container} found. Killing now")
                    node._run_cmd(f"id=`docker ps | grep {container} |"
                                  f" awk {{'print $1'}}`;"
                                  f"docker kill --signal=9 $id",
                                  ssh_key=KEY_FILE)
                    break
            else:
                raise Exception('ucp-cluster-agent was not found on any nodes')
            waiters.wait(lambda: check_if_cluster_agent_is_running(),
                         timeout=60, interval=10)
        self.check_k8s_pods(timeout=1200, interval=30)

    def check_ha_stop_containerd(self):
        """Stop containerd service on every cluster node (one at a time)
        Precondition - all expected pods and their replicas must be presented
        The following scenario is executed for every node

        Scenario:
            1. SSH to node
            2. Get number of containerd-shim processes
            3. Stop containerd service and make sure it is stopped
            4. Get number of containerd-shim processes and
               compare with previous value (must be the same)
            5. Check that all pods are Running and Ready
            6. Start containerd service

        Expected result - containerd-shim processes are running independently
        from containerd service and all containers in the pods are
        still operational even if containerd is down.
        """
        if settings.RUN_HA_TESTS:
            nodes = self.cluster.get_machines()
            LOG.info("Checking precondition: all expected pods and "
                     "their replicas must be presented")
            self.check_actual_expected_pods()
            for node in nodes:
                LOG.info(f"Accessing {node.name}")
                proc = node._run_cmd(
                    "pgrep -x containerd-shim | wc -l",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                LOG.info(f"There are containerd-shim {proc} processes")
                LOG.info("Stopping containerd")
                node._run_cmd("sudo service containerd stop",
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                time.sleep(5)
                node._run_cmd("pgrep -x containerd",
                              check_exit_code=True,
                              expected_codes=[1],
                              ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                proc2 = node._run_cmd(
                    "pgrep -x containerd-shim | wc -l",
                    ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE).stdout_str
                LOG.info(f"Now it is {proc} containerd-shim processes")
                try:
                    assert int(proc) == int(proc2) + 1
                    self.check_k8s_pods(timeout=1200, interval=30)
                except Exception as e:
                    LOG.error(e)
                finally:
                    LOG.info("Start containerd")
                    node._run_cmd("sudo service containerd start",
                                  ssh_key=settings.HA_TEST_PRIVATE_KEY_FILE)
                # TODO: maybe check pod restarts also?
        else:
            LOG.info("\nHA tests to stop containerd-shim were skipped"
                     "as RUN_HA_TESTS was set to False\n")

    def check_helmbundles(self, expected_spec=None, timeout=1800, interval=60):
        def helmbundles_ready(expected_spec):
            try:
                hb = self.cluster.get_helmbundle().data
                expected_spec = expected_spec or hb['spec']['releases']
            except Exception as e:
                LOG.error(e)
                return False

            hb_status = hb.get('status') or {}
            release_status = hb_status.get('releaseStatuses', {})

            spec_chart_names = ["{0}:{1}".format(
                                x['name'],
                                helpers.get_chart_version(x))
                                for x in expected_spec]
            status_chart_names = ["{0}:{1}".format(k, v['version'])
                                  for k, v in release_status.items()]
            LOG.debug(f"Current status "
                      f"of the charts:\n{yaml.dump(release_status)}")

            # Check that all charts with expected versions
            # are present in the helmbundle status
            charts_diff = set(spec_chart_names)\
                .difference(set(status_chart_names))
            if charts_diff:
                LOG.error(f"Mismatch between expected and actual helmbundles. "
                          f"Expected: {spec_chart_names}, "
                          f"actual: {status_chart_names}, "
                          f"difference: {charts_diff}")
                return False

            # Check that no charts in the helmbundle status has False
            failed_charts = [v for v in release_status.values()
                             if not v['success']]
            if failed_charts:
                LOG.error(f"The following charts "
                          f"in the failed state:\n"
                          f"{yaml.dump(failed_charts)}")
                return False

            LOG.info(f"The following charts "
                     f"have been deployed "
                     f"in the cluster:\n{yaml.dump(status_chart_names)}")
            return True

        LOG.info("Check helmbundles status")
        try:
            waiters.wait(lambda: helmbundles_ready(expected_spec),
                         timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            releaseStatuses = self.cluster.get_helmbundle().data\
                .get('status', {}).get('releaseStatuses', {})
            err = f"Not all helmbundles have been deployed: {releaseStatuses}"
            raise TimeoutError(err)
        LOG.info("All charts have been successfully deployed")

    def check_svc_access(self, remote):
        tmpl = templates.render_template(settings.SVC_CHECK_POD_YAML)
        LOG.debug(tmpl)
        json_body = json.dumps(yaml.load(tmpl, Loader=yaml.SafeLoader))
        LOG.info("Creating nginx pod")
        pod = self.cluster.k8sclient.pods.create(
            name="svc-check", namespace="default",
            body=json.loads(json_body)
        )
        pod.wait_phase('Running', timeout=1800, interval=10)
        LOG.info("Creating service")
        service = kubernetes.client.V1Service()
        service.metadata = kubernetes.client.V1ObjectMeta(
            name="svc-check-svc")
        service.spec = kubernetes.client.V1ServiceSpec(type="LoadBalancer")
        service.spec.selector = {"run": "svc-check"}
        service.spec.ports = [
            kubernetes.client.V1ServicePort(port=80, target_port=80)
        ]
        svc = self.cluster.k8sclient.services.create(
            name="svc-check-svc", namespace='default', body=service)
        LOG.info("Waiting for LB to have external ip/hostname")
        waiters.wait(lambda: svc.read().to_dict()['status'][
                                 'load_balancer']['ingress'] is not None,
                     timeout=500, interval=15,
                     timeout_msg='Timeout waiting for svc loadbalancer')
        ingress_addr = \
            svc.read().to_dict()['status']['load_balancer']['ingress'][0]
        if ingress_addr['ip']:
            address = ingress_addr['ip']
        else:
            address = ingress_addr['hostname'] or ''
        LOG.info(f"LB iaddress is {address}")

        command = f'wget -O- --no-proxy {address}' \
            if settings.KAAS_OFFLINE_DEPLOYMENT else f'wget -O- {address}'
        ret = remote.check_call(command, raise_on_err=False)
        pod.delete()
        svc.delete()
        assert ret.exit_code == 0 and "Welcome to nginx" in ret.stdout_str,\
            f"Service is not available. Response {ret.stdout_str}"

    def check_resource_discovery(self):
        if not self.cluster.is_management and not self.cluster.is_regional:
            LOG.warning("check_resource_discovery() is not suitable "
                        "for child clusters")
            return None
        if self.cluster.provider == 'AWSClusterProviderSpec':
            LOG.info("Checking aws resource discovery")
            kaas_awsresources = \
                self.cluster.k8sclient.kaas_awsresources.list(
                    namespace=self.cluster.namespace)
            if self.cluster.is_management:
                awsresources = [x for x in kaas_awsresources
                                if x.name == 'cloud-config'][0]
            else:
                awsresources = [x for x in kaas_awsresources
                                if x.name == self.cluster.name][0]
            aws_region = self.cluster.data['spec']['providerSpec'][
                'value']['region']
            regions = awsresources.regions
            LOG.debug(f"AWS regions: {regions}")
            instance_types = awsresources.instance_types(aws_region)
            LOG.debug(f"AWS instance types: {instance_types}")
            images = awsresources.images(aws_region)
            LOG.debug(f"AWS images: {images}")
            assert regions, \
                f"No regions were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
            assert instance_types, \
                f"No instance_types were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
            assert images, \
                f"No images were found in " \
                f"{awsresources.name} kaas_awsresources crd object"
        else:
            LOG.info("Skipping check_resource_discovery")

    def check_machines_status(self, expected_status='Ready',
                              timeout=6300, interval=30):
        def machines_status(expected_status):
            machines = self.cluster.get_machines()
            LOG.debug(f"Got machines: {machines}")
            machines_in_status = [x for x in machines
                                  if x.machine_status == expected_status]
            LOG.debug(f"Machines in {expected_status} "
                      f"status: {machines_in_status}")
            if len(machines) != len(machines_in_status):
                for machine in [x for x in machines
                                if x not in machines_in_status]:
                    LOG.info(f"{machine.name} machine is "
                             f"in status {machine.machine_status}")
                LOG.info(f"===Not all machines are "
                         f"in status {expected_status}===")
                raise Exception(f"Not all machines are "
                                f"in status {expected_status}")
            return True

        LOG.info(f"Wait {timeout} seconds until machines will be in "
                 f"the status <{expected_status}>")
        waiters.wait_pass(lambda: machines_status(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(Exception, ApiException))
        LOG.info("All machines are in expected status")

    def check_any_machine_exp_status(self, expected_status='Ready',
                                     timeout=1800, interval=30):
        def machines_status(exp_status):
            machines = self.cluster.get_machines()
            LOG.debug(f"Got machines: {machines}")
            machines_in_status = [x for x in machines
                                  if x.machine_status == exp_status]
            LOG.info(f"Machines in {exp_status} "
                     f"status: {machines_in_status}")
            if not machines_in_status:
                raise Exception(f"There are not any machines "
                                f"in status {exp_status}")
            return True

        LOG.info(f"Wait {timeout} seconds until any of machines will be in "
                 f"the status <{expected_status}>")
        waiters.wait_pass(lambda: machines_status(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(Exception, ApiException))
        LOG.info("There is machine in expected status")

    def check_machines_access_with_given_key(self,
                                             machines_list=None,
                                             verbose=False,
                                             ssh_key_file=None,
                                             ssh_login='mcc-user'):
        machines = machines_list if machines_list \
            else self.cluster.get_machines()
        ssh_key_file = ssh_key_file \
            if ssh_key_file else self.cluster.private_key
        machines_names = [m.name for m in machines]
        not_ready_machines = []
        for machine in machines:
            try:
                machine._run_cmd("hostname", verbose=verbose,
                                 ssh_key=ssh_key_file,
                                 ssh_login=ssh_login,
                                 reconnect=True)
            except Exception as e:
                LOG.debug("Can't connect to {}: {}".format(
                    machine.name, e))
                not_ready_machines.append(machine.name)
        if not_ready_machines:
            LOG.warning("No access to next machines: {} "
                        "using key {}".format(
                            not_ready_machines, ssh_key_file))
            return False
        LOG.info("Succesfully connected to all machines: {} "
                 "using key {}".format(
                     machines_names, ssh_key_file))
        return True

    def check_update_finished(self, timeout=6300, interval=15,
                              check_managed_clusters=False,
                              expected_status="Ready"):
        def update_check(expected_status):
            machines_status_ok = True
            lcmmachines_status_ok = True
            clusterrelease_status_ok = True

            previous_clusterrelease = None
            status = self.cluster.data['status']
            if status:
                previous_clusterrelease = status.get('providerStatus', {}).get(
                    'releaseRefs', {}).get('previous', {}).get('name', {})
            spec_release = \
                self.cluster.spec['providerSpec']['value']['release']
            LOG.info(f"Expected clusterrelease - {spec_release}, "
                     f"status.providerStatus.releaseRefs.previous.name - "
                     f"{previous_clusterrelease}")
            if spec_release != previous_clusterrelease:
                clusterrelease_status_ok = False

            if check_managed_clusters:
                # mgmt cluster has access to all machines,
                # so we check their status
                if self.cluster.is_management:
                    machines = \
                        self.k8sclient.kaas_machines.list_raw()
                    for machine in machines.to_dict()['items']:
                        m_name = machine['metadata']['name']
                        m_ns = machine['metadata']['namespace']
                        m_labels = machine['metadata'].get('labels', {})
                        cl_name = None if not m_labels else \
                            m_labels.get('cluster.sigs.k8s.io/cluster-name',
                                         None)
                        m_ps = machine['status'].get('providerStatus', {})
                        m_status = None if not m_ps else m_ps.get('status',
                                                                  None)
                        # we check that regional and mgmt machines are in
                        # expected_status otherwise we continue to wait
                        # CLUSTER_NAMESPACE is usually default and we expect
                        # regional cluster in this ns as well
                        if m_ns == settings.CLUSTER_NAMESPACE \
                                and m_status != expected_status:
                            LOG.info(f"Found {m_name} machine of {cl_name} "
                                     f"cluster in {m_status} status")
                            machines_status_ok = False
                        # we check that all machines, except mgmt
                        # and regional ones are in expected_status
                        if m_ns != settings.CLUSTER_NAMESPACE \
                                and m_status != expected_status:
                            LOG.error(f"Machine {m_name} of {cl_name} cluster "
                                      f"has status {m_status} while "
                                      f"Ready is expected")
                            raise Exception(f"Machine {m_name} has status "
                                            f"{m_status}")

                # mgmt & regional clusters have access to regional lcmmachines,
                # so we check their status here
                lcmmachines = \
                    self.cluster.get_all_lcmmachines(
                        states=["Upgrade", "Pending", "Deploy", "Prepare"]
                    )
                lcmm_not_ready = {}
                for lcmm in lcmmachines:
                    if lcmm.namespace != self.cluster.namespace:
                        lcmm_not_ready[lcmm.name] = {'ns': lcmm.namespace}
                    else:
                        LOG.info(f"{lcmm.name} lcmmachine of current "
                                 f"cluster is not Ready")
                        lcmmachines_status_ok = False
                if lcmm_not_ready:
                    LOG.error(f"These managed cluster lcmmachines "
                              f"are not ready: {lcmm_not_ready}")
                    raise Exception(f"These regional or child lcmachines "
                                    f"are not ready: {lcmm_not_ready}")
            else:
                # check lcmmachines that belong to current cluster
                my_lcmms = self.cluster.get_lcmmachines(
                    states=["Upgrade", "Pending", "Deploy", "Prepare"]
                )
                if my_lcmms:
                    lcmmachines_status_ok = False
                    LOG.info(f"Not Ready lcmmachines: {my_lcmms}")

                # check machines that belong to current cluster
                machines = self.cluster.get_machines()
                machines_in_status = [x for x in machines
                                      if x.machine_status == expected_status]
                LOG.debug(f"Machines in {expected_status} "
                          f"status: {machines_in_status}")
                if len(machines) != len(machines_in_status):
                    machines_status_ok = False
                for machine in [x for x in machines
                                if x not in machines_in_status]:
                    LOG.info(f"{machine.name} machine is "
                             f"in status {machine.machine_status}")

            if not machines_status_ok or not lcmmachines_status_ok \
                    or not clusterrelease_status_ok:
                LOG.info(f"===Not all machines or lcmmachines are in "
                         f"status {expected_status}===")
                raise RuntimeError(f"Not all machines or lcmmachines "
                                   f"are in status {expected_status}")
            return True

        LOG.info(f"Waiting {timeout} sec until cluster is fully updated")
        waiters.wait_pass(lambda: update_check(expected_status),
                          timeout=timeout, interval=interval,
                          expected=(RuntimeError, ApiException,
                                    KeyError, TypeError, MaxRetryError))
        # we need to refresh cached expected pods/objects
        # lists in cluster object
        self.cluster._refresh_expected_objects()
        LOG.info("Update finished")

    def check_persistent_volumes_mounts(self, timeout=900, interval=10):
        def check_volumes_mounts():
            LOG.info("Checking local-volume-provisioner mounts!")
            hb = self.cluster.get_helmbundle().data
            releases = hb['spec']['releases']
            chart_name = [
                x['name'] for x in releases
                if x.get('chart') == 'kaas-bm/local-volume-provisioner'
            ]
            mounts = [x['values']['classes'] for x in releases
                      if x.get('chart') == 'kaas-bm/local-volume-provisioner']
            # Check if the chart present (for kaas < 1.13.0(2.0.0))
            if not chart_name:
                LOG.warning("LVP chart is not used for current release!")
                return True
            else:
                assert mounts, "Missing mounts!"
                LOG.info("Expected mounts on the nodes:")
                mount_paths = []
                for mount in mounts[0]:
                    mount_paths.append(mount['hostDir'])
                    LOG.info(mount['hostDir'])
                if not mount_paths:
                    LOG.warning("Local volumes mounts are not described"
                                "(or not used) in hb! Skip checking")
                else:
                    lvp_pods_list = \
                        self.cluster.k8sclient.pods.list_starts_with(
                            pattern=chart_name[0])
                    missing_mounts = []
                    for pod in lvp_pods_list:
                        LOG.info(f"Check mounts from pod: {pod.name}")
                        for path in mount_paths:
                            LOG.info(f"Mount: '{path}'")
                            if not pod.check_folder_exists(dir_path=path):
                                missing_mounts.append(path)
                        if missing_mounts:
                            raise FileExistsError(
                                f"Mount {missing_mounts} not found "
                                f"on the host!")
                return True

        waiters.wait_pass(lambda: check_volumes_mounts(),
                          timeout=timeout, interval=interval,
                          expected=(MaxRetryError, ApiException))

    def wait_machine_status_by_name(self, machine_name, expected_status,
                                    timeout=1800, interval=30):
        def machine_status(name):
            status = self.cluster.get_machine(name).machine_status
            LOG.info(f"Machine {name} has <{status}> status")
            return status

        LOG.info(f"Wait {timeout} seconds until machine {machine_name} "
                 f"will be in the status <{expected_status}>")
        waiters.wait(lambda: machine_status(machine_name) == expected_status,
                     timeout=timeout, interval=interval)
        LOG.info(f"{machine_name} is in expected status {expected_status}")

    def check_cluster_release(self, expected_clusterrelease,
                              timeout=1000, interval=30):

        if expected_clusterrelease:
            LOG.info(f"Checking cluster clusterrelease version. "
                     f"Expected: {expected_clusterrelease}")
            try:
                ec = expected_clusterrelease
                waiters.wait(
                    lambda: self.cluster.clusterrelease_version == ec,
                    timeout=timeout,
                    interval=interval
                )
            except exceptions.TimeoutError:
                raise TimeoutError(
                    f"Clusterrelease version mismatch. "
                    f"Actual: {self.cluster.clusterrelease_version}, "
                    f"expected: {expected_clusterrelease}")
            LOG.info("Clusterrelease version is correct")

    def check_actual_expected_pods(self,
                                   expected_pods=None,
                                   timeout=300,
                                   interval=10,
                                   exclude_pods=None,
                                   exclude_jobs=True,
                                   check_all_nss=False):
        """Compare expected list of pods (which is fetched automatically,
           unless explicitly provided) with actual list of pods in this
           cluster. Comparison is conducted for all namespaces by default"""

        def compare(expected_pods_dict=None,
                    exclude_jobs=True):
            failed = {}
            namespaces = expected_pods_dict.keys()
            actual_list = [pod for pod in
                           self.k8sclient.pods.list_raw().to_dict()['items']]

            if exclude_jobs:
                before_filtering = \
                    set([x['metadata']['name'] for x in actual_list])
                actual_list = \
                    [x for x in actual_list if not (
                            x['metadata']['owner_references'] and
                            x['metadata'][
                                'owner_references'][0]['kind'] == 'Job')]
                after_filtering = \
                    set([x['metadata']['name'] for x in actual_list])
                LOG.debug(f"These pods are jobs and will be filtered out: "
                          f"{before_filtering - after_filtering}")
            # constant
            excluded_pods = ['elasticsearch-curator',
                             # SI pods
                             'ui-e2e-test', 'iam-bdd', 'conformance',
                             'test-stacklight', 'kaas-os-provider-e2e',
                             # to be checked
                             'patroni-lbauth',
                             'patroni-12-endpoint-cleanup',
                             'neutron-netns-cleanup-cron-default',
                             # temporarily exclude these MOS pods
                             'os-redis-operator-redisoperator',
                             'tungstenfabric-operator']
            excluded_pods += self.EXCLUDED_PODS
            # special exclusions
            for excl_pod in excluded_pods:
                actual_list = \
                    [x for x in actual_list
                     if not x['metadata']['name'].startswith(excl_pod)]

            other_ns_pods = [pod for pod in actual_list
                             if pod['metadata']['namespace'] not in namespaces]
            actual_list = [pod for pod in actual_list
                           if pod['metadata']['namespace'] in namespaces]

            not_checked = [pod['metadata']['name'] for pod in actual_list]

            # sort names of expected pods so that
            # long names come first. "/no_owner" tag is excluded
            for ns in namespaces:
                expected_pods_lst = sorted(
                    list(expected_pods_dict[ns].keys()),
                    key=lambda x: len(x.split("/")[0]), reverse=True
                )
                for pod in expected_pods_lst:
                    num = expected_pods_dict[ns][pod]
                    prefix = pod.split("/")[0]
                    compare_list = \
                        [pod for pod in actual_list
                         if pod['metadata']['name'].startswith(prefix) and
                         pod['metadata']['name'] in not_checked]

                    if not compare_list:
                        failed[prefix] = {"actual": 0,
                                          "desired/expected": num}
                        continue

                    if '/no_owner' in pod:
                        # for pods that are marked "no_owner" we do not fetch
                        # owner and use # of pods from the file
                        LOG.debug(f"Number of pods "
                                  f"for {prefix} group will be checked "
                                  f"according to expected pods list")
                        desired_num = num
                    else:
                        # get owner kind and name for the first pod in list
                        first_pod = compare_list[0]
                        owner_references = \
                            first_pod['metadata']['owner_references']
                        kind_name = f"{owner_references[0]['kind']}/" \
                                    f"{owner_references[0]['name']}"

                        if kind_name.split('/')[0] == "Node":
                            # Starting from 1.17 mirror pods have
                            # Node/node-name in owner_references
                            LOG.warning(f"No replica count info for pod "
                                        f"{first_pod['metadata']['name']}. "
                                        f"Using expected number ({num}) "
                                        f"from the list")
                            # Better to handle this type of pods as '/no_owner'
                            LOG.info("Please add '/no_owner' for this group of"
                                     " pods to expected pods list file")
                            desired_num = num
                        else:
                            try:
                                desired_num = self.__get_replica_number(
                                    kind_name,
                                    first_pod['metadata']['namespace']
                                )
                                LOG.debug(f"First pod is "
                                          f"{first_pod['metadata']['name']}, "
                                          f"owner: {kind_name}, "
                                          f"replica #: {desired_num}")
                                if int(desired_num) != int(num):
                                    LOG.warning(f"Desired num ({desired_num}) "
                                                f"is not equal to the "
                                                f"number ({num}) in expected "
                                                f"pod list. Pod: {prefix}")
                            except (Exception, ApiException) as e:
                                LOG.error(e)
                                LOG.error(f"Cannot process {prefix} "
                                          f"group of pods. Skipping")
                                failed[prefix] = {"actual": 0,
                                                  "desired/expected": num}
                                continue
                    if int(desired_num) != len(compare_list):
                        failed[prefix] = {"actual": len(compare_list),
                                          "desired/expected": desired_num}
                    not_checked = [x for x in not_checked
                                   if x not in [pod['metadata']['name']
                                                for pod in compare_list]]

                    actual_ns = set([pod['metadata']['namespace']
                                     for pod in compare_list])
                    if set([ns]) != actual_ns:
                        failed[prefix] = {"actual namespace": actual_ns,
                                          "desired/expected": ns}

            if other_ns_pods:
                for pod in other_ns_pods:
                    LOG.error(f"Extra pod {pod['metadata']['name']} "
                              f"found in {pod['metadata']['namespace']} "
                              f"namespace")
                    if check_all_nss:
                        not_checked.append(pod['metadata']['name'])

            if failed or not_checked:
                result = {"Pods mismatch": failed,
                          "Not checked pods": not_checked}
                LOG.warning(f"Compare pod check failed: {result}")
                return result

        if not expected_pods:
            if exclude_pods:
                expected_pods, do = self.cluster.get_expected_objects(
                    exclude_pods=exclude_pods)
            else:
                expected_pods = self.cluster.expected_pods

        try:
            waiters.wait(lambda: not compare(
                expected_pods_dict=expected_pods,
                exclude_jobs=exclude_jobs),
                         timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            result = compare(expected_pods_dict=expected_pods,
                             exclude_jobs=exclude_jobs)
            if result:
                err = f"Timeout waiting for pods. " \
                      f"After {timeout}s there are some fails: " \
                      f"{result}"
                raise TimeoutError(err)
        LOG.info("All pods and their replicas are found")

    @retry(AssertionError, delay=60, tries=10, logger=LOG)
    def check_k8s_resources_after_update(self, cluster_client,
                                         cluster_resources_before):
        LOG.info("Collect cluster fixed resources after update")
        cluster_resources_after = cluster_client.describe_fixed_resources()
        LOG.debug(f"Cluster resources after update:\n"
                  f"{yaml.dump(cluster_resources_after)}")

        before_keys = cluster_resources_before.keys()
        after_keys = cluster_resources_after.keys()
        common_keys = before_keys & after_keys

        new_objects = {key: cluster_resources_after[key]
                       for key in (after_keys - before_keys)}
        if new_objects:
            LOG.info(f"The following NEW objects appeared:\n"
                     f"{yaml.dump(new_objects)}")
        gone_objects = {k: cluster_resources_before[k]
                        for k in (before_keys - after_keys)}

        if gone_objects:
            LOG.error(f"The following OLD objects disappeared:\n"
                      f"{yaml.dump(gone_objects)}")

        LOG.info("Check that fixed resources weren't changed after update")
        before = yaml.dump(
            {k: cluster_resources_before[k] for k in common_keys
             if 'capacity/pods' not in k})
        # temporary add exclude for capacity/pods see PRODX-11503
        after = yaml.dump({k: cluster_resources_after[k] for k in common_keys
                           if 'capacity/pods' not in k})
        assert before == after, (f"Some cluster resources were "
                                 f"changed "
                                 f"before: \n {before} "
                                 f"after: \n {after}")
        assert not gone_objects, (f"The following OLD objects disappeared:\n"
                                  f"{yaml.dump(gone_objects)}")

    def get_ucp_worker_agent_name(self):
        """Find the service name for 'ucp-worker-agent-*' for the current MKE
        """
        def filter_ucp_worker_agent_service(service):
            service_name = service['Spec']['Name']
            return (
                service_name.startswith('ucp-worker-agent-') and
                'ucp-worker-agent-win' not in service_name)

        dashboardclient = self.cluster.mkedashboardclient
        docker_services = dashboardclient.get_swarm_services()

        filtered_services = list(filter(filter_ucp_worker_agent_service,
                                        docker_services))
        if len(filtered_services) < 1:
            raise Exception(f"Docker service 'ucp-worker-agent-*' not found "
                            f"on nodes of the cluster {self._cluster.name}")

        service_names = {
            s['Spec']['Labels']['com.docker.ucp.version']: s['Spec']['Name']
            for s in filtered_services
        }
        LOG.debug(f"Found the following worker agents:\n{service_names}")
        latest_ucp_version = max(service_names.keys())
        service_name = service_names[latest_ucp_version]
        LOG.info(f"Current name of ucp-worker-agent-* service for MKE version "
                 f"{latest_ucp_version} is: {service_name}")
        return service_name

    def check_actual_expected_docker_services(self,
                                              expected_services=None,
                                              excluded_services=None,
                                              changed_after_upd=None,
                                              timeout=900,
                                              interval=10):
        """Compare expected list of docker services (which are fetched
           automatically, unless explicitly provided) with actual
           list of docker services in this UCP cluster"""

        # storage for the status message from the latest check
        latest = {"status": ""}

        def compare(expected_services, excluded_services):
            """Get actual/expected docker service replicas and
               compare it to the expected_services from tests
               or from si_tests/templates/expected-pods/<cluserrelease>.yaml

               :rtype bool: False means that the errors were found
                            True means the check passed
            """
            failed = {}
            service_replicas = {}

            try:
                dashboardclient = self.cluster.mkedashboardclient
                docker_services = dashboardclient.get_swarm_service_replicas()
            except Exception as e:
                msg = f"Error reading docker swarm services:\n{e}"
                LOG.warning(msg)
                latest["status"] = msg
                self.cluster.docker_ps()
                return False

            for name, replicas in docker_services.items():
                LOG.debug(f"Docker service: {name}: {replicas}")
                r_actual = int(replicas["RunningTasks"])
                r_expected = int(replicas["DesiredTasks"])
                if r_expected == 0:
                    LOG.debug(f"Skip docker service '{name}' "
                              f"because in docker defined 0 replicas")
                    continue
                if name in excluded_services:
                    LOG.debug(f"Skip docker service '{name}' "
                              f"because it is in the skip list")
                    continue
                service_replicas[name] = {}
                service_replicas[name]["actual"] = r_actual
                service_replicas[name]["expected"] = r_expected

            not_checked = service_replicas.keys()

            for svc in expected_services.keys():
                if svc in excluded_services:
                    LOG.debug(f"Skip expected service '{svc}' "
                              f"because it is in the skip list")
                    continue

                num = expected_services[svc]

                if svc not in service_replicas.keys():
                    failed[svc] = {"actual": 0,
                                   "test expected": num}
                    LOG.debug(f"Docker service '{svc}' not found "
                              f"while it is expected with '{num}' replicas")
                    continue

                r_actual = int(service_replicas[svc]["actual"])
                r_expected = int(service_replicas[svc]["expected"])
                if int(num) != r_expected or r_expected != r_actual:
                    failed[svc] = {"actual": r_actual,
                                   "docker expected": r_expected,
                                   "test expected": num}
                    LOG.debug(f"Docker service '{svc}' has wrong number "
                              f"of replicas: {failed[svc]}")

                not_checked = [x for x in not_checked if x != svc]

            if failed or not_checked:
                result = {"Docker services mismatch": failed,
                          "Not checked docker services": not_checked}
                LOG.warning(f"Compare docker services check failed: {result}")
                latest["status"] = result
                return False
            else:
                latest["status"] = ""
                LOG.info("Compare docker services check passed "
                         "(should not fail for 3 minutes)")
                return True

        def compare_twice(expected_services, excluded_services):
            """Ensure service statuses several times during 3 minutes

            If compare() method is passed - then repeat it, until
            timeout reached.
            TimeoutError means that the compare() method was
            never failed during the specified timeout, so return True
            """
            try:
                waiters.wait(lambda: not compare(
                    expected_services, excluded_services),
                             timeout=180, interval=20)
            except exceptions.TimeoutError:
                return True
            return False

        if not expected_services:
            docker_objects = self.cluster.expected_docker_objects
            expected_services = docker_objects["service"]
            if changed_after_upd:
                # proceed with a static copy of keys, to not replace keys twice
                expected_services_keys = list(expected_services.keys())
                for key in changed_after_upd:
                    if key in expected_services_keys:
                        key_data = expected_services[key]
                        del expected_services[key]
                        expected_services[changed_after_upd[key]] = key_data

        # list of excluded docker services
        excluded_services = excluded_services or []
        LOG.info(f"Skip checks for the following docker services: "
                 f"{excluded_services}")

        try:
            waiters.wait(lambda: compare_twice(
                expected_services, excluded_services),
                         timeout=timeout, interval=interval)
        except exceptions.TimeoutError:
            if latest['status']:
                err = (f"Timeout waiting for docker services. "
                       f"After {timeout}s there are some fails: "
                       f"{latest['status']}")
            else:
                err = "Timeout waiting for docker services"
            raise TimeoutError(err)
        LOG.info("All docker services and their replicas were found")

    def __get_replica_number(self, obj, namespace):
        """Return desired number of replicas/scheduled pods according to
           the pod owner specification (e.g. statefulset).
        """
        kind, name = obj.split('/')
        desired_num = 0
        if kind == 'ReplicaSet':
            rs = self.k8sclient.replicasets.get(
                name=name, namespace=namespace).read().to_dict()
            desired_num = rs['metadata']['annotations'][
                'deployment.kubernetes.io/desired-replicas']
        elif kind == 'StatefulSet':
            ss = self.k8sclient.api_apps.read_namespaced_stateful_set(
                name=name, namespace=namespace).to_dict()
            desired_num = ss['spec']['replicas']
        elif kind == 'DaemonSet':
            ds = self.k8sclient.daemonsets.get(
                name=name, namespace=namespace).read().to_dict()
            desired_num = ds['status']['desired_number_scheduled']
        else:
            LOG.warning(f"Unknown kind {kind}")
        return desired_num

    def check_cluster_deleted(self, timeout=1800, interval=15):
        waiters.wait(lambda: not self.cluster.is_existed(),
                     timeout=timeout, interval=interval,
                     timeout_msg='Timeout waiting for cluster deletion')
        LOG.info("Cluster has been deleted")

    def _get_cluster_readiness(self,
                               exp_provider_status=True,
                               expected_fails=None,
                               expected_condition_fails=None):
        """Get conditions from Cluster and Machine objects

        :rtype bool: True only if all conditions are true, False in other case
        """
        machines_result = [m.are_conditions_ready(expected_condition_fails,
                                                  verbose=True)
                           for m in self.cluster.get_machines()]
        cluster_result = [
            self.cluster.are_conditions_ready(expected_condition_fails,
                                              verbose=True),
            self.cluster.is_ready(exp_provider_status, expected_fails)]

        return all(machines_result + cluster_result)

    def _get_cluster_status_msg(self):
        """To get latest status of the Cluster and Machine objects"""
        status = self.cluster.data['status']['providerStatus']
        status_keys = [
            'bastion',
            'conditions',
            'helm',
            'kind',
            'loadBalancerHost',
            'loadBalancerStatus',
            'nodes',
            'notReadyObjects',
            'ready',
        ]
        cluster_status = {k: v for k, v in status.items() if k in status_keys}
        machines = {m.name: m.data.get('status', {}).get('providerStatus', {})
                    for m in self.cluster.get_machines()}
        machines_msg = []
        for name, status in machines.items():
            prefix = f"Machine {self.cluster.namespace}/{name}"
            if 'conditions' not in status:
                msg = f"{prefix}: NO CONDITIONS IN PROVIDER STATUS"
                machines_msg.append(msg)
            else:
                errors = [f"{c['type']} condition: \"{c['message']}\""
                          for c in status['conditions'] if not c['ready']]
                if errors:
                    machines_msg.append(f"[ {prefix} ] " + "; ".join(errors))

        machines_str = "\n".join(machines_msg)
        return (f"Cluster data status:\n{yaml.dump(cluster_status)}\n"
                f"{machines_str}")

    def check_cluster_readiness(self,
                                timeout=1800,
                                interval=60,
                                exp_provider_status=True,
                                expected_fails=None,
                                expected_condition_fails=None):
        """
        Check that overall cluster is Ready
        and no unexpected deployments failed
        :param timeout: timeout for waiter
        :type timeout: int
        :param interval: interval to check status
        :type interval: int
        :param exp_provider_status: expected overall cluster status
        :type exp_provider_status: bool
        :param expected_fails: list of strings with
        expectedly failed deployment names
        :type expected_fails: List[str]
        :param expected_condition_fails: dict with conditions like
            { <condition type>: <part of condition message to match> , }
        :type expected_condition_fails: Dict[str]

        :rtype bool: bool
        """
        LOG.info("Checking cluster readiness")
        expected_fails = expected_fails or []
        cluster_release = self.cluster.clusterrelease_version
        LOG.info(f"Cluster release: {cluster_release}")
        can_check_mke = 'mke' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mke-5-13-0-3-3-6')
        can_check_mos = 'mos' in cluster_release and \
                        version.parse(cluster_release) \
                        > version.parse('mos-6-12-0-21-1')

        if can_check_mke or can_check_mos:

            timeout_msg = "Timeout waiting for cluster readiness"
            waiters.wait(lambda: self._get_cluster_readiness(
                             exp_provider_status,
                             expected_fails,
                             expected_condition_fails),
                         timeout=timeout,
                         interval=interval,
                         timeout_msg=timeout_msg,
                         status_msg_function=self._get_cluster_status_msg)
            LOG.info("Cluster checked")
        else:
            LOG.info(f"Skip cluster readiness check for cluster release "
                     f"{cluster_release} as feature is not supported yet")

    def check_offline_isolation(self, proxy_access_str):
        resources_to_check = [
            # Binaries
            "https://binary.mirantis.com",
            # Docker images
            "http://mirantis.azurecr.io",
            # APT mirror
            "https://mirror.mirantis.com",
            # Docker Engine mirror
            "https://repos.mirantis.com",
        ]

        machines = self.cluster.get_machines()
        machines_to_check = {machine.machine_type: machine
                             for machine in machines}
        for machine_type, machine in machines_to_check.items():
            LOG.info(f"Offline isolation check on machine [{machine_type}] "
                     f"{self._cluster.name}/{machine.name}")

            # Negative check: direct access from the offline node
            # to external resources should fail.
            # * "hostname -s" is used because equinix labs has wrong name
            #   order in /etc/hosts: "hostname -f" returns "localhost"
            for resource in resources_to_check:
                negative_cmd = (
                    f"set -x; hostname -s;"
                    f"unset HTTP_PROXY; unset HTTPS_PROXY;"
                    f"unset http_proxy; unset https_proxy;"
                    f"curl --connect-timeout 10 -sS {resource}"
                )
                result = machine.exec_pod_cmd(negative_cmd, verbose=False)
                assert result['exit_code'] != 0, (
                    f"No offline isolation on the machine "
                    f"{self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")
                assert (
                    "Connection timed out" in result['logs'] or
                    "Failed connect to" in result['logs'] or
                    "Failed to connect to" in result['logs'] or
                    "Could not resolve host" in result['logs']
                    ), (
                    f"Unexpected result while checking offline isolation "
                    f"on the machine {self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")

            if proxy_access_str:
                # Positive check: access from the offline node to
                # an external resource using the specified proxy should pass
                positive_cmd = (
                    f"set -x; hostname -s;"
                    f"curl --connect-timeout 10 -sS -x '{proxy_access_str}' "
                    f"{resources_to_check[0]}"
                )
                result = machine.exec_pod_cmd(positive_cmd, verbose=False)
                assert result['exit_code'] == 0, (
                    f"No proxy access from the machine "
                    f"{self._cluster.name}/{machine.name}:"
                    f"\n{result['logs']}")
            else:
                LOG.warning("Skip external access check through proxy "
                            "because proxy setting is empty")

            LOG.info(f"Offline isolation check on machine [{machine_type}] "
                     f"{self._cluster.name}/{machine.name}: PASSED")

    @staticmethod
    def generate_cert(remote, name, ip=None, dns_name=None):
        LOG.info("Generate certs")
        bootstrap_dir = settings.KAAS_BOOTSTRAP_TARGET_DIR

        remote.check_call(f"mkdir -p certs; "
                          f"./{bootstrap_dir}/kaas get ca "
                          f"--path certs/ "
                          f"--cn si-tests.test",
                          raise_on_err=True,
                          verbose=True)

        cmd = f"./{bootstrap_dir}/kaas get certificate " \
              f"--ca-path certs/ " \
              f"--dir {name} "

        if dns_name:
            add = f"--cn {dns_name} --dnsNames {dns_name}"
        elif ip:
            add = f"--cn {ip} --ips {ip}"
        else:
            raise ValueError("Neither hostname nor ip address "
                             "were provided to get/set certificate")
        LOG.info(f"Generate cert for {name} endpoint")
        remote.check_call(cmd + add, raise_on_err=True, verbose=True)

    @staticmethod
    def set_cert(remote, svc, name, ip=None,
                 dns_name=None, kubeconfig='kubeconfig'):
        LOG.info(f"Apply cert for {name} endpoint")
        bootstrap_dir = settings.KAAS_BOOTSTRAP_TARGET_DIR

        cmd = f"./{bootstrap_dir}/kaas set certificate " \
              f"--kubeconfig {kubeconfig} " \
              f"--cacert-file certs/ca.pem " \
              f"--cert-file certs/{name}/cert.pem " \
              f"--key-file certs/{name}/key.pem " \
              f"--for {svc} "

        if dns_name:
            add = f"--hostname {dns_name}"
        elif ip:
            add = f"--hostname {ip}"
        else:
            raise ValueError("Neither hostname nor ip address "
                             "were provided to get/set certificate")
        remote.check_call(cmd + add,
                          raise_on_err=True,
                          verbose=True,
                          timeout=1800)

    def check_applied_cert(self, remote):
        """
        Check that cluster deployed with OIDC disabled has custom certificates
        generated and applied for endpoints.
        Triggers automatically when settings.DISABLE_OIDC=True
        :rtype bool: bool
        """
        LOG.info("Apply and check certs")

        services = {
            'ui': {
                'namespace': 'kaas',
                'name': 'kaas-kaas-ui'
            },
            'keycloak': {
                'namespace': 'kaas',
                'name': 'iam-keycloak-http'
            }
        }

        for service in services:
            name = services[service]['name']
            namespace = services[service]['namespace']
            svc = self.cluster.k8sclient.services.get(
                name=name,
                namespace=namespace)
            ip = svc.get_external_addr()
            dns_name = svc.get_external_hostname()
            host = dns_name or ip

            self.generate_cert(remote, name, dns_name=dns_name, ip=ip)
            self.set_cert(remote, service, name, dns_name=dns_name, ip=ip)

            try:
                ssl_connection_setting = Context(SSLv3_METHOD)
            except ValueError:
                ssl_connection_setting = Context(TLSv1_2_METHOD)
            ssl_connection_setting.set_timeout(5)
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.connect((host, 443))
                c = Connection(ssl_connection_setting, s)
                c.set_tlsext_host_name(str.encode(host))
                c.set_connect_state()
                c.do_handshake()
                cert = c.get_peer_certificate()
                LOG.info(f'Certificate: {cert}')
                subject_list = cert.get_subject().get_components()
                LOG.info(f'Subject list: {subject_list}')
                cert_byte_arr_decoded = {}
                for item in subject_list:
                    cert_byte_arr_decoded.update(
                        {item[0].decode('utf-8'): item[1].decode('utf-8')})
                LOG.info(f"Decoded cert byte array: {cert_byte_arr_decoded}")

                assert len(cert_byte_arr_decoded) > 0
                assert cert_byte_arr_decoded["CN"] == host, \
                    'Cert common name is not equal to expected one'
