import base64
import json
import os
import string
from operator import itemgetter

import exec_helpers
import yaml
from kubernetes.client.rest import ApiException
from packaging import version
from retry import retry

import si_tests.utils.templates as template_utils
from si_tests import logger
from si_tests import settings
from si_tests.clients.k8s import K8sCluster
from si_tests.clients.dashboard import mcc_dashboard_client
from si_tests.clients.dashboard import mke_dashboard_client
from si_tests.clients.docker.docker_cli import DockerCliClient
from si_tests.deployments.utils.namespace import NAMESPACE
from si_tests.managers import clustercheck_manager
from si_tests.managers import provider_resource_manager as prm
from si_tests.utils import utils, waiters
from si_tests.clients.k8s.models import V1KaaSCluster

from typing import List

LOG = logger.logger

system_namespaces = ['default', 'kube-system', 'kube-public',
                     'openstack-provider-system', 'system',
                     'stacklight', 'ceph',
                     'kaas', 'lcm-system']
if settings.CLUSTER_NAMESPACE not in system_namespaces:
    system_namespaces.append(settings.CLUSTER_NAMESPACE)


class Manager(object):
    kubeconfig = None

    def __init__(self, kubeconfig=settings.KUBECONFIG_PATH):
        self._api = None
        self.kubeconfig = kubeconfig

    @property
    def api(self):
        """
            :rtype: cluster.K8sCluster
        """
        if self._api is None:
            if not os.path.isfile(self.kubeconfig):
                raise FileNotFoundError(
                    'kubeconfig file={} not found!'.format(self.kubeconfig))
            self._api = K8sCluster(kubeconfig=self.kubeconfig)
        return self._api

    def get_regional_clusters(self):
        return [x for x in self.get_clusters() if x.is_regional]

    def get_child_clusters(self):
        return [x for x in self.get_clusters() if x.is_child]

    def get_kaasrelease(self, name, namespace=settings.CLUSTER_NAMESPACE):
        """Get KaaSRelease CRD object like kaas-0-1-0-rc"""
        return self.api.kaas_kaasreleases.get(name=name,
                                              namespace=namespace)

    def get_kaasreleases(self):
        return self.api.kaas_kaasreleases.list_all()

    def get_kaasrelease_names(self):
        releases = self.get_kaasreleases()
        return [release.name for release in releases]

    def get_clusterrelease(self, name, namespace=settings.CLUSTER_NAMESPACE):
        """Get ClusterRelease CRD object like kubernetes-0-3-1-1-15"""
        return self.api.kaas_clusterreleases.get(name=name,
                                                 namespace=namespace)

    def get_proxies(self):
        return self.api.kaas_proxies.list_all()

    def get_proxy_by_name(self, name, ns=None):
        if ns:
            return [proxy for proxy in self.get_proxies()
                    if name in proxy.name and proxy.namespace == ns]
        else:
            return [proxy for proxy in self.get_proxies()
                    if name in proxy.name]

    def get_clusterreleases(self):
        return self.api.kaas_clusterreleases.list_all()

    def get_helmbundles(self):
        return self.api.kaas_helmbundles.list_all()

    def get_helmbundle(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_helmbundles.get(name=name,
                                             namespace=namespace)

    def get_clusters(self):
        clusters = self.api.kaas_clusters.list_all()
        return [Cluster(self, cluster) for cluster in clusters]

    def get_mgmt_cluster(self):
        """Get the KaaS management cluster

        Find the cluster where the following spec is set:
          spec:
            providerSpec:
              value:
                kaas:
                  management:
                    enabled: true  # or whatever
        """
        clusters = self.api.kaas_clusters.list_all()
        for cluster in clusters:
            c = Cluster(self, cluster)
            if c.is_management:
                return c
        raise Exception("Management cluster was not found among these clusters"
                        " {}".format(','.join([c.name for c in clusters])))

    def get_clusterrelease_names(self):
        releases = self.get_clusterreleases()
        return [release.name for release in releases]

    def get_supported_clusterreleases(self, k_release=None, provider=None):
        """Get all clusterreleases for the specified provider

        Return all clusterreleases if no provider specified

        :rtype dict: {<version>: <data>} ,
            where <version> is an internal release version like '7.0.0',
            and <data> is the original dict with data for this release from
            the 'spec.supportedClusterReleases' list
        """
        if not k_release:
            k_r = max(self.get_kaasreleases(),
                      key=lambda x: version.parse(x.name))
        else:
            k_r = self.get_kaasrelease(k_release)
        cl_releases_data = k_r.data['spec']['supportedClusterReleases']
        cl_releases = {
            x['version']: x
            for x in cl_releases_data
            if provider is None
            or not x.get('providers', {}).get('supported', [])
            or provider in x.get('providers', {}).get('supported', [])
        }
        assert (len(cl_releases.keys()) == len(cl_releases_data)), (
            f"Duplicate clusterrelease 'version' number found in "
            f"spec.supportedClusterReleases for kaasrelease "
            f"{k_r.data['spec']['version']}, please check:\n"
            f"{yaml.dump(cl_releases_data)}"
        )
        return cl_releases

    def get_supported_clusterrelease_names(
            self, k_release=None, provider=None):
        """Get the list of cluster release names"""
        supported_crs = self.get_supported_clusterreleases(
            k_release=k_release, provider=provider)
        return [data['name'] for ver, data in supported_crs.items()]

    def get_machines(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_machines.list(namespace=namespace)

    def get_machine(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_machines.get(name=name, namespace=namespace)

    def get_lcmmachines(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_lcmmachines.list(namespace=namespace)

    def get_lcmmachine(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_lcmmachines.get(name=name, namespace=namespace)

    def get_baremetalhosts(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_baremetalhosts.list(namespace=namespace)

    def get_baremetalhost(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.kaas_baremetalhosts.get(name=name, namespace=namespace)

    def get_ipam_hosts(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_hosts.list(namespace=namespace)

    def get_ipam_host(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_hosts.get(name=name, namespace=namespace)

    def get_ipam_ipaddrs(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_ipaddrs.list(namespace=namespace)

    def get_ipam_ipaddr(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_ipaddrs.get(name=name, namespace=namespace)

    def get_ipam_subnets(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_subnets.list(namespace=namespace)

    def get_ipam_subnet(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.ipam_subnets.get(name=name, namespace=namespace)

    def get_l2templates(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.l2templates.list(namespace=namespace)

    def get_l2template(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.l2templates.get(name=name, namespace=namespace)

    def get_baremetalhostprofiles(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.baremetalhostprofiles.list(namespace=namespace)

    def get_baremetalhostprofile(self, name,
                                 namespace=settings.CLUSTER_NAMESPACE):
        return self.api.baremetalhostprofiles.get(name=name,
                                                  namespace=namespace)

    def get_byocredentials(self):
        return self.api.kaas_byocredentials.list_all()

    def create_namespace(self, namespace):
        ns = self.api.namespaces.create(
            namespace=namespace,
            body={
                "apiVersion": "v1",
                "kind": "Namespace",
                "metadata": {
                    "name": namespace
                }
            }
        )
        return Namespace(self, ns)

    def get_namespace(self, namespace: str) -> "Namespace":
        ns = self.api.namespaces.get(name=namespace, namespace=namespace)
        return Namespace(self, ns)

    def get_namespaces(
            self,
            hide_system_namespaces: bool = True) -> List["Namespace"]:
        namespaces = self.api.namespaces.list_all()
        if hide_system_namespaces:
            namespaces = [
                ns for ns in namespaces
                if ns.name not in system_namespaces
            ]
        return [Namespace(self, ns) for ns in namespaces]

    def get_events(self, namespace, event_prefix=None, sort=True):
        """Return list of dicts with events data for the namespace"""
        result = []
        events = self.api.events.list(namespace=namespace)
        for event in events:
            if event_prefix and not event.name.startswith(event_prefix):
                continue
            try:
                data = event._read()
            except ApiException:
                continue
            if data.first_timestamp:
                data_first_date = str(data.first_timestamp.date())
                data_first_time = str(data.first_timestamp.time())
            else:
                data_first_date = ''
                data_first_time = ''
            if data.last_timestamp:
                data_last_time = str(data.last_timestamp.time())
            else:
                data_last_time = ''

            result.append({
                'data': data,
                'namespace': namespace,
                'event_date': data_first_date,
                'event_start': data_first_time,
                'event_end': data_last_time,
                'message': data.message,
                'reason': data.reason,
                'event_type': data.type,
                'component': data.source.component,
                'host': data.source.host or '-',
                'involved_object_uid': data.involved_object.uid,
            })
        if sort:
            return sorted(result, key=itemgetter('event_start'))
        else:
            return result

    def get_events_by_uid(self, namespace, uid, event_prefix=None, sort=True):
        """Return list of events for the specified object"""
        events = self.get_events(namespace, event_prefix, sort)
        filtered_events = [event for event in events
                           if event['involved_object_uid'] == uid]
        return filtered_events

    def get_events_by_uid_str(self, namespace, uid, event_prefix=None,
                              sort=True):
        events = self.get_events_by_uid(namespace, uid, event_prefix, sort)
        messages = []
        for event in events:
            data = event['data']
            messages.append(
                f"{data.last_timestamp or data.event_time}  "
                f"{data.type}  {data.reason}  {data.message}")
        return '\n'.join(messages)

    def get_system_events(self):
        result = []
        for ns in system_namespaces:
            result.extend(self.get_events(namespace=ns))
        return sorted(result, key=itemgetter('event_start'))

    def print_events(self, events):
        """Print events

        :param events: list of dicts, prepared in self.get_events()
        """
        print("  Events:\n")
        for event in events:
            print("  >> {event_date} {event_start}-{event_end} "
                  "{event_type} {reason} {ns}/{component}/{host} {message}"
                  .format(event_date=event['event_date'],
                          event_start=event['event_start'],
                          event_end=event['event_end'],
                          ns=event['namespace'],
                          component=event['component'],
                          host=event['host'],
                          event_type=event['event_type'],
                          reason=event['reason'],
                          message=event['message']))

    def wait_kaasrelease(self, releases_count, timeout=30 * 60,
                         interval=30):

        def check_available_kaas_releases():
            releases = self.get_kaasrelease_names()
            LOG.info('Found kaasreleases: {}'.format(releases))
            if len(releases) != releases_count:
                return False
            return True

        waiters.wait(check_available_kaas_releases,
                     timeout=timeout,
                     interval=interval,
                     timeout_msg='Expected count: {0}, got {1}'.format(
                         releases_count, len(self.get_kaasrelease_names())))

    def wait_kaasrelease_label(self, release, timeout=3400, interval=30):
        """Wait until the specified kaasrelease has the 'active' label

        During waiting, show the current and next clusterrelease versions
        for all the management and region clusters, and LCMMachine status
        to track the KaaS update process.
        """
        clusters = [self.get_mgmt_cluster()] + self.get_regional_clusters()
        lcmmachines = {}
        for cluster in clusters:
            lcmmachines[cluster.name] = cluster.get_lcmmachines()

        def check_label():
            # Get releases from mgmt and region clusters with lcmmachine status
            try:
                status_msg = ""
                for cluster in clusters:
                    current = cluster.clusterrelease_version
                    available = cluster.available_clusterrelease_version
                    lcmstates = [
                        lcmmachine.data.get('status', {}).get('state', '-')
                        for lcmmachine in lcmmachines[cluster.name]
                    ]
                    status_msg += (f"[{cluster.namespace}/{cluster.name}] "
                                   f"Processing: {current}  Next: {available}"
                                   f"  {lcmstates}\n")
                # Get kaasrelease and check if it is active
                kr_data = self.get_kaasrelease(release).data
                LOG.debug("Release_details {}".format(kr_data))
                kr_meta = kr_data['metadata']['labels']
                LOG.info(
                    f"Release: {release}. Waiting for the ACTIVE label. "
                    f"Current {kr_meta}\n{status_msg}"
                )
                if kr_meta:
                    if 'true' in kr_meta.get('kaas.mirantis.com/active', ''):
                        return True
                return False
            except ApiException:
                return False

        waiters.wait(check_label,
                     timeout=timeout,
                     interval=interval,
                     timeout_msg="Missed active label for release {}".format(
                         release))

    @staticmethod
    def get_bmApiversion(release_name):
        """
        For all kaas-releases including 'kaas-1-5-0-rc' we use new format
        :param release_name:
        :return:
        """
        old = 'baremetal.cluster.k8s.io/v1alpha1'
        new = 'baremetal.k8s.io/v1alpha1'
        if version.parse(release_name) <= version.parse("kaas-1-5-0"):
            return old
        return new

    def get_aws_credential(self, name_prefix, namespace, region=None):
        """Find AWS credentials for the specified cluster in the namespace

        :return: tuple (aws_key_id, aws_secret_access_key)
        """
        creds_list = self.api.kaas_awscredentials.list(
            namespace=namespace)

        cred_names = [cred.name for cred in creds_list]
        # Filter creds that match the specified cluster_name
        creds = [cred.data for cred in creds_list
                 if cred.name.startswith(name_prefix)]
        if region is not None:
            # Filter creds that match the specified region
            creds = [cred for cred in creds
                     if cred['metadata']['labels'][
                         'kaas.mirantis.com/region'] == region]

        assert creds, (
            f"AWSCredential object not found for the cluster "
            f"{namespace}/{name_prefix} : {cred_names}")

        cred = creds[0]
        aws_key_id = cred["spec"]["accessKeyID"]
        secret_ak = cred["spec"]["secretAccessKey"]
        if "value" in secret_ak:
            aws_secret_access_key = secret_ak["value"]
        else:
            secret_name = secret_ak["secret"]["name"]
            secret = self.api.secrets.get(name=secret_name,
                                          namespace=namespace)
            value = secret.read().data["value"]
            aws_secret_access_key = base64.b64decode(value).decode('utf-8')
        return aws_key_id, aws_secret_access_key

    def get_secret_data(self, secret_name, namespace, data_key):
        secret = self.api.secrets.get(
            secret_name, namespace=namespace)
        secret_data = secret.read().to_dict()['data']
        data = base64.b64decode(secret_data[data_key]).decode("utf-8")
        LOG.debug(f"Secret {namespace}/{secret_name} data from {data_key}:"
                  f" {data}")
        return data

    def get_keycloak_ip(self):
        keycloak_svc_status = self.api.api_core.read_namespaced_service(
            name='iam-keycloak-http',
            namespace='kaas').to_dict()['status']
        keycloak_svc = keycloak_svc_status['load_balancer']['ingress'][0]
        keycloak_ip = \
            keycloak_svc['ip'] or keycloak_svc['hostname']
        LOG.debug(f"Keycloak IP: {keycloak_ip}")
        return keycloak_ip

    def get_iam_ip_port(self):
        iam_svc = self.api.api_core.read_namespaced_service(
            name='iam-api-http',
            namespace='kaas').to_dict()
        iam_ip = \
            iam_svc['status']['load_balancer']['ingress'][0]['ip'] or \
            iam_svc['status']['load_balancer']['ingress'][0]['hostname']
        iam_port = iam_svc['spec']['ports'][0]['port']
        iam_ip_port = iam_ip + ":" + str(iam_port)
        LOG.debug(f"IAM URL: {iam_ip_port}")
        return iam_ip_port

    def get_iamrolebindings(self, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.iam_rolebindings.list(namespace=namespace)

    def get_iamrolebinding(self, name, namespace=settings.CLUSTER_NAMESPACE):
        return self.api.iam_rolebindings.get(name=name, namespace=namespace)

    def get_vsphere_credential(self, name_prefix, namespace, region=None):
        """Find VSpere credentials for the specified cluster in the namespace

        """

        creds_list = self.api.kaas_vspherecredentials.list(
            namespace=namespace)
        cred_names = [cred.name for cred in creds_list]
        creds = [cred.data for cred in creds_list
                 if cred.name.startswith(name_prefix)]

        if region is not None:
            # Filter creds that match the specified region
            creds = [cred for cred in creds
                     if cred['metadata']['labels'][
                         'kaas.mirantis.com/region'] == region]
        assert creds, (
            f"VsphereCredential object not found for the cluster "
            f"{namespace}/{name_prefix} : {cred_names}")

        one_cred = creds[0]
        cluster_api_creds = one_cred['spec']['clusterApi']
        cloud_provider_creds = one_cred['spec']['cloudProvider']
        vsphere = one_cred['spec']['vsphere']
        vsphere_creds = {
            "cluster_api_user": cluster_api_creds['username'],
            "cluster_api_pwd": settings.KAAS_VSPHERE_CAPI_PROVIDER_PASSWORD,
            "cloud_provider_user": cloud_provider_creds['username'],
            "cloud_provider_pwd":
                settings.KAAS_VSPHERE_CLOUD_PROVIDER_PASSWORD,
            "datacenter_name": vsphere['datacenter'],
            "server_insecure": vsphere['insecure'],
            "vsphere_server_ip": vsphere['server'],
            "vsphere_server_port": vsphere.get('port', None)
        }

        return vsphere_creds

    def get_rhel_license(self, name_prefix, namespace, region=None):
        licenses_list = self.api.kaas_rhellicenses.list(
            namespace=namespace)
        licenses_names = [lic.name for lic in licenses_list]
        licenses = [lic.data for lic in licenses_list
                    if lic.name.startswith(name_prefix)]

        if region is not None:
            # Filter licenses that match the specified region
            licenses = [lic for lic in licenses
                        if lic['metadata']['labels'][
                            'kaas.mirantis.com/region'] == region]
        assert licenses, (
            f"RHELLicense object not found for the cluster "
            f"{namespace}/{name_prefix} : {licenses_names}")

        one_license = licenses[0]

        return one_license

    def find_rhel_license(self, name_prefix, namespace, region=None):
        licenses_list = self.api.kaas_rhellicenses.list(
            namespace=namespace)
        licenses_names = [lic.name for lic in licenses_list]
        licenses = [lic.data for lic in licenses_list
                    if lic.name.startswith(name_prefix)]

        if region is not None:
            # Filter licenses that match the specified region
            licenses = [lic for lic in licenses
                        if lic['metadata']['labels'][
                            'kaas.mirantis.com/region'] == region]
        assert licenses, (
            f"RHELLicense object not found for the cluster "
            f"{namespace}/{name_prefix} : {licenses_names}")

        one_license = licenses[0]

        return one_license


class Namespace(object):
    __manager = None
    __namespace = None

    def __init__(self, manager, k8s_namespace):
        """
        manager: <Manager> instance
        k8s_namespace: <K8sNamespace> instance
        """
        self.__manager = manager
        self.__namespace = k8s_namespace

    @property
    def name(self):
        """Namespace name"""
        return self.__namespace.name

    @property
    def uid(self):
        """Namespace uid"""
        return self.data['metadata']['uid']

    @property
    def data(self):
        """Returns dict of k8s object

        Data contains keys like api_version, kind,

        metadata, spec, status or items
        """
        return self.__namespace.read().to_dict()

    def delete(self):
        """Deletes the current namespace"""
        self.__namespace.delete()

    def is_existed(self):
        """Verifies the current namespace deletion"""
        all_ns = self.__manager.get_namespaces()
        for n in all_ns:
            if n.name == self.name:
                LOG.info("Cluster namespace found")
                return True
        return False

    def wait_for_deletion(self, timeout=1800, interval=15):
        """Wait for the current namespace deletion"""
        waiters.wait(lambda: not self.is_existed(),
                     timeout=timeout, interval=interval,
                     timeout_msg='Timeout waiting for namespace deletion')
        LOG.info("Namespace has been deleted.")

    def get_publickeys(self):
        return self.__manager.api.kaas_publickeys.list(namespace=self.name)

    def get_publickey(self, name):
        return self.__manager.api.kaas_publickeys.get(name=name,
                                                      namespace=self.name)

    def create_publickey(self, name, key):
        """Create public key in the current namespace

        :param name: str, public key name
        :param key: str, public key content
        """
        return self.__manager.api.kaas_publickeys.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "PublicKey",
                "metadata": {
                    "name": name,
                },
                "spec": {
                    "publicKey": key
                }
            }
        )

    def get_secrets(self):
        return self.__manager.api.secrets.list(namespace=self.name)

    def get_secret(self, name):
        return self.__manager.api.secrets.get(name=name, namespace=self.name)

    def create_openstack_secret(self, name, clouds_yaml_str):
        encoded_clouds_yaml_str = base64.b64encode(
            clouds_yaml_str.encode('utf-8')).decode()
        data = {
            'clouds.yaml': encoded_clouds_yaml_str
        }
        return self.create_secret(name, data)

    def create_secret(self, name, data, region="region-one",
                      provider="openstack"):
        return self.__manager.api.secrets.create(
            namespace=self.name,
            body={
                "apiVersion": "v1",
                "kind": "Secret",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        "kaas.mirantis.com/credentials": "true",
                        "kaas.mirantis.com/region": region,
                        "kaas.mirantis.com/provider": provider
                    }
                },
                "data": data
            }
        )

    def create_openstack_credential(self, name, clouds_yaml_str,
                                    region="region-one", validate=True):
        clouds = yaml.load(clouds_yaml_str,
                           Loader=yaml.SafeLoader).get("clouds", {})
        if len(clouds) > 1:
            raise Exception("Multiple clouds are present in OpenStack config")
        for cloud in clouds.values():
            auth = cloud.get("auth", {})
            res = self.__manager.api.kaas_openstackcredentials.create(
                namespace=self.name,
                body={
                    "apiVersion": "kaas.mirantis.com/v1alpha1",
                    "kind": "OpenStackCredential",
                    "metadata": {
                        "name": name,
                        "namespace": self.name,
                        "labels": {
                            "kaas.mirantis.com/region": region
                        }
                    },
                    "spec": {
                        "auth": {
                            "authURL": auth.get("auth_url", ""),
                            "userName": auth.get("username", ""),
                            "userID": auth.get("user_id", ""),
                            "password": {
                                "value": auth.get("password", "")
                            },
                            "projectDomainName": auth.get(
                                "project_domain_name", ""),
                            "projectDomainID": auth.get(
                                "project_domain_id", ""),
                            "projectName": auth.get("project_name", ""),
                            "projectID": auth.get("project_id", ""),
                            "userDomainName": auth.get(
                                "user_domain_name", ""),
                            "userDomainID": auth.get("user_domain_id", ""),
                            "domainName": auth.get("domain_name", ""),
                            "domainID": auth.get("domain_id", ""),
                            "defaultDomain": auth.get("default_domain", "")

                        },
                        "regionName": cloud.get("region_name", ""),
                    }
                }
            )
            if validate:
                # keep in mind that object may be differ to 'res'
                # after validation
                self.wait_provider_cred_is_valid(
                        credential_name=name,
                        provider=settings.OPENSTACK_PROVIDER_NAME)
            return res
        raise Exception("There are no clouds provided in OpenStack config")

    def wait_provider_cred_is_valid(self,
                                    credential_name,
                                    provider,
                                    retries=60,
                                    interval=60):
        def wait_cred_status():
            try:
                LOG.info("Wait until credential {} become valid".format(
                    credential_name))
                if provider == settings.OPENSTACK_PROVIDER_NAME:
                    cred = self.get_openstackcredential(
                            name=credential_name).data
                elif provider == settings.VSPHERE_PROVIDER_NAME:
                    cred = self.__manager.api.kaas_vspherecredentials.get(
                            name=credential_name,
                            namespace=self.name).data
                elif provider == settings.AWS_PROVIDER_NAME:
                    cred = self.__manager.api.kaas_awscredentials.get(
                            name=credential_name,
                            namespace=self.name).data
                elif provider == settings.EQUINIX_PROVIDER_NAME:
                    cred = self.__manager.api.kaas_equinixmetalcredentials.get(
                            name=credential_name,
                            namespace=self.name).data
                elif provider == settings.AZURE_PROVIDER_NAME:
                    cred = self.__manager.api.kaas_azurecredentials.get(
                            name=credential_name,
                            namespace=self.name).data
                else:
                    raise Exception("Unsupported provider for cred validation")

                LOG.info('cred data {}'.format(cred))
                if cred['status']:
                    if cred['status']['valid']:
                        return True
            except ApiException as e:
                LOG.debug(
                    'Failed to wait credential '
                    'become valid {}'.format(e))
            return False

        waiters.wait(lambda: wait_cred_status(),
                     timeout=retries * interval,
                     interval=interval,
                     timeout_msg="Credential {0} isn't valid "
                                 "after {1} retries".format(
                         credential_name, retries
                     ))

    def create_rhel_license(self,
                            name,
                            rhel_creds,
                            region="region-one",
                            is_license_key=False):

        if is_license_key:
            (key, org) = rhel_creds
            spec = {
                "activationKey": {
                    "value": key
                },
                "orgID": org
            }
        else:
            (username, password) = rhel_creds
            spec = {
                "username": username,
                "password": {
                    "value": password
                }
            }

        return self.__manager.api.kaas_rhellicenses.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "RHELLicense",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        "kaas.mirantis.com/region": region
                    }
                },
                "spec": spec
            }
        )

    def create_vsphere_credential(self, name,
                                  vsphere_creds,
                                  region="region-one",
                                  validate=True):

        body = {
            "apiVersion": "kaas.mirantis.com/v1alpha1",
            "kind": "VsphereCredential",
            "metadata": {
                "name": name,
                "namespace": self.name,
                "labels": {
                    "kaas.mirantis.com/region": region
                }
            },
            "spec": {
                "clusterApi": {
                    "username": vsphere_creds['cluster_api_user'],
                    "password": {
                        "value": vsphere_creds['cluster_api_pwd']
                    }
                },
                "cloudProvider": {
                    "username": vsphere_creds['cloud_provider_user'],
                    "password": {
                        "value": vsphere_creds['cloud_provider_pwd']
                    },
                },
                "vsphere": {
                    "server": vsphere_creds['vsphere_server_ip'],
                    "insecure": vsphere_creds['server_insecure'],
                    "datacenter": vsphere_creds['datacenter_name']
                }
            }
        }

        if vsphere_creds.get('vsphere_server_port'):
            body["spec"]["vsphere"]["port"] = vsphere_creds[
                'vsphere_server_port']

        res = self.__manager.api.kaas_vspherecredentials.create(
            namespace=self.name,
            body=body)

        if validate:
            # keep in mind that object may be differ to 'res' after validation
            self.wait_provider_cred_is_valid(
                    credential_name=name,
                    provider=settings.VSPHERE_PROVIDER_NAME)
        return res

    def create_aws_credential(self, name, aws_key_id, aws_secret_access_key,
                              region="region-one", validate=True):
        res = self.__manager.api.kaas_awscredentials.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "AWSCredential",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        "kaas.mirantis.com/region": region
                    }
                },
                "spec": {
                    "accessKeyID": aws_key_id,
                    "secretAccessKey": {
                        "value": aws_secret_access_key
                    }
                }
            }
        )
        if validate:
            # keep in mind that object may be differ to 'res' after validation
            self.wait_provider_cred_is_valid(
                    credential_name=name,
                    provider=settings.AWS_PROVIDER_NAME)
        return res

    def create_equinix_credential(self, name, api_token, project_id,
                                  region="region-one", validate=True):
        res = self.__manager.api.kaas_equinixmetalcredentials.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "EquinixMetalCredential",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        "kaas.mirantis.com/region": region
                    }
                },
                "spec": {
                    "apiToken": {
                        "value": api_token
                    },
                    "projectID": project_id,
                }
            }
        )
        if validate:
            # keep in mind that object may be differ to 'res' after validation
            self.wait_provider_cred_is_valid(
                    credential_name=name,
                    provider=settings.EQUINIX_PROVIDER_NAME)
        return res

    def create_azure_credential(self,
                                name,
                                subscription_id,
                                tenant_id,
                                client_id,
                                client_secret,
                                region="region-one",
                                validate=True):
        res = self.__manager.api.kaas_azurecredentials.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "AzureCredential",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        "kaas.mirantis.com/region": region
                    }
                },
                "spec": {
                    "subscriptionID": subscription_id,
                    "tenantID": tenant_id,
                    "clientID": client_id,
                    "clientSecret": {
                        "value": client_secret
                    },
                }
            }
        )
        if validate:
            # keep in mind that object may be differ to 'res' after validation
            self.wait_provider_cred_is_valid(
                    credential_name=name,
                    provider=settings.AZURE_PROVIDER_NAME)
        return res

    def get_events(self):
        return self.__manager.get_events(namespace=self.name)

    def print_events(self):
        self.__manager.print_events(self.get_events())

    def get_openstackresource(self, name):
        return self.__manager.api.kaas_openstackresources.get(
            name=name, namespace=self.name)

    def get_openstackcredential(self, name):
        return self.__manager.api.kaas_openstackcredentials.get(
            name=name, namespace=self.name)

    def get_openstackresources(self):
        return self.__manager.api.kaas_openstackresources.list(
            namespace=self.name)

    def get_clusters(self):
        clusters = self.__manager.api.kaas_clusters.list(namespace=self.name)
        return [Cluster(self.__manager, cluster) for cluster in clusters]

    def get_cluster(self, name: str) -> "Cluster":
        cluster = self.__manager.api.kaas_clusters.get(name=name,
                                                       namespace=self.name)
        return Cluster(self.__manager, cluster)

    def get_provider_credentials(self, provider='', region=''):
        if provider == 'AWSClusterProviderSpec':
            creds = self.__manager.api.kaas_awscredentials.list_all()
        elif provider == 'OpenstackClusterProviderSpec':
            creds = self.__manager.api.kaas_openstackcredentials.list_all()
        elif provider == 'VsphereClusterProviderSpec':
            creds = self.__manager.api.kaas_vspherecredentials.list_all()
        elif provider == 'EquinixMetalClusterProviderSpec':
            creds = self.__manager.api.kaas_equinixmetalcredentials.list_all()
        else:
            creds = None
        if region and creds is not None:
            creds = [cred for cred in creds if
                     cred.data['metadata']['labels']
                     ['kaas.mirantis.com/region'] == region]
        return creds

    def create_byocredential(self, name, ucp_host, cacert,
                             clientcert, clientkey, ucp_kubeconfig,
                             region='region-one'):
        bc = self.__manager.api.kaas_byocredentials.create(
            namespace=self.name,
            body={
                "apiVersion": "kaas.mirantis.com/v1alpha1",
                "kind": "BYOCredential",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    'labels': {
                        'kaas.mirantis.com/region': region,
                        'kaas.mirantis.com/provider': 'byo'
                    }
                },
                "spec": {
                    "docker": {
                        "host": ucp_host,
                        "caCert": cacert,
                        "clientCert": clientcert,
                        "clientKey": {"value": clientkey}
                    },
                    "kubeConfig": {'value': ucp_kubeconfig}
                }
            }
        )
        return bc

    def get_byocredential(self, name):
        return self.__manager.api.kaas_byocredentials.get(name=name,
                                                          namespace=self.name)

    def get_byocredentials(self):
        return self.__manager.api.kaas_byocredentials.list(namespace=self.name)

    def get_cephclusters(self):
        return self.__manager.api.kaas_cephclusters.list(
            namespace=self.name)

    def get_cephcluster(self, name):
        return self.__manager.api.kaas_cephclusters.get(
            name=name, namespace=self.name)

    def get_helmbundle(self, name):
        return self.__manager.api.kaas_helmbundles.get(name=name,
                                                       namespace=self.name)

    def get_helmbundles(self):
        return self.__manager.api.kaas_helmbundles.list(namespace=self.name)

    def attach_ucp_cluster(self, name, creds_name, region='region-one',
                           lma_enabled=settings.KAAS_CHILD_CLUSTER_DEPLOY_LMA):
        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Cluster",
            "metadata": {
                "name": name,
                "namespace": self.name,
                'labels': {
                    'kaas.mirantis.com/region': region,
                    'kaas.mirantis.com/provider': 'byo'
                },
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "byo.kaas.mirantis.com/v1alpha1",
                        "kind": "BYOClusterProviderSpec",
                        "credentials": creds_name,

                    }
                }
            }
        }
        if not lma_enabled:
            # Explicitly disable LMA for BYO cluster
            body["spec"]["providerSpec"]["value"]["helmReleases"] = [
                {
                    "name": "stacklight",
                    "enabled": False
                }
            ]
        return self.__manager.api.kaas_clusters.create(namespace=self.name,
                                                       body=body)

    def create_cluster(self,
                       name,
                       release_name,
                       credentials_name="",
                       external_network_id="",
                       region="region-one",
                       provider="openstack",
                       aws_region="us-east-2",
                       aws_az="us-east-2a",
                       aws_priv_subnet_cidr="10.0.0.0/24",
                       aws_pub_subnet_cidr="10.0.1.0/24",
                       loadbalancer_host="172.16.48.223",
                       metallb_ip_range=None,
                       services_cidr="10.96.0.0/12",
                       pods_cidr="192.168.0.0/16",
                       nodes_cidr="10.10.10.0/24",
                       dns=settings.KAAS_CHILD_CLUSTER_DNS,
                       lma_enabled=True,
                       lma_extra_options=None,
                       extra_helm_releases=None,
                       dedicated_controlplane=True,
                       public_key_name=None,
                       proxy_name=None,
                       ipam_enabled=False,
                       ipam_cidr=None,
                       ipam_gw=None,
                       ipam_ns=None,
                       ipam_incl_range=None):

        if extra_helm_releases is None:
            extra_helm_releases = []

        if public_key_name:
            public_key = {"name": public_key_name}
        else:
            public_key = {}
            LOG.warning("Cluster will be created without public key")

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Cluster",
            "metadata": {
                "name": name,
                "namespace": self.name,
                "annotations": {
                    "kaas.mirantis.com/lcm": "true"
                },
                "labels": {
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "clusterNetwork": {
                    "services": {
                        "cidrBlocks": [
                            services_cidr
                        ]
                    },
                    "pods": {
                        "cidrBlocks": [
                            pods_cidr
                        ]
                    },
                    "serviceDomain": ""
                },
                "providerSpec": {
                    "value": {
                        "apiVersion": "openstackproviderconfig.k8s.io/v1alpha1",  # noqa
                        "kind": "OpenstackClusterProviderSpec",
                        "dnsNameservers": dns.split(','),
                        "externalNetworkId": external_network_id,
                        "nodeCidr": nodes_cidr,
                        "release": release_name,
                        "publicKeys": [
                            public_key
                        ],
                        "credentials": credentials_name,
                        "dedicatedControlPlane": dedicated_controlplane,
                        "helmReleases": extra_helm_releases
                    }
                }
            }
        }

        lma_defaults = {
            "name": "stacklight",
            "enabled": True,
            "values": {
                "alertmanagerSimpleConfig": {
                    "email": {
                        "enabled": False
                    },
                    "slack": {
                        "enabled": False
                    }
                },
                "highAvailabilityEnabled": settings.STACKLIGHT_ENABLE_HA,
                "prometheusServer": {
                    "customAlerts": [],
                    "persistentVolumeClaimSize": "16Gi",
                    "retentionSize": "15GB",
                    "retentionTime": "15d",
                    "watchDogAlertEnabled": False
                }
            }
        }
        lma_logging = {
            "elasticsearch": {
                "logstashRetentionTime": "30",
                "persistentVolumeClaimSize": "30Gi"
            },
            "logging": {
                "enabled": settings.STACKLIGHT_ENABLE_LOGGING
            },
        }
        if settings.STACKLIGHT_ENABLE_LOGGING:
            utils.merge_dicts(lma_defaults['values'], lma_logging)
        metallb_helm_release = {
            "name": "metallb",
            "values": {"configInline": {}}
        }
        if metallb_ip_range:
            metallb_helm_release['values']['configInline'] = {
                "address-pools": [
                    {
                        "addresses": [
                            metallb_ip_range
                        ],
                        "name": "default",
                        "protocol": "layer2"
                    }
                ]
            }

        aws_subnet_defaults = {
            "subnets": [
                {
                    "cidrBlock": aws_priv_subnet_cidr,
                    "isPublic": False,
                    "availabilityZone": aws_az
                },
                {
                    "cidrBlock": aws_pub_subnet_cidr,
                    "isPublic": True,
                    "availabilityZone": aws_az
                }
            ]
        }

        if proxy_name:
            body['spec']['providerSpec']['value']['proxy'] = proxy_name

        if lma_enabled:
            if lma_extra_options:
                assert type(lma_extra_options) == dict, (
                    "Please, use correct format for lma_extra_options. "
                    "Use next dict as a reference: \n{}").format(
                    yaml.dump(lma_defaults, default_flow_style=False))
                lma_options = utils.merge_dicts(lma_defaults,
                                                lma_extra_options)
                body.get("spec").get("providerSpec").get(
                    "value").get("helmReleases").append(lma_options)
            else:
                body.get("spec").get("providerSpec").get(
                    "value").get("helmReleases").append(lma_defaults)
        else:
            body.get("spec").get("providerSpec").get(
                "value").get("helmReleases").append({"name": "stacklight",
                                                     "enabled": False})

        if provider == settings.BAREMETAL_PROVIDER_NAME:
            body['spec']['providerSpec']['value']['helmReleases'].append(
                metallb_helm_release)
            body['spec']['providerSpec']['value'][
                'apiVersion'] = self.__manager.get_bmApiversion(release_name)
            body['spec']['providerSpec']['value'][
                'kind'] = "BaremetalClusterProviderSpec"
            body['spec']['providerSpec']['value'][
                'loadBalancerHost'] = loadbalancer_host
            body['spec']['providerSpec']['value'][
                'region'] = region

        elif provider == settings.AWS_PROVIDER_NAME:
            body['spec']['providerSpec']['value'][
                'apiVersion'] = "aws.kaas.mirantis.com/v1alpha1"
            body['spec']['providerSpec']['value'][
                'kind'] = "AWSClusterProviderSpec"
            body['spec']['providerSpec']['value'][
                'region'] = aws_region
            body['spec']['providerSpec']['value'][
                'networkSpec'] = aws_subnet_defaults

        elif provider == settings.VSPHERE_PROVIDER_NAME:
            body['spec']['providerSpec']['value'].update(
                {
                    "apiVersion": "vsphere.cluster.k8s.io/v1alpha1",
                    "kind": "VsphereClusterProviderSpec",
                    "loadBalancerHost": loadbalancer_host,
                    "vsphere": {
                        "cloudProviderDatastore":
                            settings.KAAS_VSPHERE_DATASTORE_NAME,
                        "clusterApiDatastore":
                            settings.KAAS_VSPHERE_DATASTORE_NAME,
                        "clusterApiDatastoreFolder":
                            settings.KAAS_VSPHERE_DATASTORECLUSTER_PATH,
                        "machineFolderPath":
                            settings.KAAS_VSPHERE_MACHINES_FOLDER,
                        "networkPath": settings.KAAS_VSPHERE_NETWORK_PATH,
                        "resourcePoolPath":
                            settings.KAAS_VSPHERE_RESOURCE_POOL_PATH,
                        "scsiControllerType": "pvscsi"
                    }
                }
            )
            body.get("spec").get("providerSpec").get(
                "value").get("helmReleases").append(metallb_helm_release)

            if ipam_enabled:
                LOG.info("Going to update provider spec")
                body.get("spec").get("providerSpec").get(
                    "value")["clusterNetwork"] = {
                    "clusterNetwork": True,
                    "cidr": ipam_cidr,
                    "ipamEnabled": ipam_enabled,
                    "gateway": ipam_gw,
                    "includeRanges": [ipam_incl_range],
                    # Exclude ranges aren't use now in tests. But defined
                    # in original teamples. Let's left them are disabled.
                    # "excludeRanges": [],
                    "nameservers": [ipam_ns]
                }

        elif provider == settings.EQUINIX_PROVIDER_NAME:
            machines_count = \
                int(settings.KAAS_CHILD_CLUSTER_MASTER_NODES_COUNT) + \
                int(settings.KAAS_CHILD_CLUSTER_SLAVE_NODES_COUNT)
            facility = utils.get_available_equinix_facility(
                settings.KAAS_EQUINIX_USER_API_TOKEN,
                settings.KAAS_EQUINIX_MACHINE_TYPE,
                machines_count)

            body['spec']['providerSpec']['value'][
                'apiVersion'] = "equinix.kaas.mirantis.com/v1alpha1"
            body['spec']['providerSpec']['value'][
                'kind'] = "EquinixMetalClusterProviderSpec"
            body['spec']['providerSpec']['value'][
                'facility'] = facility

        elif provider == settings.AZURE_PROVIDER_NAME:
            body['spec']['providerSpec']['value'][
                'apiVersion'] = "azure.kaas.mirantis.com/v1alpha1"
            body['spec']['providerSpec']['value'][
                'kind'] = "AzureClusterProviderSpec"
            body['spec']['providerSpec']['value'][
                'location'] = settings.KAAS_AZURE_CLUSTER_LOCATION

        cluster = self.__manager.api.kaas_clusters.create(
            namespace=self.name, body=body)
        return Cluster(self.__manager, cluster)

    def _get_machines_with_label(self, label, machines=None, strict=False):
        """

        :param label: check for label exist
        :param strict: check label value also
        :param machines: optional list with machines data as dict
        :return: list with machines
        """
        if machines:
            machines_data = machines
        else:
            machines_data = [machine.read().to_dict() for machine in
                             self.get_machines()]
        label_name = tuple(label.keys())[0]
        label_value = tuple(label.values())[0]
        catched = []
        for machine in machines_data:
            labels = machine.get('metadata', {}).get('labels', {})
            if strict and labels.get(label_name) == label_value:
                catched.append(machine)
            if not strict and labels.get(label_name):
                catched.append(machine)
        return catched

    def _get_storage_machines(self):
        """
        Get machines,that marked with label for ceph `storage`
        :param self:
        :return:
        """
        # labels hardcoded in kaas/core/frontend/src/consts.js
        # Alined with Machine.machine_type
        hostlabels = utils.get_labels_by_type('storage')
        return self._get_machines_with_label(hostlabels)

    def create_ceph_cluster(self, name,
                            cluster_name,
                            nodes_data=None,
                            version='',
                            monitors_count=3,
                            manager_count=3,
                            hdd_type="hdd",
                            pools=None,
                            rgw=None,
                            network=None):
        """
        :param name:
        :param cluster_name:
        :param nodes_data:
        :param version:
        :param monitors_count:
        :param manager_count:
        :param hdd_type:
        :param pools: The ceph pools data
        :param rgw: The ceph rgw data
        :param network: The ceph network configuration
        :return:
        """
        # TODO(alexz) Must be refactored, as soon bmhp will
        #  support disk_labeling.
        # also skip second disk, as default for local-volume-provisioner
        system_disk_pattern = ['/dev/vda', '/dev/hda', '/dev/sda',
                               '/dev/vdb', '/dev/hdb', '/dev/sdb']
        assert 0 < monitors_count <= 3, (
            "Num of monitors could not be more then 3 "
            "and should be at least 1")
        assert 0 < manager_count <= 3, (
            "Num of managers could not be more then 3 "
            "and should be at least 1")
        if not nodes_data:
            nodes_data = {}
            roles = ['mgr', 'mon', 'rgw']
            storage_machines_data = self._get_storage_machines()
            for storage_machine in storage_machines_data:
                storage_devices_names = [
                    device.get('name').split('/')[-1] for device
                    in storage_machine.get('status', {}).get(
                        'providerStatus', {}).get('hardware', {}).get(
                        'storage', {}) if device.get(
                        'name', {}) not in system_disk_pattern]

                nodes_data[storage_machine.get('metadata', {}).get(
                    'name', {})] = {'roles': roles, 'storageDevices': []}
                for storage_device in storage_devices_names:
                    nodes_data.get(storage_machine.get('metadata', {}).get(
                        'name', {})).get('storageDevices', {}).append(
                        {'name': storage_device, 'config': {
                            'deviceClass': hdd_type}})
        if pools is None:
            pools = [
                {
                    "deviceClass": hdd_type,
                    "name": "kubernetes",
                    "role": "kubernetes",
                    # NOTE(vsaienko): allow pool to survive when
                    # shutting down 1 osd.
                    # related https://mirantis.jira.com/browse/PRODX-9815
                    "replicated": {
                        "size": 2
                    },
                    "default": True
                }
            ]

        body = {'apiVersion': "kaas.mirantis.com/v1alpha1",
                'kind': "KaaSCephCluster",
                'metadata': {
                    "name": name,
                    "namespace": self.name,
                }}
        body["spec"] = {
            "cephClusterSpec": {
                "failureDomain": "host",
                "nodes": nodes_data,
                "pools": pools,
                "version": version
            },
            "k8sCluster": {
                "name": cluster_name,
                "namespace": self.name
            }
        }
        if network:
            body["spec"]["cephClusterSpec"]["network"] = network
        if rgw:
            body["spec"]["cephClusterSpec"]["rgw"] = rgw
        LOG.info("Creating Ceph cluster with next data: {}".format(
            body))
        return self.__manager.api.kaas_cephclusters.create(
            namespace=self.name, body=body)

    def get_machines(self):
        return self.__manager.get_machines(namespace=self.name)

    def get_machine(self, name):
        return self.__manager.get_machine(name=name, namespace=self.name)

    def get_lcmmachines(self):
        return self.__manager.get_lcmmachines(namespace=self.name)

    def get_lcmmachine(self, name):
        return self.__manager.get_lcmmachine(name=name, namespace=self.name)

    def get_baremetalhosts(self):
        return self.__manager.get_baremetalhosts(namespace=self.name)

    def get_baremetalhost(self, name):
        return self.__manager.get_baremetalhost(name=name, namespace=self.name)

    def get_ipam_hosts(self):
        return self.__manager.get_ipam_hosts(namespace=self.name)

    def get_ipam_host(self, name):
        return self.__manager.get_ipam_host(name=name, namespace=self.name)

    def get_ipam_ipaddrs(self):
        return self.__manager.get_ipam_ipaddrs(namespace=self.name)

    def get_ipam_ipaddr(self, name):
        return self.__manager.get_ipam_ipaddr(name=name, namespace=self.name)

    def get_ipam_subnets(self):
        return self.__manager.get_ipam_subnets(namespace=self.name)

    def get_ipam_subnet(self, name):
        return self.__manager.get_ipam_subnet(name=name, namespace=self.name)

    def create_bm_statics(self, child_data):
        """
        Create BM static resources, in NS:
        bmh
        subnets
        l2templates
        - only create new resources, never updated already existence.
        """
        subnets_data = child_data.get('ipam_subnets', [])
        bmhp_data = child_data.get('bmh_profiles', {})
        l2templates = child_data.get('l2templates', [])
        ipaddrs = child_data.get('ipaddrs', [])

        if subnets_data:
            existing_subnets = [p.name for p in self.get_ipam_subnets()]
            for subnet in subnets_data:
                if subnet['name'] in existing_subnets:
                    LOG.info(f"Subnet {subnet['name']} already exists")
                else:
                    LOG.info(f"Create Subnet '{subnet['name']}'")
                    self.create_ipam_subnet(subnet['name'], subnet['cidr'],
                                            labels=subnet.get('labels'),
                                            include_ranges=subnet.get(
                                                'includeRanges'),
                                            exclude_ranges=subnet.get(
                                                'excludeRanges'),
                                            gateway=subnet.get('gateway'),
                                            nameservers=subnet.get(
                                                'nameservers'))

        if bmhp_data:
            existing_bmhp = [p.name for p in self.get_baremetalhostprofiles()]
            for n, bmhp in bmhp_data.items():
                if n in existing_bmhp:
                    LOG.info(f"BareMetalHostProfile {n} already exists")
                else:
                    self.create_baremetalhostprofile(name=n, spec=bmhp['spec'],
                                                     labels=bmhp['labels'])
        if l2templates:
            existing_l2templates = [l2t.name for l2t in self.get_l2templates()]
            for l2template in l2templates:
                if l2template['name'] not in existing_l2templates:
                    LOG.debug(f"L2Template to create:\n{l2template}")
                    LOG.info(f"Creating L2Template: {l2template['name']}")
                    self.create_l2template(
                        l2template['name'],
                        l2template.get('npTemplate', None),
                        labels=l2template.get('labels'),
                        cluster_ref=l2template.get('clusterRef', None),
                        if_mapping=l2template.get('ifMapping'),
                        auto_if_mapping_prio=l2template.get(
                            'autoIfMappingPrio'),
                        spec=l2template.get('spec', None),
                    )
                    LOG.info(f"L2Template '{l2template['name']}' "
                             f"has been created")
                else:
                    LOG.warning(
                        f'l2template {l2template["name"]} already exist,'
                        f'skipping')
        if ipaddrs:
            existing_ipaddr = [ipaddr.name for ipaddr in
                               self.get_ipam_ipaddrs()]
            for ipaddr in ipaddrs:
                if ipaddr['name'] not in existing_ipaddr:
                    LOG.debug(f"ipaddrs to create:\n{ipaddr}")
                    LOG.info(f"Creating ipaddr: {ipaddr['name']}")
                    self.create_ipam_ipaddr(ipaddr['name'],
                                            spec=ipaddr['spec'],
                                            labels=ipaddr['labels'])
                    LOG.info(f"ipaddr '{ipaddr['name']}' has been created")
                else:
                    LOG.warning(
                        f'ipaddr {ipaddr["name"]} already exist, skipping')

    def create_ipam_subnet(self, name, cidr,
                           labels=None,
                           include_ranges=None,
                           exclude_ranges=None,
                           gateway=None,
                           nameservers=None):

        """
        Create IPAM Subnet object

        :param name: Subnet's name
        :param cidr: network address with mask in form of 0.0.0.0/0
        :param labels: Dict with labels for Subnet object
        :param include_ranges: list of pairs (firstIP,lastIP) to choose IP from
        :param exclude_ranges: list of single IP and/or ranges of IPs
                               (pairs of (firstIP, lastIP)) to exclude from
                               distribution
        :param gateway: Gateway IP in the subnet
        :param nameservers: list of IPs represents nameservers in the subnet

        :return: SubnetIpamMirantis object
        """
        labels = labels or {}
        body = {
            "apiVersion": "ipam.mirantis.com/v1alpha1",
            "kind": "Subnet",
            "metadata": {
                "name": name,
                "namespace": self.name,
                "labels": {
                    **labels
                }
            },
            "spec": {
                "cidr": cidr,
            },
        }
        if include_ranges:
            body['spec']['includeRanges'] = include_ranges
        if exclude_ranges:
            body['spec']['excludeRanges'] = exclude_ranges
        if gateway:
            body['spec']['gateway'] = gateway
        if nameservers:
            body['spec']['nameservers'] = nameservers
        return self.__manager.api.ipam_subnets.create(
            namespace=self.name,
            body=body
        )

    def create_ipam_ipaddr(self, name, spec, labels=None):

        """
        Create IPAM ipaddr object

        :param name: Ipaddr's name
        :param spec: spec
        :param labels: labels
        :return: IPaddrIpamMirantis object
        """
        labels = labels or {}
        body = {
            "apiVersion": "ipam.mirantis.com/v1alpha1",
            "kind": "IPaddr",
            "metadata": {
                "name": name,
                "namespace": self.name,
                "labels": {
                    **labels
                }
            },
            "spec": {
                **spec
            },
        }
        return self.__manager.api.ipam_ipaddrs.create(
            namespace=self.name,
            body=body
        )

    def get_iamrolebindings(self):
        return self.__manager.get_iamrolebindings(namespace=self.name)

    def get_iamrolebinding(self, name):
        return self.__manager.get_iamrolebinding(name=name,
                                                 namespace=self.name)

    def create_iamrolebinding(self, name, role, user):
        """
        Create IAMRoleBinding object

        :param name: IAMRoleBinding's name
        :param role: Role name that should be assigned
        :param user: Username that should be associated with role

        :return: IAMRoleBindings object
        """
        body = {
            "apiVersion": "iam.mirantis.com/v1alpha1",
            "kind": "IAMRoleBinding",
            "metadata": {
                "name": name,
                "namespace": self.name,
            },
            "role": {
                "name": role,
            },
            "user": {
                "name": user,
            },
        }
        return self.__manager.api.iam_rolebindings.create(namespace=self.name,
                                                          body=body)

    def create_iamclusterrolebinding(self, name, role, user, cluster):
        """
        Create IAMRoleBinding object

        :param name: IAMRoleBinding's name
        :param role: Role name that should be assigned
        :param user: Username that should be associated with role
        :param cluster: Cluster name

        :return: IAMClusterRoleBindings object
        """
        body = {
            "apiVersion": "iam.mirantis.com/v1alpha1",
            "kind": "IAMClusterRoleBinding",
            "metadata": {
                "name": name,
                "namespace": self.name,
            },
            "role": {
                "name": role,
            },
            "user": {
                "name": user,
            },
            "cluster": {
                "name": cluster,
            },
        }
        return self.__manager.api.iam_clusterrolebindings.create(
            namespace=self.name,
            body=body)

    def get_l2templates(self):
        return self.__manager.get_l2templates(namespace=self.name)

    def get_l2template(self, name):
        return self.__manager.get_l2template(name=name, namespace=self.name)

    def get_baremetalhostprofile(self, name):
        return self.__manager.get_baremetalhostprofile(name=name,
                                                       namespace=self.name)

    def get_baremetalhostprofiles(self):
        return self.__manager.get_baremetalhostprofiles(namespace=self.name)

    def create_l2template(self, name, np_template,
                          labels=None,
                          cluster_ref=None,
                          if_mapping=None,
                          l3Layout=None,
                          auto_if_mapping_prio=None,
                          spec=None):
        """
        Create L2Template object

        :param name: L2Template's name
        :param np_template: Netplan template as a string
        :param labels: Dict with labels for L2Template object
        :param cluster_ref: Name of cluster to apply the template. 'default'
                           means all cluster in the namespace
        :param if_mapping: list of interfaces to use in npTemplate
        :param auto_if_mapping_prio: list of prefixes to form ifMapping
                                     automatically in this order
        :param spec: Full spec wo any checks
        :return: L2TemplateIpamMirantis object
        """
        labels = labels or {}
        if spec:
            body = {
                "apiVersion": "ipam.mirantis.com/v1alpha1",
                "kind": "L2Template",
                "metadata": {
                    "name": name,
                    "namespace": self.name,
                    "labels": {
                        **labels
                    }
                },
                "spec": {**spec},
            }
            return self.__manager.api.l2templates.create(
                namespace=self.name,
                body=body)
        body = {
            "apiVersion": "ipam.mirantis.com/v1alpha1",
            "kind": "L2Template",
            "metadata": {
                "name": name,
                "namespace": self.name,
                "labels": {
                    **labels
                }
            },
            "spec": {
                "npTemplate": np_template,
            },
        }
        if cluster_ref:
            body['spec']['clusterRef'] = cluster_ref
        if if_mapping and isinstance(if_mapping, list):
            body['spec']['ifMapping'] = if_mapping
        elif auto_if_mapping_prio and isinstance(auto_if_mapping_prio, list):
            body['spec']['autoIfMappingPrio'] = auto_if_mapping_prio
        else:
            # Set default autoIfMappingPrio if both 'if_mapping' and
            # 'auto_if_mapping_prio' have not been provided
            body['spec']['autoIfMappingPrio'] = [
                'provision',  # is used in test envs so added to simplify life
                'eno',
                'ens',
                'enp',
            ]
        return self.__manager.api.l2templates.create(
            namespace=self.name,
            body=body
        )

    def create_baremetalhostprofile(self, name, spec, labels=None):
        body = {
            "apiVersion": "metal3.io/v1alpha1",
            "kind": "BareMetalHostProfile",
            "metadata": {
                "name": name,
                "namespace": self.name,
            },
            "spec": {**spec,
                     },
        }
        if labels:
            body['metadata']['labels'] = {**labels}
        return self.__manager.api.baremetalhostprofiles.create(
            namespace=self.name,
            body=body
        )

    def create_baremetalhost(self,
                             bmh_name,
                             bmh_secret,
                             bmh_mac,
                             bmh_ipmi,
                             hardwareProfile,
                             labels=None,
                             annotations=None,
                             bootUEFI=True):
        """
        Create node with kind: BareMetalHost
        :param bmh_name:
        :param bmh_secret:
        :param bmh_mac:
        :param bmh_ipmi:
        :return:
        """
        bootmode = 'UEFI' if bootUEFI else 'legacy'
        body = {
            "apiVersion": "metal3.io/v1alpha1",
            "kind": "BareMetalHost",
            "metadata": {
                "name": bmh_name,
                "labels": labels if labels else {},
                "annotations": annotations if annotations else {},
            },
            "spec": {
                "online": True,
                "bootMode": bootmode,
                "bootMACAddress": bmh_mac,
                "bmc": {
                    "address": bmh_ipmi,
                    "credentialsName": bmh_secret,
                }
            }
        }
        if hardwareProfile:
            body['spec']['hardwareProfile'] = hardwareProfile
        return self.__manager.api.kaas_baremetalhosts.create(
            # bmh objects must be set in default ns.
            namespace=self.name,
            body=body
        )

    def get_baremetalhost_status(self, name):
        return self.get_baremetalhost(name).data.get(
            "status", {}).get("provisioning", {}).get("state")

    def set_baremetalhost_power(self, ns, bmh_name, online=True):
        """
        online=True - Powered on
        online=False - hard power off
        """
        bmh = self.__manager.get_baremetalhost(bmh_name, namespace=ns)
        return bmh.patch({'spec': {'online': online}})

    def wait_baremetalhost_deletion(self, bmh_name, timeout=600):

        def status_msg():
            """To show the remaining machines status after wait timeout"""
            bmhs = [bmh.read()
                    for bmh in self.get_baremetalhosts()]
            bmhs_status = {
                m.metadata.name: m.status.provisioning.state for m in bmhs}
            return "Baremetal hosts status: {0}".format(bmhs_status)

        def check_deleted():
            bmh = [x for x in self.get_baremetalhosts() if x.name == bmh_name]
            if not bmh:
                LOG.info("BMH %s hasn't been found in cluster, "
                         "looks like it was deleted", bmh_name)
                return True

            LOG.info("BMH %s has not been deleted yet and has status - %s",
                     bmh_name, self.get_baremetalhost_status(bmh_name))
            raise Exception("Not deleted yet")

        timeout_msg = ("Machine {} has not been deleted during "
                       "timeout {} seconds").format(bmh_name, timeout)
        waiters.wait_pass(lambda: check_deleted(), timeout=timeout,
                          timeout_msg=timeout_msg,
                          status_msg_function=status_msg)

    def wait_openstack_resources_available(self, credential_name,
                                           retries=20,
                                           interval=15):

        def wait_resource():
            LOG.info("Wait until openstack resources are available")
            try:
                resource = [x for x in self.get_openstackresources()
                            if x.name == credential_name]
                if resource:
                    resource[0].read()
                else:
                    raise ApiException
                LOG.debug("Available openstackresource {0}"
                          .format(credential_name))
                return True
            except Exception as e:
                LOG.debug("Waiting for openstackresource {0}: {1}"
                          .format(credential_name, e))
            return False

        waiters.wait(lambda: wait_resource(),
                     timeout=retries * interval,
                     interval=interval,
                     timeout_msg="Openstack resource {0} is still "
                                 "not available after {1} retries".format(
                         credential_name, retries
                     ))

    def wait_baremetalhosts_statuses(self, retries=10, interval=5,
                                     wait_status='provisioned',
                                     wait_nodes_msg=None,
                                     nodes=None,
                                     check_error=True):
        """nodes: list of node names
         ['cz7987-child-controller-1', 'cz7842-child-worker-1']
         check_error: raise, if any of bmh in error state.
        """

        def get_statuses(bmhs, wait_nodes_msg):
            err_statuses = ['error']
            bm_statuses = {}
            if not wait_nodes_msg:
                wait_nodes_msg = \
                    "BMH {name} has <{status}> provisioning status"
            for bm in bmhs:
                bm_data = bm.read()
                name = bm_data.metadata.name
                if bm_data.status is None:
                    LOG.info("BMH %s status isn't ready yet", name)
                    bm_statuses[name] = ("BMH %s status isn't ready yet", name)
                    continue
                pstatus = bm_data.status.get(
                    'provisioning').get('state', 'Nil')
                LOG.info(wait_nodes_msg.format(name=name, status=pstatus))
                bm_statuses[name] = pstatus
                if check_error and bm_data.status.get(
                        'operationalStatus',
                        'Nil').lower() in err_statuses:
                    err_msg = ("Aborting. One of BMH in error status.See: "
                               "{machines}").format(machines=bm_statuses)
                    raise Exception(err_msg)
            return bm_statuses

        total_time = retries * interval
        bmhs = self.get_baremetalhosts()
        if nodes:
            bmhs = [bmh for bmh in bmhs if bmh.name in nodes]
        LOG.info("Wait {0} seconds until BMH will be in the status <{1}>"
                 .format(total_time, wait_status))
        try:
            waiters.wait(
                lambda: set(get_statuses(bmhs, wait_nodes_msg).values()) == set([wait_status]),  # noqa
                timeout=total_time, interval=interval)
        except TimeoutError as e:
            self.print_events()
            LOG.error("Not all baremetalhosts are in "
                      "{} status".format(wait_status))
            raise e

    def create_proxyobject(self, name=settings.KAAS_PROXYOBJECT_NAME,
                           region=settings.KAAS_PROXYOBJECT_REGION,
                           proxy_str=settings.KAAS_EXTERNAL_PROXY_ACCESS_STR,
                           no_proxy=settings.KAAS_PROXYOBJECT_NO_PROXY):
        """Create a proxy inside child cluster namespaces

        :param no_proxy: No proxy variable
        :param name: Name of proxyobject
        :param region: Region of proxyobject
        :param proxy_str: Generated access string
        :return:
        """

        LOG.info('Creating proxy object with name - {0}'.format(name))
        options = {
            'KAAS_PROXYOBJECT_NAME': name,
            'KAAS_PROXYOBJECT_REGION': region,
            'KAAS_PROXYOBECT_NAMESPACE': self.name,
            'KAAS_PROXYOBJECT_ACCESS_STR': proxy_str,
            'KAAS_PROXYOBJECT_NO_PROXY': no_proxy

        }

        templates_prx = template_utils.render_template(
            settings.KAAS_PROXYOBJECT_YAML, options)
        LOG.debug(templates_prx)
        proxy_obj = self.__manager.api.kaas_proxies.create(
            name=name,
            namespace=self.name,
            body=yaml.load(templates_prx, Loader=yaml.SafeLoader))
        return proxy_obj


class Cluster(object):
    def __init__(self, manager: Manager, kaas_cluster: V1KaaSCluster):
        """
        manager: <Manager> instance
        kaas_cluster: <KaaSCluster> instance
        """
        self.__manager = manager
        self.__cluster = kaas_cluster
        self.__metadata = None
        self.__bastion_ip = None
        self.__k8sclient = None
        self.__dockerclient = None
        self.__mkedashboardclient = None
        self.__provider_resources = None
        self.__expected_pods = None
        self.__expected_docker_objects = None
        self.__ssh_user = None
        self.__private_key = None
        self.__key_file_var = None
        self.__mcp_docker_registry = None
        self.__cluster_type = None
        self.__provider = None
        self.check = clustercheck_manager.ClusterCheckManager(self)

    @property
    @retry(KeyError, delay=10, tries=30, logger=LOG)
    def _cluster_type(self):
        if not self.__cluster_type:
            if self.provider == 'BYO':
                self.__cluster_type = 'byo_child'
                return self.__cluster_type
            kaas_spec = self.spec['providerSpec']['value'].get('kaas')
            if not kaas_spec:
                raise KeyError("No kaas object found in providerSpec value. "
                               "Kaas cluster is not ready")
            mgmt = kaas_spec.get('management', {}).get('enabled', False)
            if mgmt:
                self.__cluster_type = 'management'
            elif 'regional' in kaas_spec:
                self.__cluster_type = 'regional'
            else:
                self.__cluster_type = 'child'
        return self.__cluster_type

    @property
    def k8sclient(self):
        """Cached cluster k8sclient"""
        if self.__k8sclient is None:
            if self.is_management:
                self.__k8sclient = self.__manager.api
            else:
                _, kubeconfig = self.get_kubeconfig_from_secret()
                config_data = yaml.load(kubeconfig, Loader=yaml.SafeLoader)
                self.__k8sclient = K8sCluster(config_data=config_data)
        return self.__k8sclient

    @property
    def dockerclient(self):
        """Cached docker ssh client"""
        if self.__dockerclient is None:
            self.__dockerclient = self.get_dockerclient()
        return self.__dockerclient

    @property
    @retry(KeyError, delay=10, tries=30, logger=LOG)
    def mkedashboardclient(self):
        """Cached docker ssh client"""
        if self.__mkedashboardclient is None:
            self.__mkedashboardclient = self.get_mke_dashboardclient()
        return self.__mkedashboardclient

    @property
    def ssh_user(self):
        """Cached ssh_user property"""
        if self.__ssh_user is None:
            if self.is_management:
                self.__ssh_user = settings.KAAS_MGMT_CLUSTER_SSH_LOGIN
            elif self.is_regional:
                self.__ssh_user = settings.KAAS_REGIONAL_CLUSTER_SSH_LOGIN
            else:
                self.__ssh_user = settings.KAAS_CHILD_CLUSTER_SSH_LOGIN
        return self.__ssh_user

    @ssh_user.setter
    def ssh_user(self, ssh_user):
        self.__ssh_user = ssh_user

    @property
    def private_key(self):
        """Cached ssh_user property"""
        if self.__private_key is None:
            if self.is_management:
                self.__private_key = \
                    settings.KAAS_MGMT_CLUSTER_PRIVATE_KEY_FILE
            elif self.is_regional:
                self.__private_key = \
                    settings.KAAS_REGIONAL_CLUSTER_PRIVATE_KEY_FILE
            else:
                self.__private_key = \
                    settings.KAAS_CHILD_CLUSTER_PRIVATE_KEY_FILE
        return self.__private_key

    @private_key.setter
    def private_key(self, private_key):
        self.__private_key = private_key

    @property
    def key_file_var(self):
        """Cached key_file_var property"""
        if self.__key_file_var is None:
            if self.is_management:
                self.__key_file_var = 'KAAS_MGMT_CLUSTER_PRIVATE_KEY_FILE'
            elif self.is_regional:
                self.__key_file_var = 'KAAS_REGIONAL_CLUSTER_PRIVATE_KEY_FILE'
            else:
                self.__key_file_var = 'KAAS_CHILD_CLUSTER_PRIVATE_KEY_FILE'
        return self.__key_file_var

    @key_file_var.setter
    def key_file_var(self, key_file_var):
        self.__key_file_var = key_file_var

    def get_parent_cluster(self):
        """Returns parent Cluster object for current child"""
        if self.is_management:
            raise Exception("You are trying to get parent cluster "
                            "of management cluster.")
        else:
            if self.is_regional:
                LOG.debug("{} is regional cluster".format(self.name))
                return self
            provider_region_labels = ['kaas.mirantis.com/provider',
                                      'kaas.mirantis.com/region']
            # let's use cached metadata
            labels = {k: v for k, v in self.metadata['labels'].items()
                      if k in provider_region_labels}
            # return mgmt cluster as parent for byo
            if labels['kaas.mirantis.com/provider'] == 'byo':
                return [x for x in self.__manager.get_clusters()
                        if x.is_management][0]

            cluster_list = [x for x in self.__manager.get_clusters()
                            if x.name != self.name and
                            (x.is_management or x.is_regional)]
            if len(cluster_list) == 1:
                # we have only 1 cluster and it is mgmt
                return cluster_list[0]

            cluster_provider = labels['kaas.mirantis.com/provider']
            parents = []
            for cl in cluster_list:
                cl_data = cl.data
                candidate_provider = cl_data[
                    'metadata']['labels']['kaas.mirantis.com/provider']
                candidate_region = cl_data[
                    'metadata']['labels']['kaas.mirantis.com/region']

                # if region doesn't match - skip
                if candidate_region != labels['kaas.mirantis.com/region']:
                    continue
                # check if regional cluster is a parent itself
                elif candidate_provider == cluster_provider:
                    parents.append(cl)
                else:
                    # check if regional cluster contains
                    # desired provider as nested
                    nested_providers = cl_data[
                        'spec']['providerSpec']['value']['kaas']['regional']
                    for p in nested_providers:
                        if p['provider'] == cluster_provider:
                            parents.append(cl)
                            break

            if len(parents) != 1:
                parents = [x.metadata['name'] for x in parents]
                raise Exception(f"Found {len(parents)} parent clusters: "
                                f"{parents} but should have found 1")

            parent_name = parents[0].metadata['name']
            LOG.info("Cluster {0} belongs to parent cluster {1}".format(
                self.name, parent_name))
            return parents[0]

    @retry(Exception, delay=30, tries=10, logger=LOG)
    def __get_openid_k8sclient(self):
        LOG.info("Getting openid kubeconfig for {} cluster".format(self.name))
        providerStatus = self.data['status']['providerStatus']
        oidc = providerStatus['oidc']
        assert oidc['ready'], (f"OIDC is not ready for {self.name}")
        external_ip = providerStatus['loadBalancerHost']
        # initiate openid kubeconfig
        return K8sCluster(idp_issuer_url=oidc['issuerUrl'], host=external_ip,
                          idp_ca=oidc['certificate'],
                          client_id=oidc['clientId'],
                          cluster_name=self.name)

    def kubeconfig_secret_exists(self, secret_name=''):
        """Check if kubeconfig secret exists

        :param secret_name: str, (optional) use the specified secret
                            name instead of generated from cluster_name
        """
        client = self.get_parent_client()
        secret_name = secret_name or "{}-kubeconfig".format(self.name)
        LOG.debug("Fetching kubeconfig of {} cluster".format(self.name))
        kubeconfig_secret = [
            x for x in client.secrets.list(namespace=self.namespace)
            if x.name == secret_name
        ]
        return True if kubeconfig_secret else False

    def get_parent_client(self):
        """Returns parent client.
           If parent is a regional cluster, openid client will be returned.
           Note: openid client may not have access to some objects.
        """
        client = self.__manager.api
        try:
            if not self.is_management:
                parent = self.get_parent_cluster()
                if not parent.is_management:
                    # get openid config for regional cluster only
                    client = parent.__get_openid_k8sclient()
        except Exception as e:
            LOG.error("An error occurred {}".format(e))
            LOG.error("Fallback to old behavior (use mgmt client for "
                      "get_kubeconfig_from_secret)")
        return client

    def get_kubeconfig_from_secret(self, secret_name=''):
        """Get kubeconfig for the cluster

        :param secret_name: str, (optional) use the specified secret
                            name instead of generated from cluster_name
        """
        client = self.get_parent_client()
        secret_name = secret_name or "{}-kubeconfig".format(self.name)
        external_ip = self.data[
            'status']['providerStatus']['loadBalancerHost']
        LOG.debug("Fetching kubeconfig of {} cluster".format(self.name))
        kubeconfig_secret = client.secrets.get(
            secret_name, namespace=self.namespace)
        kubeconfig = kubeconfig_secret.read()
        insecure = 'insecure' in kubeconfig.data
        if 'external_ip' in kubeconfig.data:
            external_ip_encoded = kubeconfig.data['external_ip']
            external_ip = base64.b64decode(external_ip_encoded).decode('utf-8')
        kubeconfig = kubeconfig.data['admin.conf']
        kubeconfig = base64.b64decode(kubeconfig)
        kubeconfig = yaml.load(kubeconfig, Loader=yaml.SafeLoader)
        endpoint = "https://{}".format(external_ip)
        if endpoint not in kubeconfig['clusters'][0]['cluster']['server']:
            api_url = 'https://{external_ip}:443'.format(
                external_ip=external_ip
            )
            LOG.info("Replace public endpoint to %s", api_url)
            kubeconfig['clusters'][0]['cluster']['server'] = api_url
        else:
            curr_address = kubeconfig['clusters'][0]['cluster']['server']
            LOG.info("Current server address in kubeconfig "
                     "will not be replaced: {}".format(curr_address))
        if insecure:
            LOG.info("Set insecure-skip-tls-verify flag")
            kubeconfig['clusters'][0][
                'cluster']['insecure-skip-tls-verify'] = insecure
            if 'certificate-authority-data' in \
                    kubeconfig['clusters'][0]['cluster']:
                LOG.info("Remove certificate-authority-data")
                del kubeconfig['clusters'][0][
                    'cluster']['certificate-authority-data']
        kubeconfig = yaml.dump(kubeconfig)

        name = secret_name
        content = kubeconfig
        return (name, content)

    def get_dockerclient(self):
        """Get docker ssh client working directly or through bastion node"""
        return DockerCliClient(self)

    def get_mke_dashboardclient(self):
        dashboard_url = self.data.get('status', {}).get(
            'providerStatus', {}).get('ucpDashboard')
        if not dashboard_url:
            raise Exception("Cannot find ucpDashboard url in "
                            "cluster status: {}".format(self.data))

        LOG.info(">>> Cluster {0}/{1} dashboard URL: {2}"
                 .format(self.namespace,
                         self.name,
                         dashboard_url))
        if self.is_management:
            # TODO(ddmitriev): enable openid when allowed, PRODX-16245
            # client_id = 'kaas'
            secret_name = "ucp-admin-password-{0}".format(self.name)
            secret = self.__manager.api.secrets.get(
                secret_name, namespace=self.namespace)
            pwd_data = secret.read().to_dict()['data']['ucpAdminPassword']
            pwd = base64.b64decode(pwd_data).decode("utf-8")
            LOG.info(f">>> Use the UCP admin password '{pwd}' "
                     f"from the secret {secret_name}")

            d_client = mke_dashboard_client.MKEDashboardClientPlain(
                dashboard_url, settings.UCP_UI_USER, pwd)

        else:
            client_id = 'k8s'
            keycloak_ip = self.__manager.get_keycloak_ip()
            d_client = mke_dashboard_client.MKEDashboardClientOpenid(
                dashboard_url, keycloak_ip,
                settings.KAAS_UI_USER, settings.KAAS_UI_PASSWORD,
                client_id=client_id)

        return d_client

    def get_mcc_dashboardclient(self):
        child_client = self.k8sclient

        # Get kubernetes ip and port
        service = child_client.services.get(name='kaas-kaas-ui',
                                            namespace='kaas')
        ip = service.get_external_ip()
        service_port = [s for s in service.get_ports()
                        if s.name == 'http' or s.name == 'https']
        assert len(service_port) > 0, (
            "No http port found for the service kubernetes-dashboard-proxy")
        port = service_port[0].port

        # Get objects from kubernetes dashboard
        dashboard_url = "http://{0}:{1}".format(ip, port)
        LOG.info(f"MCC UI found at {dashboard_url}")

        d_client = mcc_dashboard_client.MCCDashboardClient(
            dashboard_url,
            settings.KAAS_UI_USER, settings.KAAS_UI_PASSWORD)
        return d_client

    @property
    def provider_resources(self):
        """Cached cluster provider class"""
        try:
            if self.__provider_resources is None:
                if self.provider == "OpenstackClusterProviderSpec":
                    self.__provider_resources = \
                        prm.OSProviderResources(self.__manager, self)
                elif self.provider == "AWSClusterProviderSpec":
                    self.__provider_resources = \
                        prm.AWSProviderResources(self.__manager, self)
                elif self.provider == "VsphereClusterProviderSpec":
                    self.__provider_resources = \
                        prm.VsphereProviderResources(self.__manager, self)
                else:
                    self.__provider_resources = \
                        prm.ProviderResources(self.__manager, self)
            return self.__provider_resources
        except Exception as e:
            LOG.error(e)

    @property
    def is_management(self):
        return True if self._cluster_type == 'management' else False

    @property
    def is_regional(self):
        """Returns True only if this is a standalone regional cluster"""
        return True if self._cluster_type == 'regional' else False

    @property
    def is_child(self):
        """Returns True only if this is a child cluster"""
        return True if self._cluster_type == 'child' else False

    def is_existed(self):
        ns = self.__manager.get_namespace(self.namespace)
        all_clusters = ns.get_clusters()
        for c in all_clusters:
            if c.name == self.name:
                LOG.info("Cluster object found")
                return True
        return False

    def is_ready(self, exp_provider_status, expected_fails):
        """
        Check if all cluster components and machines are ready
        :param exp_provider_status: expected overall cluster status
        :type exp_provider_status: bool
        :param expected_fails: list of strings with
        expectedly failed deployment names
        :type expected_fails: List[str]
        :rtype bool: bool
        """
        data = self.data
        if data['status']:
            unexpected_fails = []
            provider_status = data['status'] \
                .get('providerStatus', {}) \
                .get('ready')
            not_ready_obj = data['status'] \
                .get('providerStatus', {}). \
                get('notReadyObjects', {})
            LOG.debug(f"Provider ready status: {provider_status}")
            LOG.debug(f"Not ready objects: {not_ready_obj}")
            LOG.debug(f"Deployments expected to fail: {expected_fails}")
            not_ready_depl = not_ready_obj.get('deployments', [])

            if not_ready_depl:
                unexpected_fails = [d for d in not_ready_depl
                                    if d.get('name', '') not in expected_fails]

            if (provider_status and not unexpected_fails) or \
                    (provider_status == exp_provider_status
                     and not unexpected_fails):
                LOG.info(f"Provider ready status is {provider_status} "
                         f"as expected and no unexpected fails. "
                         f"Expected fails are: {expected_fails}")
                return True
            else:
                error_msg = "Current status: "
                if provider_status != exp_provider_status:
                    error_msg = error_msg + f"provider ready status " \
                                            f"{provider_status} expected " \
                                            f"to be {exp_provider_status} "
                if unexpected_fails:
                    error_msg = error_msg + f"unexpected fails " \
                                            f"{unexpected_fails}"
                LOG.error(error_msg)
                return False
        else:
            raise ValueError(f"No status in provider data, "
                             f"provider data: {data}")

    def get_conditions_status(self, data, expected_fails=None, verbose=False):
        expected_fails = expected_fails or {}
        result = {
            'ready': [],
            'not_ready': [],
            'skipped': [],
        }

        conditions = data.get('status', {}).get(
            'providerStatus', {}).get('conditions', [])
        if not conditions:
            result['not_ready'].append("NO CONDITIONS IN PROVIDER STATUS")

        for condition in conditions:
            ready = condition['ready']
            condition_type = condition['type']
            if (not ready and
                    condition_type in expected_fails.keys() and
                    expected_fails[condition_type] in condition['message']):
                result['skipped'].append(condition_type)
            elif ready is True:
                result['ready'].append(condition_type)
            else:
                result['not_ready'].append(condition_type)
        if verbose:
            message = (f"{data['kind']} {self.namespace}/"
                       f"{data['metadata']['name']}: {result}")
            LOG.info(message)
        return result

    def are_conditions_ready(self, expected_fails=None, verbose=False):
        """
        Check if all Cluster conditions in ready status
        :param expected_fails: None or dict like the following
            {<condition.type>: <expected message pattern>, ...}
            If expected message pattern is empty, then the whole
            condition is ignored. If it is not empty - then
            this pattern is matched to the message from the cluster
            to check if the message is expected or not.
        :rtype bool: bool
        """
        data = self.data
        result = self.get_conditions_status(data, expected_fails, verbose)
        if result['not_ready']:
            return False
        else:
            return True

    @property
    def name(self):
        """Cluster name"""
        return self.__cluster.name

    @property
    def namespace(self):
        """Cluster namespace"""
        return self.__cluster.namespace

    @property
    def metadata(self):
        """Cached cluster metadata"""
        if self.__metadata is None:
            self.__metadata = self.data['metadata']
        return self.__metadata

    @property
    def spec(self):
        """Cluster spec"""
        return self.data['spec']

    @property
    def uid(self):
        """Cluster uid"""
        return self.metadata['uid']

    @property
    def data(self):
        """Returns dict of k8s object

        Data contains keys like api_version, kind,

        metadata, spec, status or items
        """
        return self.__cluster.read().to_dict()

    def patch(self, *args, **kwargs):
        self.__cluster.patch(*args, **kwargs)

    def delete(self):
        """Deletes the current cluster"""
        # TODO(ddmitriev) delete cluster nodes first
        self.__cluster.delete()
        # TODO(ddmitriev) wait for deletion complete

    @property
    def cluster_status(self):
        """Determine the cluster status

        :rtype string: status of the cluster,
                       one of ('Pending', 'Updating', 'Ready')
        """
        status = self.data['status']
        if not status:
            return 'Pending'
        nodes = status.get('providerStatus', {}).get('nodes', {})
        if not nodes:
            return 'Pending'
        if nodes['ready'] != nodes['requested']:
            return 'Updating'
        return 'Ready'

    @property
    def clusterrelease_version(self):
        """Determine the cluster clusterrelease version

        :rtype string: clusterrelease of the cluster
        """
        status = self.data['status']
        if not status:
            return None
        current_clusterrelease = status.get('providerStatus', {}).get(
            'releaseRefs', {}).get('current', {}).get('name', '')
        if not current_clusterrelease:
            return None
        return current_clusterrelease

    @property
    def available_clusterrelease_version(self):
        """Determine first available clusterrelease version

        :rtype string clusterrelease of the cluster or None
        """
        status = self.data['status']
        if not status:
            return None
        available_clusterreleases_list = status.get('providerStatus', {}).get(
            'releaseRefs', {}).get('available', [{}])
        if len(available_clusterreleases_list) == 0:
            return None
        available_clusterrelease = available_clusterreleases_list[0].get(
            'name', '')
        if not available_clusterrelease:
            return None
        return available_clusterrelease

    @property
    def lcm_type(self):
        """Determine the cluster LCM engine type

        :rtype string: LCM type of the cluster, or "kubespray" as default,
                       or None if no status yet
        """
        status = self.data['status']
        if not status:
            return None
        current_release = status.get('providerStatus', {}).get(
            'releaseRefs', {}).get('current', {})
        if not current_release:
            return None
        return current_release.get('lcmType', 'kubespray')

    @property
    def lcm_type_is_ucp(self):
        return 'ucp' in self.lcm_type

    @property
    def lcm_type_is_kubespray(self):
        return 'kubespray' in self.lcm_type

    @property
    def cluster_proxy_name(self):
        current_cluster_proxy_name = self.data.get('providerSpec', {}).get(
            'proxy', {}).get('name')
        return current_cluster_proxy_name

    @property
    def provider(self):
        """Determine the cluster provider

        :rtype string: provider of the cluster,
                       one of ('AWSClusterProviderSpec',
                               'OpenstackClusterProviderSpec',
                               'BaremetalClusterProviderSpec',
                               'BYOMachineProviderSpec',
                               'VsphereMachineProviderSpec',
                               'EquinixMetalMachineProviderSpec',
                               'AzureClusterProviderSpec')
        """
        if self.__provider is None:
            self.__provider = self.spec['providerSpec']['value']['kind']
            assert self.__provider is not None
        return self.__provider

    @property
    def get_region_name(self):
        """Determine the cluster region"""
        cluster_data = self.data
        return cluster_data['metadata']['labels']['kaas.mirantis.com/region']

    def get_kaasrelease_version(self):
        if self.is_management or self.is_regional:
            cluster_data = self.data
            return cluster_data['spec'][
                'providerSpec']['value']['kaas']['release']
        else:
            LOG.warning(f"Cannot get kaasrelease version "
                        f"for cluster: {self.name} - "
                        f"its non management|regional cluster ")
            return False

    def get_cluster_annotations(self):
        cluster_data = self.data
        return cluster_data['metadata'][
            'annotations']

    def update_proxy(self, name, update_proxy):

        body = {
            "spec": {
                "httpProxy": update_proxy,
                "httpsProxy": update_proxy
            }
        }
        proxy_object = self.__manager.api.kaas_proxies.update(
            name=name, namespace=self.namespace, body=body)
        return proxy_object

    def get_cluster_annotation(self, annotation_key):
        """
        ---------
        metadata:
          annotations:
            kaas.mirantis.com/lcm: "true"
        ---------
        :param annotation_key: "kaas.mirantis.com/lcm"  # example value
        :return: {kaas.mirantis.com/lcm: "true"} # example return
        """
        annotation = None
        all_annotations = self.get_cluster_annotations()
        for ak, av in all_annotations.items():
            if ak == annotation_key:
                annotation = {ak: av}
        return annotation

    def add_cluster_annotation(self, annotation_key, annotation_value):
        """
        Adds the annotation "annotation_key" with value
        "annotation_value" and will overwrite the "annotation_key"
        label if it already exists.
        :return: api result
        """
        body = {
            "metadata": {
                "annotations": {
                    annotation_key: annotation_value}
            }
        }
        api_responce = self.__manager.api.kaas_clusters.update(
            name=self.name,
            namespace=self.namespace,
            body=body
        )
        return api_responce

    def remove_cluster_annotation(self, annotation_key):
        """
        For remove annotation, we should add annotation
        "annotation_key" with "annotation_value" = None
        """
        return self.add_cluster_annotation(annotation_key, None)

    def store_k8s_artifacts(self, secret_name=''):
        kubeconfig_name, kubeconfig = \
            self.get_kubeconfig_from_secret(secret_name)
        kubeconfig_path = os.path.join(
            settings.ARTIFACTS_DIR, kubeconfig_name)

        LOG.info("Save cluster kubeconfig to %s", kubeconfig_path)
        with open(kubeconfig_path, 'w') as f:
            f.write(kubeconfig)

        try:
            ver = self.k8sclient.api_version.get_code()
        except OSError as e:
            LOG.error(e)
            ver = self.k8sclient.api_version.get_code()
        k8s_version = "{0}.{1}.x".format(ver.major, ver.minor).replace("+", "")
        kubeconfig_version_path = os.path.join(
            settings.ARTIFACTS_DIR, kubeconfig_name + "_k8s_version")
        LOG.info("Save child cluster k8s version {0} to {1}"
                 .format(k8s_version, kubeconfig_version_path))
        with open(kubeconfig_version_path, 'w') as f:
            f.write(k8s_version)

    def store_machines_artifacts(self, machine_types=None):
        artifacts_dir = settings.ARTIFACTS_DIR
        m_types = machine_types or ["control", "worker"]
        LOG.info(f"Store IPs for the following machine types in the cluster "
                 f"{self.name}: {m_types}")

        for m_type in m_types:
            typed_machines = self.get_machines(machine_type=m_type)
            addrs = []
            for machine in typed_machines:
                addrs.append(machine.public_ip)
            filename = f"{artifacts_dir}/machines_{m_type}"
            with open(filename, "w") as f:
                f.write(",".join(addrs))
            LOG.info(f"'{m_type}' machine IPs have been saved to {filename}")

    def update_cluster(self, update_release_name):

        body = {
            "spec": {
                "providerSpec": {
                    "value": {
                        "release": update_release_name
                    }
                }
            }
        }

        cluster = self.__manager.api.kaas_clusters.update(
            name=self.name, namespace=self.namespace, body=body)
        return Cluster(self.__manager, cluster)

    def update_cluster_proxy(self, update_proxy_name=None):
        body = {
            "spec": {
                "providerSpec": {
                    "value": {
                        "proxy": update_proxy_name
                    }
                }
            }
        }

        self.__manager.api.kaas_clusters.update(
            name=self.name, namespace=self.namespace, body=body)

    def describe_public_services(self):
        """Find the services with external IP and http/https ports

        return: dict, key/values where key is service.name,
                value is list of URLs:
                {'iam-proxy-alerta': ['http://172.19.116.138'], ...}
        """
        services = self.k8sclient.services.list_all()
        result = {}
        for service in services:
            ip = service.get_external_ip()
            ports = service.get_ports()
            if ip:
                urls = []
                for port in ports:
                    if port.target_port in ['http', 'https']:
                        urls.append("{0}://{1}".format(port.target_port, ip))
                if urls:
                    result[service.name] = urls
        return result

    def get_helmbundles(self):
        client = self.__manager.api
        if not self.is_management:
            client = self.get_parent_cluster().k8sclient
        return client.kaas_helmbundles.list(
            namespace=self.namespace)

    def get_helmbundle(self, name=None, namespace=None):
        """Return specified helmbundle.
        By default helmbundle for current cluster is returned
        """
        if not namespace:
            namespace = self.namespace
        if not name:
            name = self.name
        client = self.__manager.api
        if not self.is_management:
            client = self.get_parent_cluster().k8sclient
        return client.kaas_helmbundles.get(name=name,
                                           namespace=namespace)

    def wait_machine_deletion(self, machine_name,
                              retries=20, interval=30):
        def status_msg():
            """To show the remaining machines status after wait timeout"""
            machines = [machine.read()
                        for machine in self.get_machines()]
            machines_status = {m.metadata.name: m.status for m in machines}
            return "Machines status: {0}".format(machines_status)

        def check_deleted():
            machine = self.get_machine(machine_name)
            if not machine:
                LOG.info("Machine %s was not found in cluster, "
                         "looks like it was deleted", machine_name)
                return True
            LOG.info("Machine %s has not been deleted yet and has status - %s",
                     machine_name, machine.machine_status)
            raise Exception("Machine has been deleted")

        total_time = retries * interval
        LOG.info("Wait {0} seconds until machine will be deleted"
                 .format(total_time))

        timeout_msg = ("Machine has not been deleted during "
                       "timeout {} seconds").format(total_time)
        waiters.wait_pass(lambda: check_deleted(), timeout=total_time,
                          timeout_msg=timeout_msg,
                          status_msg_function=status_msg)

    def get_lcmmachines(self, states=None):
        """Return list of lcmmachines filtered by states list"""
        return self.get_all_lcmmachines(states=states,
                                        namespace=self.namespace)

    def get_all_lcmmachines(self, states=None, namespace=None):
        """Return list of all lcmmachines filtered by states
        and namespace name
        """
        client = self.__manager.api
        if not self.is_management:
            client = self.get_parent_cluster().k8sclient
        if namespace:
            lcmm = client.kaas_lcmmachines.list_all(states=states)
            return [x for x in lcmm if x.namespace == namespace]
        else:
            return client.kaas_lcmmachines.list_all(states=states)

    def get_machine(self, name):
        all_machines = self.get_machines()
        ret = next((m for m in all_machines if m.name == name), None)
        return ret

    def get_machines(self, machine_type=None, machine_status=None):
        """Return list of Machine objects and set bastion_ip"""
        cluster_machines = self._get_machines()
        if machine_type is not None:
            cluster_machines = [m for m in cluster_machines
                                if m.is_machine_type(machine_type)]
        if machine_status is not None:
            cluster_machines = [m for m in cluster_machines
                                if m.machine_status == machine_status]
        return cluster_machines

    def get_machine_openstack_name(self, machine=None):
        """ Return name of instance in Openstack cloud
        :param machine: Machine Name or Object
        """
        if isinstance(machine, str):
            os_machine = self.get_machine(machine)
        else:
            os_machine = machine
        openstack_instance_uid = os_machine.data['metadata'][
            'annotations']['kaas.mirantis.com/uid']
        openstack_instance_name = "kaas-node-" + openstack_instance_uid
        return openstack_instance_name

    def get_related_cephclusters(self):
        return self.__manager.api.kaas_cephclusters.list(
            namespace=self.namespace)

    def get_controller_ip(self):
        machines = self._get_machines()
        for machine in machines:
            if machine.is_machine_type('control'):
                return machine.public_ip
        raise Exception("No controllers found in the cluster")

    def _get_machines(self):
        """Return list of Machine objects which are belong to the cluster"""
        # collect raw data of all machines in namespace
        machines = self.__manager.api.kaas_machines.list_raw(
            namespace=self.namespace).to_dict()['items']
        cl_name = 'cluster.sigs.k8s.io/cluster-name'
        # filter out machines not from current cluster
        cl_machines = [
            x for x in machines
            if x['metadata'].get('labels', {}).get(cl_name, '') == self.name
        ]

        # assuming all machines in cluster has the same provider
        if cl_machines:
            # get MachineClass based on the 1st machine in list
            MachineClass = self._get_machine_class(cl_machines[0])
        else:
            raise Exception(f"No Machines have been found for the cluster"
                            f" {self.namespace}/{self.name}")
        cl_machine_names = [x['metadata']['name'] for x in cl_machines]

        # collect kaas_machines objects
        machines_obj = [
            x for x in self.__manager.get_machines(namespace=self.namespace)
            if x.name in cl_machine_names
        ]

        cluster_machines = []
        for machine in machines_obj:
            cluster_machines.append(
                MachineClass(self.__manager, self, machine))
        return cluster_machines

    def _get_machine_class(self, kaasmachine):
        provider_machines = {
            'OpenstackMachineProviderSpec': OpenstackProviderMachine,
            'BareMetalMachineProviderSpec': BaremetalProviderMachine,
            'AWSMachineProviderSpec': AwsProviderMachine,
            'BYOMachineProviderSpec': ByoProviderMachine,
            'VsphereMachineProviderSpec': VsphereProviderMachine,
            'EquinixMetalMachineProviderSpec': EquinixMetalProviderMachine,
            'AzureMachineProviderSpec': AzureProviderMachine
        }
        machine_spec = kaasmachine['spec']
        machine_provider = machine_spec['providerSpec']['value']['kind']
        if machine_provider not in provider_machines:
            raise Exception(
                "Provider 'spec/providerSpec/value/kind: {0}' is not"
                " supported in si-tests, please"
                " update the code".format(machine_provider))
        AdapterClass = provider_machines[machine_provider]
        return AdapterClass

    @property
    def bastion_ip(self):
        if self.__bastion_ip:
            return self.__bastion_ip

        provider_status = self.data["status"].get("providerStatus", {})
        if 'bastion' in provider_status and provider_status.get('bastion'):
            bastion = provider_status["bastion"]
            # We have different field names for different providers right now
            if "publicIp" in bastion:
                self.__bastion_ip = bastion["publicIp"]
            elif "publicIP" in bastion:
                self.__bastion_ip = bastion["publicIP"]
            else:
                self.__bastion_ip = ""
        elif self.provider == "AWSClusterProviderSpec":
            # AWS machines never have public IPs
            raise Exception("No <bastion> data in the status "
                            "for cluster {0}".format(self.name))
        elif self.provider == "AzureClusterProviderSpec":
            # Azure machine access using host API server IP address as bastion
            assert 'loadBalancerHost' in provider_status, \
                    (f"loadBalancerHost must present in cluster {self.name} \
                     provider status as bastion host for azure machines")
            self.__bastion_ip = provider_status['loadBalancerHost']
            return self.__bastion_ip
        elif self.provider == "BaremetalClusterProviderSpec":
            return None
        else:
            # When bastion is not available
            cluster_machines = self._get_machines()
            # Use a k8s control node with public_ip for a bastion_ip
            # if any of control nodes have public_ip and is in Ready status
            for m in cluster_machines:
                LOG.debug("Machine {0} has type {1} and status {2}"
                          .format(m.name, m.machine_type, m.machine_status))
                if all([m.is_machine_type('control'),
                        m.machine_status == "Ready",
                        m.public_ip is not None]):
                    self.__bastion_ip = m.public_ip
                    break
        return self.__bastion_ip

    @bastion_ip.setter
    def bastion_ip(self, ip):
        self.__bastion_ip = ip

    def _get_bastion_remote(self, ssh_login="mcc-user",
                            ssh_key=None, key_file_var="Private key"):
        if not self.bastion_ip:
            raise Exception(
                "Bastion IP is not set for cluster <{0}>,"
                " SSHClient cannot be initialized".format(self.name))

        if not os.path.isfile(ssh_key):
            raise Exception(
                "{2} is not set for bastion "
                "IP {0} in the cluster <{1}>, "
                "SSHClient cannot be "
                "initialized".format(self.bastion_ip, self.name, key_file_var))

        keys = utils.load_keyfile(ssh_key)
        pkey = utils.get_rsa_key(keys['private'])

        auth = exec_helpers.SSHAuth(username=ssh_login,
                                    password='', key=pkey)
        ssh = exec_helpers.SSHClient(host=self.bastion_ip, port=22, auth=auth)
        ssh.logger.addHandler(logger.console)

        # Upload private key to the bastion node, if missing
        priv_key_dir = os.path.join("/home",
                                    ssh_login,
                                    ".ssh")
        priv_key_path = os.path.join(priv_key_dir, self.name)
        res = ssh.execute("test -f {0}".format(priv_key_path), verbose=False)
        if res.exit_code != 0:
            ssh.check_call("mkdir -p {0}; chmod 0700 {0}"
                           .format(priv_key_dir), verbose=False)
            ssh.upload(ssh_key,
                       priv_key_path)
            ssh.check_call("mkdir -p {0}; chmod 0600 {0}"
                           .format(priv_key_path), verbose=False)
        return ssh

    def get_bastion_remote(self):
        return self._get_bastion_remote(ssh_login=self.ssh_user,
                                        ssh_key=self.private_key,
                                        key_file_var=self.key_file_var)

    def download_kubernetes_logs(self, name=None, raise_on_error=False):
        """Try to download kubernetes logs, print exception to the log

        :param name: archive name (without .tar.gz)
        """
        try:
            self._download_kubernetes_logs(name=name)
        except Exception as e:
            LOG.error(e)
            LOG.error("Failed to download kubernetes logs for cluster <{0}>"
                      .format(self.name))
            if raise_on_error:
                raise

    def _download_kubernetes_logs(self, name=None):
        """Download pod logs and describe kubernetes resources

        Based on:
        https://github.com/openstack/openstack-helm-infra/tree/master/\
            roles/describe-kubernetes-objects/tasks
        https://github.com/openstack/openstack-helm-infra/tree/master/\
            roles/gather-pod-logs/tasks
        """
        if not name:
            if self.is_management:
                name = "mgmt_k8s_{0}".format(self.name)
            else:
                name = "child_k8s_{0}".format(self.name)

        control_nodes = self.get_machines(machine_type='control',
                                          machine_status='Ready')
        if not control_nodes:
            raise Exception("No control nodes in 'Ready' status found "
                            "in the cluster {0}".format(self.name))

        ctl = control_nodes[0]
        kube_dir = "{0}_{1}".format(name, self.cluster_status)
        ctl.run_cmd("sudo rm -rf /tmp/{0}* || true;"
                    "mkdir /tmp/{1}/;"
                    "sudo cp /root/ucp-bundle/kube.yml /tmp/{1}/kubeconfig &&"
                    " sudo chmod o+r /tmp/{1}/kubeconfig;"
                    .format(name, kube_dir),
                    verbose=True, check_exit_code=True)

        # TODO: In case vSphere supports Ubuntu need to find another way to
        # detect host OS
        if ctl.provider == "vsphere":
            ctl.run_cmd("sudo yum install -y rsync", verbose=False)
        else:
            ctl.run_cmd("sudo apt-get -qq install -y rsync", verbose=False)

        logs_dir_command = ("""
          set -ex
          export LOGS_DIR=/tmp/{0}
          export KUBECONFIG=/tmp/{1}/kubeconfig
        """.format(kube_dir, kube_dir))

        pod_logs_commands = ("""
PARALLELISM_FACTOR=4
function get_namespaces () {
  kubectl get namespaces -o name | awk -F '/' '{ print $NF }'
}
function get_pods () {
  NAMESPACE=$1
  kubectl get pods -n ${NAMESPACE} -o name | \
  awk -F '/' '{ print $NF }' | \
  xargs -L1 -P 1 -I {} echo ${NAMESPACE} {}
}
export -f get_pods
function get_pod_logs () {
  NAMESPACE=${1% *}
  POD=${1#* }
  INIT_CONTAINERS=$(kubectl get pod/$POD -n ${NAMESPACE} \
      -o jsonpath={.spec.initContainers[*].name})
  CONTAINERS=$(kubectl get pod/$POD -n ${NAMESPACE} \
      -o jsonpath={.spec.containers[*].name})
  for CONTAINER in ${INIT_CONTAINERS} ${CONTAINERS}; do
    echo "${NAMESPACE}/${POD}/${CONTAINER}"
    mkdir -p "${LOGS_DIR}/pod-logs/${NAMESPACE}/${POD}"
    mkdir -p "${LOGS_DIR}/pod-logs/failed-pods/${NAMESPACE}/${POD}"
    kubectl logs ${POD} -n ${NAMESPACE} -c ${CONTAINER} \
     > "${LOGS_DIR}/pod-logs/${NAMESPACE}/${POD}/${CONTAINER}.txt"
    kubectl logs --previous ${POD} -n ${NAMESPACE} -c ${CONTAINER} \
     > "${LOGS_DIR}/pod-logs/failed-pods/${NAMESPACE}/${POD}/${CONTAINER}.txt"\
     || true
  done
}
export -f get_pod_logs
get_namespaces | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
    bash -c 'get_pods "$@"' _ {} | \
  xargs -r -n 2 -P ${PARALLELISM_FACTOR} -I {} \
    bash -c 'get_pod_logs "$@"' _ {} \
  > ${LOGS_DIR}/dump_pod_logs_commands.log 2>&1
  """)

        non_namespaced_objects_commands = ("""
export OBJECT_TYPES=$(kubectl api-resources -o name --namespaced=false)
export PARALLELISM_FACTOR=4

function list_objects () {
  echo ${OBJECT_TYPES} | xargs -d ' ' -I {} -P1 -n1 bash -c 'echo "$@"' _ {}
}
export -f list_objects

function name_objects () {
  export OBJECT=$1
  kubectl get ${OBJECT} -o name | xargs -L1 -I {} -P1 -n1 \
      bash -c 'echo "${OBJECT} ${1#*/}"' _ {}
}
export -f name_objects

function get_objects () {
  input=($1)
  export OBJECT=${input[0]}
  export NAME=${input[1]#*/}
  echo "${OBJECT}/${NAME}"
  DIR="${LOGS_DIR}/objects/cluster/${OBJECT}"
  mkdir -p ${DIR}
  kubectl get ${OBJECT} ${NAME} -o yaml > "${DIR}/${NAME}.yaml"
  kubectl describe ${OBJECT} ${NAME} > "${DIR}/${NAME}.txt"
}
export -f get_objects

list_objects | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'name_objects "$@"' _ {} | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'get_objects "$@"' _ {} \
  > ${LOGS_DIR}/dump_non_namespaced_objects_commands.log 2>&1""")

        namespaced_objects_commands = ("""
export OBJECT_TYPES=$(kubectl api-resources -o name --namespaced=true)
export PARALLELISM_FACTOR=4
function get_namespaces () {
  kubectl get namespaces -o name | awk -F '/' '{ print $NF }'
}

function list_namespaced_objects () {
  export NAMESPACE=$1
  echo ${OBJECT_TYPES} | xargs -d ' ' -I {} -P1 -n1 \
      bash -c 'echo "${NAMESPACE} $@"' _ {}
}
export -f list_namespaced_objects

function name_objects () {
  input=($1)
  export NAMESPACE=${input[0]}
  export OBJECT=${input[1]}
  kubectl get -n ${NAMESPACE} ${OBJECT} -o name | xargs -L1 -I {} -P1 -n1 \
      bash -c 'echo "${NAMESPACE} ${OBJECT} $@"' _ {}
}
export -f name_objects

function get_objects () {
  input=($1)
  export NAMESPACE=${input[0]}
  export OBJECT=${input[1]}
  export NAME=${input[2]#*/}
  echo "${NAMESPACE}/${OBJECT}/${NAME}"
  DIR="${LOGS_DIR}/objects/namespaced/${NAMESPACE}/${OBJECT}"
  mkdir -p ${DIR}
  kubectl get -n ${NAMESPACE} ${OBJECT} ${NAME} -o yaml > "${DIR}/${NAME}.yaml"
  kubectl describe -n ${NAMESPACE} ${OBJECT} ${NAME} > "${DIR}/${NAME}.txt"
}
export -f get_objects

get_namespaces | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'list_namespaced_objects "$@"' _ {} | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'name_objects "$@"' _ {} | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'get_objects "$@"' _ {} \
  > ${LOGS_DIR}/dump_namespaced_objects_commands.log 2>&1""")

        get_objects_commands = ("""
export OBJECT_TYPES=$(kubectl api-resources -o name)
export PARALLELISM_FACTOR=4

function list_resources () {
  echo ${OBJECT_TYPES} | xargs -d ' ' -I {} -P1 -n1 bash -c 'echo "$@"' _ {}
}
export -f list_resources

function get_objects () {
  export OBJECT=$1
  kubectl get ${OBJECT} -o wide --all-namespaces > \
      ${LOGS_DIR}/get_${OBJECT}.txt || true
  if [ ! -s ${LOGS_DIR}/get_${OBJECT}.txt ]; then
      rm ${LOGS_DIR}/get_${OBJECT}.txt
  fi
}
export -f get_objects

list_resources | \
  xargs -r -n 1 -P ${PARALLELISM_FACTOR} -I {} \
      bash -c 'get_objects "$@"' _ {} \
  > ${LOGS_DIR}/dump_get_objects_commands.log 2>&1""")

        res = ctl.run_cmd('hostname', verbose=False, verbose_info=True)
        LOG.info("Get kubernetes logs from the cluster <{0}> on Machine <{1}> "
                 "with local hostname <{2}>"
                 .format(self.name, ctl.name, res.stdout_str))

        # Download pod logs
        logs_target = "{0}/dump_pod_logs_commands.log".format(kube_dir)
        ret = ctl.run_cmd("{0}{1}"
                          .format(logs_dir_command, pod_logs_commands),
                          verbose=False, check_exit_code=False)
        if ret.exit_code != 0:
            LOG.error("Failed collecting pod logs from cluster <{0}> with "
                      "exit code <{1}>, see {2} for details"
                      .format(self.name, ret.exit_code, logs_target))

        # Download non-namespaced objects
        logs_target = "{0}/dump_non_namespaced_objects_commands.log".format(
            kube_dir)
        ret = ctl.run_cmd("{0}{1}"
                          .format(logs_dir_command,
                                  non_namespaced_objects_commands),
                          verbose=False, check_exit_code=False)
        if ret.exit_code != 0:
            LOG.error("Failed collecting non-namespaced k8s objects from "
                      "cluster <{0}> with exit code <{1}>, see {2} for details"
                      .format(self.name, ret.exit_code, logs_target))

        # Download namespaced objects
        logs_target = "{0}/dump_namespaced_objects_commands.log".format(
            kube_dir)
        ret = ctl.run_cmd("{0}{1}"
                          .format(logs_dir_command,
                                  namespaced_objects_commands),
                          verbose=False, check_exit_code=False)
        if ret.exit_code != 0:
            LOG.error("Failed collecting namespaced k8s objects from cluster "
                      "<{0}> with exit code <{1}>, see {2} for details"
                      .format(self.name, ret.exit_code, logs_target))

        # List all objects
        logs_target = "{0}/dump_get_objects_commands.log".format(
            kube_dir)
        ret = ctl.run_cmd("{0}{1}"
                          .format(logs_dir_command,
                                  get_objects_commands),
                          verbose=False, check_exit_code=False)
        if ret.exit_code != 0:
            LOG.error("Failed to get k8s object lists from cluster "
                      "<{0}> with exit code <{1}>, see {2} for details"
                      .format(self.name, ret.exit_code, logs_target))

        # Archive the logs
        ctl.run_cmd(
            "tar --absolute-names --warning=no-file-changed"
            " -C /tmp/{0} -czf /tmp/{0}.tar.gz ./".format(kube_dir),
            verbose=False)

        # Copy kubernetes logs from k8s master node to the bastion node
        # if the bastion node is not the same node as k8s master
        bastion_remote = self.get_bastion_remote()
        if self.bastion_ip != ctl.public_ip:
            priv_key_path = os.path.join("/home",
                                         settings.KAAS_CHILD_CLUSTER_SSH_LOGIN,
                                         ".ssh", self.name)
            LOG.info("Copy kubernetes logs from {0} to the bastion node {1}"
                     .format(ctl.name, self.bastion_ip))
            bastion_remote.check_call(
                "rsync -aruvz -e 'ssh -o StrictHostKeyChecking=no -i {0}' "
                "{1}:/tmp/{2}.tar.gz "
                "/tmp/{2}.tar.gz".format(priv_key_path,
                                         ctl.internal_ip or ctl.public_ip,
                                         kube_dir),
                verbose=False)

        # Download the logs
        artifact_name = os.path.join(settings.ARTIFACTS_DIR,
                                     "{0}.tar.gz".format(kube_dir))
        bastion_remote.download(destination="/tmp/{0}.tar.gz"
                                .format(kube_dir), target=artifact_name)
        LOG.info("Kubernetes logs were downloaded to {0}"
                 .format(os.path.join(artifact_name)))

    def download_system_logs(self, name=None, raise_on_error=False):
        """Try to download system logs, print exception to the log

        :param name: archive name (without .tar.gz)
        """
        try:
            self._download_system_logs(name=name)
        except Exception as e:
            LOG.error(e)
            LOG.error("Failed to download system logs for cluster <{0}>"
                      .format(self.name))
            if raise_on_error:
                raise

    def _download_system_logs(self, name=None):
        """Collect system logs from all cluster nodes

        1. rsync logs from cluster nodes to the node with bastion_ip
        2. tar.gz logs on bastion_ip node and download the archive
        """
        if not name:
            if self.is_management:
                name = "mgmt_system_logs_{0}".format(self.name)
            else:
                name = "child_system_logs_{0}".format(self.name)

        dump_commands = (
            r"mkdir /tmp/{0}/;"
            r"rsync -aruv --exclude journal /var/log /tmp/{0}/;"
            r"rsync -aruv --exclude /etc/*shadow* /etc /tmp/{0}/;"
            r"cp -r /root/lcm-ansible* /tmp/{0}/;"
            r"hostname -f > /tmp/{0}/hostname.txt;"
            r"dpkg -l > /tmp/{0}/dump_dpkg_l.txt ||"
            r" rpm -qa > /tmp/{0}/dump_rpm_l.txt;"
            r"df -h > /tmp/{0}/dump_df.txt;"
            r"free > /tmp/{0}/dump_free.txt;"
            r"cat /proc/cpuinfo > /tmp/{0}/dump_cpuinfo.txt;"
            r"mount > /tmp/{0}/dump_mount.txt;"
            r"blkid -o list > /tmp/{0}/dump_blkid_o_list.txt;"
            r"iptables -t nat -S > "
            r"  /tmp/{0}/dump_iptables_nat.txt;"
            r"iptables -S > /tmp/{0}/dump_iptables.txt;"
            r"ps auxwwf > /tmp/{0}/dump_ps.txt;"
            r"vgdisplay > /tmp/{0}/dump_vgdisplay.txt;"
            r"lvdisplay > /tmp/{0}/dump_lvdisplay.txt;"
            r"ip a > /tmp/{0}/dump_ip_a.txt;"
            r"ip r > /tmp/{0}/dump_ip_r.txt;"
            r"netstat -anp > /tmp/{0}/dump_netstat.txt;"
            r"brctl show > /tmp/{0}/dump_brctl_show.txt ||"
            r" bridge link show > /tmp/{0}/dump_bridge_show.txt;"
            r"arp -an > /tmp/{0}/dump_arp.txt;"
            r"uname -a > /tmp/{0}/dump_uname_a.txt;"
            r"lsmod > /tmp/{0}/dump_lsmod.txt;"
            r"cat /proc/interrupts > "
            r"  /tmp/{0}/dump_interrupts.txt;"
            r"cat /etc/*-release > /tmp/{0}/dump_release.txt;"
            r"timedatectl status > /tmp/{0}/timedata.txt;"
            r"systemctl --state=failed --no-pager --all > "
            r"  /tmp/{0}/systemctl_failed.txt;"
            r"journalctl -p err --no-pager > /tmp/{0}/journalctl_errors.txt;")

        dump_controller_commands = (
            r"etcdctl --key-file /etc/kubernetes/ssl/apiserver-etcd-client.key"
            r"      --cert-file /etc/kubernetes/ssl/apiserver-etcd-client.crt"
            r"      --ca-file /etc/kubernetes/ssl/etcd/ca.crt"
            r"      --endpoint=https://127.0.0.1:2379"
            r"      cluster-health > /tmp/{0}/etcd_cluster_health.txt;"
            r"calicoctl node status > /tmp/{0}/calico_node_status.txt;")

        bastion_remote = self.get_bastion_remote()
        # Prepare target dir on the bastion node
        cluster_dir = "{0}_{1}".format(name, self.cluster_status)
        bastion_remote.check_call("sudo rm -rf /tmp/{0}* || true;"
                                  "mkdir /tmp/{1}"
                                  .format(name, cluster_dir),
                                  verbose=False)
        # Set private key path on the bastion node
        priv_key_path = os.path.join("/home",
                                     settings.KAAS_CHILD_CLUSTER_SSH_LOGIN,
                                     ".ssh", self.name)

        cluster_machines = self.get_machines()
        for m in cluster_machines:
            res = m.run_cmd('hostname', verbose=False, verbose_info=True)
            node_dir = "{0}_{1}_{2}".format(m.machine_type, m.name,
                                            m.machine_status)
            LOG.info("Get system logs from the Machine <{0}> with local "
                     "hostname <{1}>"
                     .format(m.name, res.stdout_str))
            # Re-create the directory to store the artifacts
            m.run_cmd("sudo rm -rf /tmp/{0}_{1}_* || true;"
                      "mkdir /tmp/{2}".format(m.machine_type,
                                              m.name, node_dir),
                      verbose=False)
            m.run_cmd("sudo apt-get -qq install -y rsync", verbose=False)

            # Collect node logs. Do not fail on errors
            m.run_cmd('sudo bash -cx "{0}" > /tmp/{1}/dump_commands.log 2>&1'
                      .format(dump_commands.format(node_dir), node_dir),
                      verbose=False, check_exit_code=False)
            if m.is_machine_type('control'):
                # run k8s controller-specific commands
                m.run_cmd('sudo bash -cx "{0}" > '
                          '/tmp/{1}/dump_controller_commands.log 2>&1'
                          .format(dump_controller_commands.format(node_dir),
                                  node_dir),
                          verbose=False, check_exit_code=False)
            m.run_cmd('sudo chmod o+rx -R /tmp/{0}/*'.format(node_dir),
                      verbose=False)

            # Copy system logs from nodes to the bastion node
            # Try to use internal IP first
            bastion_remote.check_call(
                "rsync -aruvz -e 'ssh -o StrictHostKeyChecking=no -i {3}' "
                "{0}:/tmp/{1} /tmp/{2}/".format(m.internal_ip or m.public_ip,
                                                node_dir, cluster_dir,
                                                priv_key_path),
                verbose=False)
        # Archive the logs
        bastion_remote.check_call(
            "tar --absolute-names --warning=no-file-changed"
            " -C /tmp/{0} -czf /tmp/{0}.tar.gz ./".format(cluster_dir),
            verbose=False)
        # Download the logs
        artifact_name = os.path.join(settings.ARTIFACTS_DIR,
                                     "{0}.tar.gz".format(cluster_dir))
        bastion_remote.download(destination="/tmp/{0}.tar.gz"
                                .format(cluster_dir), target=artifact_name)
        LOG.info("System logs were downloaded to {0}"
                 .format(os.path.join(artifact_name)))

    def get_ceph_storage_nodes(self):
        """
        Total ceph-osd pods must be equal to Total osd disks in cluster
        """
        kubectl_client = self.k8sclient
        # Name and namespace hardcoded in rook
        rookcephcls = \
            [x for x in
             kubectl_client.rookcephclusters.list(namespace='rook-ceph')
             if x.name == 'rook-ceph']
        if not rookcephcls:
            LOG.info("Rookcephcluster 'rook-ceph' was not found")
            return []
        else:
            rookceph = rookcephcls[0].read().to_dict()
            return rookceph['spec']['storage']['nodes']

    def _render_expected_pods_template(self):
        """Render the template with expected pods for the cluster release"""

        all_machines = self.get_machines()

        ctl_nodes = [x.data['metadata']['name'] for x in all_machines
                     if x.is_machine_type('control')]
        num_ctl_nodes = len(ctl_nodes)

        num_worker_nodes = len(
            [x for x in all_machines
             if x.is_machine_type('worker')])
        num_storage_nodes = len(
            [x for x in all_machines
             if x.is_machine_type('storage')]
        )
        worker_or_storage_nodes = [
            x.data['metadata']['name'] for x in all_machines
            if x.is_machine_type('worker') or x.is_machine_type('storage')]
        # in dedicatedControlPlane=False the storage might be collocated
        # with ctl
        num_worker_or_storage_nodes = len(
            set(worker_or_storage_nodes) - set(ctl_nodes))
        if num_storage_nodes == 0:
            # for mgmt cluster we won't have machines w/ 'storage'
            # labels so let's set default value to num_ctl_nodes (3)
            LOG.debug(
                f"Number of storage nodes is set to {num_ctl_nodes} "
                "(equals to number of control nodes)"
            )
            num_storage_nodes = num_ctl_nodes
        num_nodes = len(all_machines)
        machines_info = \
            self.__manager.api.kaas_machines.list_raw(
                namespace=self.namespace).items
        os_compute_label = {'key': 'openstack-compute-node',
                            'value': 'enabled'}
        os_computes_num = len([
            x for x in machines_info
            if os_compute_label in x.spec['providerSpec']
            ['value'].get('nodeLabels', [])
        ])

        clusterrelease_version = \
            self.data['spec']['providerSpec'][
                'value']['release'].replace("-rc", "")

        if self.sl_ha_enabled():
            LOG.debug("HA in SL is enabled")
            prometheus_server_num = 2
            elasticsearch_master_num = 3
            patroni_num = 3
        else:
            LOG.debug("HA in SL is disabled")
            prometheus_server_num = 1
            elasticsearch_master_num = 1
            patroni_num = 1

        if self.logging_enabled():
            LOG.debug("Logging in SL is enabled")
            logging_enabled = True
        else:
            LOG.debug("Logging in SL is disabled")
            logging_enabled = False

        if self.tf_enabled():
            LOG.debug("TF is enabled")
            tf_enabled = True
        else:
            LOG.debug("This is OVS deployment")
            tf_enabled = False

        ns = self.__manager.get_namespace(self.namespace)

        cephclusters = [
            cephcl for cephcl in ns.get_cephclusters()
            if cephcl.data['spec']['k8sCluster']['name'] == self.name and
            cephcl.data['spec']['k8sCluster']['namespace'] == self.namespace
        ]

        rgw_num = 0
        mon = []
        crashcollector_num = 0
        rook_ceph_osd = []
        rgw_pods = []
        if cephclusters:
            cephspecs = cephclusters[0].data['spec']['cephClusterSpec']
            mon = [node for node in cephspecs['nodes'].values()
                   if 'mon' in node.get('roles', [])]
            ceph_osd = 0
            ceph_nodes = self.get_ceph_storage_nodes()
            for node in ceph_nodes:
                for disks in node.get('devices', []):
                    ceph_osd += 1
            crashcollector_num = len(cephspecs['nodes'])
            for i in range(0, ceph_osd):
                rook_ceph_osd.append("rook-ceph-osd-{}".format(i))
            if 'rgw' in cephspecs:
                rgw_num = cephspecs.get('rgw', {})\
                    .get('gateway', {}).get('instances', 0)

            for i in range(0, rgw_num):
                store_name = cephspecs['rgw']['name']
                rgw_pods.append(
                    f"rook-ceph-rgw-{store_name}-{string.ascii_letters[i]}")

        # ntpd constant
        ntpd_num = 2

        options = {
            'logging_enabled': logging_enabled,
            'num_ctl_nodes': num_ctl_nodes,
            'lma_child_enabled': self.is_sl_enabled(),
            'num_worker_nodes': num_worker_nodes,
            'num_storage_nodes': num_storage_nodes,
            'num_worker_or_storage_nodes': num_worker_or_storage_nodes,
            'num_nodes': num_nodes,
            'ntpd_num': ntpd_num,
            'rook_ceph_osd': rook_ceph_osd,
            'crashcollector_num': crashcollector_num,
            'num_ceph_mon_nodes': len(mon),
            'prometheus_alertmanager_num': 2,
            'prometheus_server_num': prometheus_server_num,
            'elasticsearch_master_num': elasticsearch_master_num,
            'patroni_num': patroni_num,
            'rgw_pods': rgw_pods,
            'os_computes_num': os_computes_num,
            'tf_enabled': tf_enabled,
        }

        templates = template_utils.render_template(
            settings.EXPECTED_PODS_TEMPLATES_DIR +
            clusterrelease_version +
            '.yaml', options
        )

        json_body = yaml.load(templates, Loader=yaml.SafeLoader)
        return json_body

    @property
    def expected_pods(self):
        if not self.__expected_pods:
            self._refresh_expected_objects()
        return self.__expected_pods

    @property
    def expected_docker_objects(self):
        if not self.__expected_docker_objects:
            self._refresh_expected_objects()
        return self.__expected_docker_objects

    def _refresh_expected_objects(self):
        """Force update expected objects after cluster scale"""
        self.__expected_pods, self.__expected_docker_objects = \
            self.get_expected_objects()

    def get_expected_objects(self, exclude_pods=None):
        """Returns list of expected pods (and their quantity for the cluster)
           plus all docker objects
        """
        if self.is_regional:
            cluster_type = 'regional'
        elif self.is_management:
            cluster_type = 'management'
        else:
            cluster_type = 'child'

        json_body = self._render_expected_pods_template()
        all_pods = {}
        all_objects = {}

        provider = self.provider

        if exclude_pods:
            LOG.warning("These pods will be excluded from "
                        "expected pod list {}".format(exclude_pods))
        else:
            exclude_pods = []
        for ns in json_body[cluster_type][provider]:
            pods = json_body[cluster_type][provider][ns]
            pods = {k: v for k, v in pods.items() if k not in exclude_pods}
            if ns not in all_pods:
                all_pods[ns] = pods
            else:
                all_pods[ns].update(pods)

        for component in self.get_components():
            if component in json_body[cluster_type]['components']:
                for ns in json_body[cluster_type]['components'][component]:
                    pods = json_body[cluster_type]['components'][component][ns]
                    pods = {k: v for k, v in pods.items()
                            if k not in exclude_pods}
                    if ns not in all_pods:
                        all_pods[ns] = pods
                    else:
                        all_pods[ns].update(pods)
            else:
                LOG.error("Component {} was not found in "
                          "expected pod templates".format(component))

        # fetching docker objects
        cluster_type = f"docker_{cluster_type}"
        for opt in json_body[cluster_type][provider]:
            objects = json_body[cluster_type][provider][opt]
            if opt not in all_objects:
                all_objects[opt] = objects
            else:
                all_objects[opt].update(objects)
        LOG.info("Expected docker objects:\n{0}"
                 .format(json.dumps(all_objects, indent=4)))

        LOG.info(json.dumps(all_pods, indent=4))
        return (all_pods, all_objects)

    def get_components(self):
        """Returns all components defined for current cluster
          (istio, harbor, etc)
        """
        release_names = []
        available_releases = self.spec['providerSpec']['value'].get(
            'helmReleases', [])
        if available_releases:
            release_names = [x['name'] for x in available_releases
                             if x.get('enabled', True)]
        if self.provider == 'EquinixMetalMachineProviderSpec' and not \
                self.is_regional and not self.is_management:
            release_names.append("ceph-controller")
            release_names.append("storage-discovery")

        return release_names

    def is_sl_enabled(self):
        """Check if StackLight enabled for the cluster"""
        return 'stacklight' in self.get_components()

    def sl_ha_enabled(self):
        if 'stacklight' not in self.get_components():
            LOG.warning("Stacklight component is not enabled")
            return None
        if self.is_management or self.is_regional:
            return True

        hb = self.get_helmbundle().data
        for chart in hb['spec']['releases']:
            if chart['name'] == 'stacklight':
                return bool(chart['values']['highAvailabilityEnabled'])

        LOG.warning("Stacklight release is not found in helm bundle")
        return None

    def tf_enabled(self):
        if self.is_management:
            return False
        if os.path.exists(settings.BM_CHILD_SETTINGS_YAML_PATH):
            child_data = yaml.safe_load(template_utils.render_template(
                settings.BM_CHILD_SETTINGS_YAML_PATH))[0]
            child_cluster_data = child_data.get('child_config')
            tf = child_cluster_data.get('tungstenfabric', [])
        else:
            tf = False
        if tf:
            return True
        else:
            return False

    def logging_enabled(self):
        if 'stacklight' not in self.get_components():
            LOG.warning("Stacklight component is not enabled")
            return None
        if self.is_management or self.is_regional:
            return True

        hb = self.get_helmbundle().data
        for chart in hb['spec']['releases']:
            if chart['name'] == 'stacklight':
                if 'logging' in chart['values']:
                    return bool(chart['values']['logging']['enabled'])
                # in old versions parameter is absent, but logging is enabled
                else:
                    return True

        LOG.warning("Stacklight release is not found in helm bundle")
        return None

    def openstack_lma_enabled(self):
        if 'stacklight' not in self.get_components():
            LOG.warning("Stacklight component is not enabled")
            return None
        if self.is_management or self.is_regional:
            return False

        hb = self.get_helmbundle().data
        for chart in hb['spec']['releases']:
            if chart['name'] == 'stacklight':
                if 'openstack' in chart['values']:
                    return bool(chart['values']['openstack']['enabled'])
                else:
                    return False

        LOG.warning("Stacklight release is not found in helm bundle")
        return None

    def get_allowed_node_labels(self):
        cluster_cr = self.clusterrelease_version or \
                     self.data['spec']['providerSpec']['value']['release']
        cr_version = \
            self.__manager.get_clusterrelease(name=cluster_cr)
        if 'allowedNodeLabels' in cr_version.data['spec'].keys():
            allowed_node_labels = cr_version.data['spec']['allowedNodeLabels']
            return {x['key']: x['value'] for x in allowed_node_labels}
        else:
            LOG.info("This cluster doesn't support Machine NodeLabels")
            return {}

    def create_machine(self,
                       node_type,
                       node_flavor,
                       node_image,
                       node_az,
                       node_sg=None,
                       region="region-one",
                       provider="openstack",
                       node_labels=None):
        name_prefix = "{0}-{1}-".format(self.name, node_type)

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": name_prefix,
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "openstackproviderconfig.k8s.io/v1alpha1",  # noqa
                        "kind": "OpenstackMachineProviderSpec",
                        "flavor": node_flavor,
                        "image": node_image,
                        "availabilityZone": node_az,
                    }
                }
            }
        }

        if node_labels:
            body['spec']['providerSpec']['value']['nodeLabels'] = node_labels

        if node_sg:
            body['spec']['providerSpec']['value']['securityGroups'] = node_sg

        if node_type == "master":
            labels = body['metadata']['labels']
            labels['cluster.sigs.k8s.io/control-plane'] = 'true'
        elif node_type == "node":
            pass  # for possible future changes
        else:
            raise NotImplementedError("Machine type {0} is not "
                                      "supported".format(node_type))

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def create_aws_machine(self,
                           node_type,
                           instance_type,
                           ami,
                           root_device_size=40,
                           region="region-one",
                           provider="aws",
                           node_labels=None):
        name_prefix = "{0}-{1}".format(self.name, node_type)

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": name_prefix,
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "aws.kaas.mirantis.com/v1alpha1",  # noqa
                        "kind": "AWSMachineProviderSpec",
                        "instanceType": instance_type,
                        "ami": {
                            "id": ami
                        },
                        "rootDeviceSize": root_device_size
                    }
                }
            }
        }

        if node_labels:
            body['spec']['providerSpec']['value']['nodeLabels'] = node_labels

        if node_type == "master":
            labels = body['metadata']['labels']
            labels['cluster.sigs.k8s.io/control-plane'] = 'true'
        elif node_type == "node":
            pass  # for possible future changes
        else:
            raise NotImplementedError("Machine type {0} is not "
                                      "supported".format(node_type))

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def create_vsphere_machine(self,
                               node_type,
                               template,
                               rhel_license,
                               root_device_size=120,
                               region="vsphere",
                               provider="vsphere",
                               node_labels=None):

        name_prefix = "{0}-{1}".format(self.name, node_type)
        if settings.KAAS_CHILD_CLUSTER_MACHINES_PREFIX:
            name_prefix = "{}-{}-".format(
                settings.KAAS_CHILD_CLUSTER_MACHINES_PREFIX,
                name_prefix)

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": name_prefix,
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "vsphere.cluster.k8s.io/v1alpha1",
                        "kind": "VsphereMachineProviderSpec",
                        "rhelLicense": rhel_license,
                        "template": template,
                        "diskGiB": root_device_size,
                        "numCPUs": 8,
                        "memoryMiB": 16384,
                    },
                    "labels": {
                        "cluster.sigs.k8s.io/cluster-name": self.name,
                        "kaas.mirantis.com/region": region,
                        "kaas.mirantis.com/provider": provider
                    }
                }
            }
        }

        if node_labels:
            body['spec']['providerSpec']['value']['nodeLabels'] = node_labels

        if node_type == "master":
            labels = body['metadata']['labels']
            labels['cluster.sigs.k8s.io/control-plane'] = 'true'
        elif node_type == "node":
            pass  # for possible future changes
        else:
            raise NotImplementedError("Machine type {0} is not "
                                      "supported".format(node_type))

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def create_equinix_machine(self,
                               node_type,
                               os,
                               billing_cycle,
                               machine_type,
                               region="region-one",
                               provider="equinixmetal",
                               facility=None,
                               node_labels=None):
        name_prefix = "{0}-{1}".format(self.name, node_type)

        facility = facility or settings.KAAS_EQUINIX_FACILITY_REGION

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": name_prefix,
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "equinix.kaas.mirantis.com/v1alpha1",  # noqa
                        "kind": "EquinixMetalMachineProviderSpec",
                        "OS": os,
                        "billingCycle": billing_cycle,
                        "machineType": machine_type,
                        "facility": facility
                    }
                }
            }
        }

        if node_labels:
            body['spec']['providerSpec']['value']['nodeLabels'] = node_labels

        if node_type == "master":
            labels = body['metadata']['labels']
            labels['cluster.sigs.k8s.io/control-plane'] = 'true'
        elif node_type == "node":
            pass  # for possible future changes
        else:
            raise NotImplementedError("Machine type {0} is not "
                                      "supported".format(node_type))

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def create_azure_machine(self,
                             node_type,
                             os,
                             vm_size,
                             vm_disk_size,
                             region="region-one",
                             provider=settings.AZURE_PROVIDER_NAME,
                             node_labels=None):
        name_prefix = "{0}-{1}".format(self.name, node_type)

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": name_prefix,
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    "kaas.mirantis.com/region": region,
                    "kaas.mirantis.com/provider": provider
                }
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": "azure.kaas.mirantis.com/v1alpha1",  # noqa
                        "kind": "AzureMachineProviderSpec",
                        "vmSize": vm_size,
                        "osDisk": {
                            "osType": os,
                            "diskSizeGB": vm_disk_size,
                            }
                    }
                }
            }
        }

        if node_labels:
            body['spec']['providerSpec']['value']['nodeLabels'] = node_labels

        if node_type == "master":
            labels = body['metadata']['labels']
            labels['cluster.sigs.k8s.io/control-plane'] = 'true'
        elif node_type == "node":
            pass  # for possible future changes
        else:
            raise NotImplementedError("Machine type {0} is not "
                                      "supported".format(node_type))

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def create_baremetal_machine(self,
                                 node_pubkey_name,
                                 labels,
                                 genname,
                                 matchlabels=False,
                                 baremetalhostprofile=False,
                                 node_labels=False,
                                 l2TemplateSelector=None):
        """
        :param genname: Machine generateName
        :param node_pubkey_name:
        :param matchlabels:  Dict with extra labels,  to be passed into
        hostSelector
        :param labels: Dict with labels to be passed into Machine
        :param node_labels: Dict with labels to be passed into
         providerSpec.nodeLabels
        :param labels: Dict with labels to be passed into Machine obj
        :param l2TemplateSelector: Dict with l2TemplateSelector data
        :return:
        """
        mlabels = dict()
        if matchlabels:
            mlabels = matchlabels

        # define actual version of kaasrelease, which needed only to guess
        # bmApiversion
        cluster = self.__manager.get_mgmt_cluster()
        actual_kaasrelease = cluster.get_kaasrelease_version()
        LOG.info(f"Actual (deployed) kaasrelease version is\n"
                 f"{actual_kaasrelease}")

        body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Machine",
            "metadata": {
                "generateName": f"{genname}-",
                "labels": {
                    "cluster.sigs.k8s.io/cluster-name": self.name,
                    **labels}
            },
            "spec": {
                "providerSpec": {
                    "value": {
                        "apiVersion": self.__manager.get_bmApiversion(
                            actual_kaasrelease),
                        "kind": "BareMetalMachineProviderSpec",
                        "hostSelector": {
                            "matchLabels": {**mlabels}
                        },
                        "publicKeys": [
                            {"name": node_pubkey_name},
                        ]
                    }
                }
            }
        }
        if baremetalhostprofile:
            body['spec']['providerSpec']['value'][
                'bareMetalHostProfile'] = baremetalhostprofile
        if l2TemplateSelector:
            body['spec']['providerSpec']['value'][
                'l2TemplateSelector'] = l2TemplateSelector
        # Process nodeLabels
        nlabels = dict()
        if node_labels:
            nlabels = node_labels
        set_node_labels = []
        allowed_node_labels = self.get_allowed_node_labels()
        for k, v in nlabels.items():
            if k.startswith('hotfix/'):
                continue
            if k in allowed_node_labels.keys() and v == allowed_node_labels[k]:
                set_node_labels.append({'key': k, 'value': v})
            else:
                set_node_labels.append({'key': k, 'value': v})
                LOG.error("Key/value {0}:{1} is not in "
                          "allowed node label list".format(k, v))
        if set_node_labels:
            body['spec']['providerSpec']['value'][
                'nodeLabels'] = set_node_labels

        return self.__manager.api.kaas_machines.create(
            namespace=self.namespace,
            body=body
        )

    def get_bmc_data(self, bmh):
        """
        return BMC data(host,user,password) for specific bmh
        """
        bmh_data = bmh.data
        bmc_host = bmh_data['spec']['bmc']['address'].split(":")[0]
        bmc_cred_name = bmh_data['spec']['bmc']['credentialsName']
        bmc_cred_data = self.__manager.api.secrets.get(
            name=bmc_cred_name, namespace=self.namespace).read().to_dict()
        bmc_username = base64.b64decode(
            bmc_cred_data['data']['username']).decode("utf-8")
        bmc_password = base64.b64decode(
            bmc_cred_data['data']['password']).decode("utf-8")
        return bmc_host, bmc_username, bmc_password

    def describe_fixed_resources(self):
        """Describe the cluster resource properties

        Describe the resource properties that should not be changed
        during the Cluster update process
        """
        resources = [
            self.k8sclient.nodes,
            self.k8sclient.services,
        ]
        data = {}
        for resource in resources:
            data.update(resource.describe_fixed_resources())
        return data

    def docker_ps(self):
        """Show info from 'docker ps', 'top' and 'df'"""
        machines = self.get_machines()
        for machine in machines:
            cmd = ("docker ps --all; top -bn 1| head -n6; df -h|"
                   "grep -e 'Filesystem\\| /$'")
            LOG.info(f"Collecting docker container statuses from "
                     f"{machine.name}, see debug log for details")
            result = machine.exec_pod_cmd(cmd, get_events=False, verbose=False)
            LOG.debug(f"Machine {machine.name}:\n" + result["logs"].strip())
            LOG.debug(f"Machine {machine.name}:\n" + result["logs"].strip())

    def get_cluster_description(self, system_info=False):
        cluster_release = self.clusterrelease_version
        provider_types = []
        if self.is_management:
            cluster_name = f"Management cluster: '{self.name}'"
        elif self.is_regional:
            cluster_name = f"Region cluster: '{self.name}'"
        else:
            cluster_name = f"Child cluster: '{self.name}'"
        lcm_type = self.lcm_type
        machines = self.get_machines()
        mstatus = {}
        docker_versions = set()
        kernel_versions = set()
        system_versions = set()
        for machine in machines:
            if machine.provider not in provider_types:
                provider_types.append(machine.provider)
            if machine.machine_type not in mstatus:
                mstatus[machine.machine_type] = {}
            if machine.machine_status not in mstatus[machine.machine_type]:
                mstatus[machine.machine_type][machine.machine_status] = 0
            mstatus[machine.machine_type][machine.machine_status] += 1
            if system_info:
                system_versions.add(utils.get_system_version(machine))
                kernel_versions.add(utils.get_kernel_version(machine))
                docker_versions.add(utils.get_docker_version(machine))
        msg = (f"{cluster_name} ({self.uid})"
               f" release: '{cluster_release}' provider: {provider_types}"
               f" lcm_type: '{lcm_type}'\n"
               f"    machines: {mstatus}\n")
        if system_info:
            msg += (f"    system versions: {system_versions}\n"
                    f"    kernel versions: {kernel_versions}\n"
                    f"    docker versions: {docker_versions}\n")
        return msg

    def apply_hotfix_labels(self, bm_nodes):
        LOG.info("Trying to apply hotfix labels")
        machines = self.get_machines()
        for machine in machines:
            hwid = machine.get_bmh_id()
            for bm_node in bm_nodes:
                k8s_new_labels = {}
                if hwid == bm_node[
                        'bmh_labels']['kaas.mirantis.com/baremetalhost-id']:
                    bm_node_labels = bm_node.get('node_labels')
                    if bm_node_labels:
                        for k, v in bm_node_labels.items():
                            if k.startswith('hotfix/'):
                                k8s_new_labels.update(
                                    {k.replace('hotfix/', ''): v}
                                )

                if k8s_new_labels:
                    k8s_node = machine.get_k8s_node()
                    LOG.info("Set new labels {0} for child K8S"
                             " node {1}, hwid {2}".format(k8s_new_labels,
                                                          k8s_node.name,
                                                          hwid))
                    k8s_node.patch({'metadata': {'labels': k8s_new_labels}})
        LOG.info("Applying hotfix labels is completed")

    def remove_k8s_node_all_labels(self, bm_machine_name):
        machine = self.get_machine(name=bm_machine_name)
        k8s_node = machine.get_k8s_node()
        k8s_node.patch({'metadata': {'labels': {'nolabels': 'nolabels'}}})

    def add_osdpl_overrides(self, data):
        """
        Get data in format:
          child_data.yaml node['name']: <node overrides dict>
        Creates a secret in child cluster openstack namespace:
          data:
            kubernetes-node-name: base64(<node overrides dict>)
        """
        if not data:
            return
        LOG.info("Add osdpl overrides")
        machines = self.get_machines()
        kubectl_client_child = self.k8sclient
        secret_name = "osdpl-nodes-override"
        secret_data = {}
        for machine in machines:
            hwid = machine.get_bmh_id()
            for node_name, node_data in data.items():
                if hwid == node_name:
                    k8s_node = machine.get_k8s_node()
                    secret_data[k8s_node.name] = base64.b64encode(
                        json.dumps(node_data).encode('utf-8')).decode()
        kubectl_client_child.secrets.create(
            namespace=NAMESPACE.openstack,
            body={
                "apiVersion": "v1",
                "kind": "Secret",
                "metadata": {
                    "name": secret_name,
                    "namespace": NAMESPACE.openstack,
                },
                "data": secret_data
            }
        )
        LOG.info("Add osdpl overrides completed")

    def get_ceph_data(self, namespace):
        name = settings.CEPH_CLUSTER_NAME
        crd = [crd for crd in self.k8sclient.kaas_cephclusters.list(
            namespace=namespace) if crd.name == name]
        assert crd, ("CRD with name {} is not found".format(name))
        return crd[0]

    def patch_ceph_data(self, data, crd):
        LOG.info("crd: {0}: ".format(crd))
        crd.patch(body=data)

    def replace_ceph_data(self, data, crd):
        LOG.info("crd: {0}: ".format(crd))
        crd.replace(body=data)

    def get_ceph_tool_pod(self):
        ceph_tools_pod = self.k8sclient.pods.list(
            namespace=settings.ROOK_CEPH_NS,
            name_prefix='rook-ceph-tools')
        assert len(ceph_tools_pod) > 0, ("No pods found with prefix "
                                         "rook-ceph-tools in namespace {}".
                                         format(settings.ROOK_CEPH_NS))
        return ceph_tools_pod[0]

    def get_ceph_operator_pod(self):
        ceph_operator_pod = self.k8sclient.pods.list(
            namespace=settings.ROOK_CEPH_NS,
            name_prefix='rook-ceph-operator')
        assert len(ceph_operator_pod) > 0, ("No pods found with prefix "
                                            "rook-ceph-operator in "
                                            "namespace {}".
                                            format(settings.ROOK_CEPH_NS))
        return ceph_operator_pod[0]

    def determine_mcp_docker_registry(self,
                                      pod_name="helm-controller-",
                                      pod_ns="kube-system",
                                      default="mirantis.azurecr.io"):
        """Determine MCP docker registry from the specified pod"""
        if self.__mcp_docker_registry is not None:
            return self.__mcp_docker_registry

        pods = self.k8sclient.pods.list(
            namespace=pod_ns,
            name_prefix=pod_name)
        if len(pods) == 0:
            LOG.warning(f"Pod {pod_ns}/{pod_name} not found. "
                        f"Use the default mcp docker registry: {default}")
            return default

        pod = pods[0].read()
        containers = pod.spec.containers
        image = containers[0].image
        patterns = [i for i in image.split("/") if i and 'http' not in i]

        self.__mcp_docker_registry = patterns[0]
        LOG.info(f"Found the MCP docker registry for "
                 f"{self.namespace}/{self.name} at: "
                 f"{self.__mcp_docker_registry}")
        return self.__mcp_docker_registry


class Machine(object):
    """Provider independent methods for the kaas machine"""

    def __init__(self, manager, cluster, kaasmachine):
        """
        manager: <Manager> instance
        cluster: <Cluster> instance with attribute "bastion_ip"
                 "bastion_ip" is a public IP address of a jump host that
                 is connected to the internal network with other nodes
                 in the cluster
        kaasmachine: <KaaSMachine> or <KaaSBaremetalHosts> instance
        """
        self.__manager = manager
        self.__cluster = cluster
        self.__kaasmachine = kaasmachine
        self.__uid = None
        self.__metadata = None
        self.__spec = None
        self.__machine_type = None
        self.__machine_types = None

    @property
    def provider(self):
        raise NotImplementedError("Abstract method")

    @property
    def internal_ip(self):
        raise NotImplementedError("Abstract method")

    @property
    def public_ip(self):
        raise NotImplementedError("Abstract method ")

    def power_off(self):
        raise NotImplementedError("Method is not implemented for current"
                                  " provider yet")

    def power_on(self):
        raise NotImplementedError("Method is not implemented for current"
                                  " provider yet")

    @property
    def _manager(self):
        return self.__manager

    @property
    def _cluster(self):
        return self.__cluster

    @property
    def name(self):
        """Machine name"""
        return self.__kaasmachine.name

    @property
    def namespace(self):
        """Machine namespace"""
        return self.__kaasmachine.namespace

    @property
    def metadata(self):
        """Cached machine metadata"""
        if self.__metadata is None:
            self.__metadata = self.data['metadata']
        return self.__metadata

    @property
    def spec(self):
        """Cached machine spec"""
        if self.__spec is None:
            self.__spec = self.data['spec']
        return self.__spec

    @property
    def uid(self):
        """Machine uid"""
        if self.__uid is None:
            self.__uid = self.metadata['uid']
        return self.__uid

    @property
    def machine_type(self):
        """Cached machine type
        :rtype string: One of ('control', 'worker', 'storage')
        """
        if self.__machine_type is None:
            self.__machine_type = \
                utils.get_type_by_labels(self.metadata['labels'])
        return self.__machine_type

    @property
    def machine_types(self):
        """Cached machine types
        :rtype list: with any of valid machine types:
            'control', 'worker', 'storage'
        """
        if self.__machine_types is None:
            self.__machine_types = \
                utils.get_types_by_labels(self.metadata['labels'])
        return self.__machine_types

    def is_machine_type(self, mtype):
        """
        :mtype string: One of ('control', 'worker', 'storage')
        """
        if mtype in self.machine_types:
            return True
        return False

    @property
    def machine_status(self):
        """Machine status from providerStatus

        Status object may disappear for a short periods, return None
        :rtype string: One of (None, 'Ready', 'Pending', 'Updating', ...)
        """
        data = self.data
        if data and data['status'] and data['status']['providerStatus']:
            return data['status']['providerStatus'].get('status')
        else:
            return None

    def get_k8s_node_name(self):
        if self.metadata['annotations']:
            node_name = 'kaas-node-' + \
                        self.metadata['annotations']['kaas.mirantis.com/uid']
            return node_name

    def get_k8s_node(self):
        node_name = self.get_k8s_node_name()
        if node_name:
            return self._cluster.k8sclient.nodes.get(
                name=node_name
            )

    def check_k8s_nodes_has_labels(self):
        node_meta = self.get_k8s_node().read().metadata
        nodeLabels = {x['key']: x['value'] for x in self.nodeLabels}
        not_assigned_labels = self._cluster.get_allowed_node_labels()
        result = {'Missing labels': [], 'Unexpected labels': []}
        if nodeLabels:
            for k, v in nodeLabels.items():
                if k not in node_meta.labels.keys() or \
                        v != node_meta.labels[k]:
                    result['Missing labels'].append({k: v})
                del not_assigned_labels[k]
        for k, v in not_assigned_labels.items():
            if k in node_meta.labels.keys():
                result['Unexpected labels'].append({k: v})

        if result['Unexpected labels'] or result['Missing labels']:
            LOG.warning("Labels mismatch: {0}. K8S node "
                        "labels: {1}".format(result, node_meta.labels))
            return False
        return True

    def get_bmh_id(self):
        _id = self.spec['providerSpec']['value'].get(
            'hostSelector', {}).get('matchLabels', {}).get(
            'kaas.mirantis.com/baremetalhost-id')
        if not _id:
            LOG.warning(f"HostSelector label for link <machine> and <bmh>"
                        f"not found for machine:{self.name}")
        return _id

    def get_bmh_name(self):
        _name = self.metadata['annotations'].get(
            'metal3.io/BareMetalHost')
        if not _name:
            LOG.warning(f"HostSelector label for link <machine> and <bmh>"
                        f"not found for machine:{self.name}")
        else:
            _name = _name.split('/')[1]
        return _name

    @property
    def data(self):
        """Returns dict of k8s object

        Data contains keys like api_version, kind,

        metadata, spec, status or items
        """

        return self.__kaasmachine.read().to_dict()

    @property
    def nodeLabels(self):
        spec = self.data['spec']
        if 'nodeLabels' in spec['providerSpec']['value']:
            return spec['providerSpec']['value']['nodeLabels']
        return []

    def add_labels(self, labels):
        if self.nodeLabels:
            existing_labels = self.nodeLabels
            LOG.info("Existing labels: {}".format(existing_labels))
            # TODO maybe check for duplicates
            existing_labels += labels
            body = \
                {"spec": {
                    'providerSpec': {
                        'value': {'nodeLabels': existing_labels}}}}
        else:
            body = \
                {"spec": {
                    'providerSpec': {
                        'value': {'nodeLabels': labels}}}}
        self.__manager.api.kaas_machines.update(
            name=self.name, namespace=self.namespace, body=body)

    def remove_labels(self, labels=None):
        if labels:
            updated_labels = [x for x in self.nodeLabels
                              if x not in labels]
            body = \
                {"spec": {
                    'providerSpec': {
                        'value': {'nodeLabels': updated_labels}}}}
        else:
            body = \
                {"spec": {
                    'providerSpec': {
                        'value': {'nodeLabels': None}}}}
        self.__manager.api.kaas_machines.update(
            name=self.name, namespace=self.namespace, body=body)

    def delete(self):
        """Deletes the current machine"""

        self.__kaasmachine.delete()
        # TODO(ddmitriev) wait for deletion complete

    def _run_cmd(self, cmd, verbose=True, timeout=None,
                 expected_codes=None, check_exit_code=True,
                 verbose_info=False, ssh_key=None, ssh_login="mcc-user",
                 sudo=False, reconnect=False):
        """Direct or through-bastion shell command execution"""

        expected_codes = expected_codes or [0]
        machine_ip = self.public_ip or self.internal_ip

        keys = utils.load_keyfile(ssh_key)
        pkey = utils.get_rsa_key(keys['private'])
        auth = exec_helpers.SSHAuth(username=ssh_login,
                                    password='', key=pkey)

        if self.public_ip or (self.internal_ip and
                              not self.__cluster.bastion_ip):
            # For public IP or directly accessible internal ip
            if verbose_info:
                LOG.info("SSH to the kaas node <{0}> using machine IP {1}"
                         .format(self.name, machine_ip))
            ssh = exec_helpers.SSHClient(
                host=machine_ip,
                port=22,
                auth=auth,
                )
            ssh.logger.addHandler(logger.console)
            if reconnect:
                ssh.reconnect()
            if sudo:
                with ssh.sudo(True):
                    result = ssh.execute(cmd, verbose=verbose, timeout=timeout)
            else:
                result = ssh.execute(cmd, verbose=verbose, timeout=timeout)
        elif self.internal_ip:
            if not self.__cluster.bastion_ip:
                raise Exception("Node {0} don't have bastion_ip to access the "
                                "internal address {1}"
                                .format(self.name, self.internal_ip))

            if verbose_info:
                LOG.info("SSH to the kaas node <{0}> using bastion IP '{1}' "
                         "and internal IP '{2}'"
                         .format(self.name,
                                 self.__cluster.bastion_ip,
                                 self.internal_ip))

            ssh = self.__cluster._get_bastion_remote(ssh_login=ssh_login,
                                                     ssh_key=ssh_key)
            result = ssh.execute_through_host(self.internal_ip, cmd,
                                              auth=auth, verbose=verbose,
                                              timeout=timeout)
        else:
            raise Exception("Node {0} don't have neither "
                            "public nor internal IP. "
                            "Current ones are Bastion {1} "
                            "internal {2}".format(self.name,
                                                  self.__cluster.bastion_ip,
                                                  self.internal_ip))

        if result.exit_code not in expected_codes and check_exit_code:
            message = (
                "Command {result.cmd!r} returned exit code "
                "{result.exit_code!s} while expected {expected!s}\n"
                "STDERR: {stderr!s}".format(
                    result=result, expected=expected_codes,
                    stderr=result.stderr_str
                )
            )
            LOG.error(msg=message)
            raise Exception(message)

        return result

    def run_cmd(self, cmd, verbose=True, timeout=None,
                expected_codes=None, check_exit_code=True,
                verbose_info=False, reconnect=False):
        """Direct or through-bastion shell command execution"""
        LOG.info(f"Use user `{self.__cluster.ssh_user}` "
                 f"for remote command run")

        use_sudo = False
        if self.provider == 'vsphere':
            use_sudo = True

        return self._run_cmd(cmd, verbose, timeout, expected_codes,
                             check_exit_code, verbose_info,
                             ssh_key=self.__cluster.private_key,
                             ssh_login=self.__cluster.ssh_user,
                             sudo=use_sudo, reconnect=reconnect)

    def exec_pod_cmd(self, cmd, verbose=True, get_events=True, timeout=600):
        """Execute cmd in a priveleged pod on the Machine

        Doesn't require SSH access to the Machine.

        * The {cmd} is executed in the priveleged container from
          the chroot to the node's root filesystem /

        * A separated pod is started and removed for each {cmd}, so
          it will spend some time to pull the image and start the pod.

        * There is no way to interact with {cmd} from shell while it
          is executed. All required parameters must be set in {cmd}

        * There is no way to separate stdout from stderr, all the
          info is collected from the pod "logs".

        :rtype dict: {'logs': <str>, 'events': <str>, 'exit_code': <int>}
        """
        node_name = self.get_k8s_node_name()
        pod_name = "exec-pod-cmd-{}".format(utils.gen_random_string(6))
        mcp_docker_registry = self._cluster.determine_mcp_docker_registry()
        pod_namespace = 'default'
        render_options = {
            'POD_NAME': pod_name,
            'POD_CMD': cmd,
            'NODE_NAME': node_name,
            'MCP_DOCKER_REGISTRY': mcp_docker_registry,
        }
        pod_template = yaml.safe_load(template_utils.render_template(
            settings.MACHINE_PRIVELEGED_POD_YAML, options=render_options,
            log_env_vars=False, log_template=False))

        pod = self._cluster.k8sclient.pods.create(name=pod_name,
                                                  namespace=pod_namespace,
                                                  body=pod_template,
                                                  log_body=False)
        if verbose:
            _LOG = LOG.info
            _ERRLOG = LOG.error
        else:
            _LOG = LOG.debug
            _ERRLOG = LOG.debug
        try:
            pod.wait_phase(['Succeeded', 'Failed'],
                           timeout=timeout, interval=1)
            pod_status = pod.read().status
            container_state = pod_status.container_statuses[0].state
            if container_state.terminated is not None:
                exit_code = container_state.terminated.exit_code
            else:
                exit_code = -1
                LOG.debug(f"Pod {pod.name} has wrong container status:\n"
                          f"{pod_status}")

            logs = pod.get_logs()
            if get_events:
                events = self._manager.get_events_by_uid_str(
                    pod.namespace, pod.uid,
                    event_prefix=pod.name, sort=False)
            else:
                events = None

            _LOG(f"[{self._cluster.name}][{node_name}] Output of the command "
                 f"\"{cmd}\":\n{logs}")
            _LOG(f"Events after command execution:\n{events}")
            _LOG(f"Exit code: {exit_code}")
            return {'logs': logs,
                    'events': events,
                    'exit_code': exit_code}
        except Exception as e:
            _ERRLOG(f"Error while running pod {pod_name}:\n{e}")
            raise (e)
        finally:
            pod.delete()

    def are_conditions_ready(self, expected_fails=None, verbose=False):
        """
        Check if all Machine conditions in ready status
        :param expected_fails: None or dict like the following
            {<condition.type>: <expected message pattern>, ...}
            If expected message pattern is empty, then the whole
            condition is ignored. If it is not empty - then
            this pattern is matched to the message from the cluster
            to check if the message is expected or not.
        :rtype bool: bool
        """
        data = self.data
        result = self._cluster.get_conditions_status(
            data, expected_fails, verbose)
        if result['not_ready']:
            return False
        else:
            return True


class OpenstackProviderMachine(Machine):
    """Openstack provider specific methods"""
    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.OPENSTACK_PROVIDER_NAME

    @property
    def internal_ip(self):
        if self.__internal_ip is not None:
            return self.__internal_ip
        annotations = self.metadata['annotations']
        if 'openstack-ip-address' in annotations:
            self.__internal_ip = annotations['openstack-ip-address']
            return self.__internal_ip
        if self.data and self.data['status'] and \
                self.data['status']['providerStatus'] and \
                self.data['status']['providerStatus']['privateIp']:
            provider_status = self.data['status']['providerStatus']
            self.__internal_ip = provider_status.get("privateIp")
            return self.__internal_ip
        else:
            return None

    @property
    def public_ip(self):
        # if self.machine_type != 'control':
        #     return None
        if self.__public_ip is not None:
            return self.__public_ip
        annotations = self.metadata['annotations']
        if 'openstack-floating-ip-address' in annotations:
            self.__public_ip = annotations['openstack-floating-ip-address']
            return self.__public_ip
        else:
            return None

    def power_off(self):
        openstack_instance_name = self.get_k8s_node_name()
        client = self._cluster.provider_resources.client
        server = client.get_server_by_name(server_name=openstack_instance_name)
        if server.status == 'ACTIVE':
            LOG.info(f"Power off {openstack_instance_name}")
            server.stop()
            client.wait_server_vm_state(server, state='shut down',
                                        timeout=180, interval=20)
            LOG.info(f"Machine {openstack_instance_name} powered off")
        else:
            raise Exception(f"Machine with name {openstack_instance_name} "
                            f"in status {server.status}, expected ACTIVE")

    def power_on(self):
        openstack_instance_name = self.get_k8s_node_name()
        client = self._cluster.provider_resources.client
        server = client.get_server_by_name(server_name=openstack_instance_name)
        if server.status == 'SHUTOFF':
            LOG.info(f"Power on {openstack_instance_name}")
            server.start()
            client.wait_server_vm_state(server, state='running',
                                        timeout=180, interval=20)
            LOG.info(f"Machine with name {openstack_instance_name} "
                     f"powered on")
        else:
            raise Exception(f"Machine with name {openstack_instance_name} "
                            f" in status {server.status}, expected SHUTOFF")


class BaremetalProviderMachine(Machine):
    """Baremetal provider specific methods"""
    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.BAREMETAL_PROVIDER_NAME

    @property
    def internal_ip(self):
        if self.__internal_ip is not None:
            return self.__internal_ip
        # PRODX-14564 providerStatus.privateIp will have the address
        self.__internal_ip = self.data.get('status', {})\
            .get('providerStatus', {}).get('privateIp', None)
        LOG.debug(f"Machine internal ip is {self.__internal_ip}")

        if not self.__internal_ip:
            # fallback to lcmmachine info
            lcmm_status = \
                [x for x in self._cluster.get_lcmmachines()
                 if x.name == self.name][0].data.get('status', {})
            LOG.debug(f"lcmmachine {self.name} status: {lcmm_status}")
            ip = [x.get('address', None)
                  for x in lcmm_status.get('addresses', [])
                  if x.get('type', '') == 'InternalIP']
            if ip:
                LOG.info(f"Machine internal ip is "
                         f"taken from lcmmachine: {ip[0]}")
                self.__internal_ip = ip[0]
        return self.__internal_ip

    @property
    def public_ip(self):
        if self.__public_ip is not None:
            return self.__public_ip
        # PRODX-15440 we don't have PublicIP in addresses array
        self.__public_ip = self.internal_ip
        return self.__public_ip


class AwsProviderMachine(Machine):
    """AWS provider specific methods"""

    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.AWS_PROVIDER_NAME

    def get_k8s_node_name(self):
        if self.data.get('status', {}).get(
                'providerStatus', {}).get(
                   'providerInstanceState', {}).get('id', None):
            provider_status = self.data['status']['providerStatus']
            _id = provider_status['providerInstanceState']['id']
            nodes = [x for x in
                     self._cluster.k8sclient.nodes.list_raw().items
                     if _id in x.spec.provider_id]
            if nodes:
                return nodes[0].metadata.name

    @property
    def internal_ip(self):
        if self.__internal_ip is not None:
            return self.__internal_ip
        if self.data and self.data['status'] and \
                self.data['status']['providerStatus'] and \
                self.data['status']['providerStatus']['privateIp']:
            provider_status = self.data['status']['providerStatus']
            self.__internal_ip = provider_status.get("privateIp")
            return self.__internal_ip
        else:
            return None

    @property
    def public_ip(self):
        LOG.debug("public_ip() is not implemented for AWS provider")
        return None

    def power_off(self):
        aws_instance_name = \
            self.data['metadata']['annotations']['kaas.mirantis.com/uid']
        client = self._cluster.provider_resources.client
        power_state = client.get_instance_state_by_name(
            instance_name=aws_instance_name)
        if power_state == 'running':
            client.instance_power_action(instance_name=aws_instance_name,
                                         action='stop')
            waiters.wait(lambda: client.get_instance_state_by_name(
                instance_name=aws_instance_name) == 'stopped', timeout=300,
                         interval=30)
            power_state = client.get_instance_state_by_name(
                aws_instance_name)
            LOG.info(f"state after power_off instance: {power_state}\n")
        else:
            raise Exception(f"Machine with name {aws_instance_name} "
                            f"in status {power_state}, expected <running>")

    def power_on(self):
        aws_instance_name = \
            self.data['metadata']['annotations']['kaas.mirantis.com/uid']
        client = self._cluster.provider_resources.client
        power_state = client.get_instance_state_by_name(aws_instance_name)
        if power_state == 'stopped':
            client.instance_power_action(aws_instance_name, action='start')
            waiters.wait(lambda: client.get_instance_state_by_name(
                instance_name=aws_instance_name) == 'running', timeout=300,
                         interval=30)
            power_state = client.get_instance_state_by_name(
                aws_instance_name)
            LOG.info(f"state after power_on instance: {power_state}\n")
        else:
            raise Exception(f"Machine with name {aws_instance_name} "
                            f"in status {power_state}, expected <stopped>")


class ByoProviderMachine(Machine):
    """BYO provider specific methods"""

    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.BYO_PROVIDER_NAME

    @property
    def internal_ip(self):
        LOG.debug("internal_ip() is not implemented for BYO provider")
        return None

    @property
    def public_ip(self):
        LOG.debug("public_ip() is not implemented for BYO provider")
        return None


class VsphereProviderMachine(Machine):
    """VSphere provider specific methods"""
    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.VSPHERE_PROVIDER_NAME

    @property
    def internal_ip(self):
        if self.__internal_ip is not None:
            return self.__internal_ip
        if self.data and self.data['status'] and \
                self.data['status']['providerStatus'] and \
                self.data['status']['providerStatus']['privateIp']:
            provider_status = self.data['status']['providerStatus']
            self.__internal_ip = provider_status.get("privateIp")
            return self.__internal_ip
        else:
            return None

    @property
    def public_ip(self):
        LOG.debug("public_ip() is not implemented for VSphere provider")
        return None


class EquinixMetalProviderMachine(Machine):
    """EquinixMetal provider specific methods"""
    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.EQUINIX_PROVIDER_NAME

    @property
    def internal_ip(self):
        # Using private IP addresses in Equinix starting from 5.16
        # Checking internal node IP to avoid wrong IP address detection
        # after fixing PRODX-14498 (node IP address is still public after the
        # upgrade from 5.15 to 5.16 while in pure 5.16 deployment it is
        # private)
        # TODO: remove getting node IP after 2.9 release
        if self.metadata['annotations']:
            nodes = self._cluster.k8sclient.nodes.list_raw().to_dict()['items']
            node_name = 'kaas-node-' + self.metadata['annotations'][
                'kaas.mirantis.com/uid']
            for node in nodes:
                if node['metadata']['name'] == node_name:
                    return [ip['address'] for ip in node['status']
                            ['addresses'] if 'InternalIP' in ip['type']][0]

        if self.__internal_ip is not None:
            return self.__internal_ip
        if self.data and self.data['status'] and \
                self.data['status']['providerStatus'] and \
                self.data['status']['providerStatus']['privateIp']:
            provider_status = self.data['status']['providerStatus']
            self.__internal_ip = provider_status.get("privateIp")
            return self.__internal_ip
        else:
            return None

    @property
    def public_ip(self):
        if self.__public_ip is not None:
            return self.__public_ip
        if self.data and self.data['status'] and \
                self.data['status']['providerStatus'] and \
                self.data['status']['providerStatus']['publicIp']:
            provider_status = self.data['status']['providerStatus']
            self.__public_ip = provider_status.get("publicIp")
            return self.__public_ip
        else:
            return None


class AzureProviderMachine(Machine):
    """Azure provider specific methods"""
    __internal_ip = None
    __public_ip = None

    @property
    def provider(self):
        return settings.AZURE_PROVIDER_NAME

    @property
    def internal_ip(self):
        if self.__internal_ip is not None:
            return self.__internal_ip

        try:
            self.__internal_ip = self.data['status'][
                    'providerStatus']['privateIp']
        except KeyError as e:
            LOG.warning(f"failed to get privateIp from providerStatus: {e}")
            return None

        return self.__internal_ip

    @property
    def public_ip(self):
        LOG.debug("public_ip() is not implemented for Azure provider")
        return None
