import pytest
from packaging import version

from si_tests import settings
from si_tests import logger
from si_tests.utils import templates
from si_tests.utils import utils
from si_tests.utils import waiters

LOG = logger.logger


@pytest.mark.usefixtures('log_method_time')
@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.KAAS_CHILD_CLUSTER_NAME)])
@pytest.mark.usefixtures("store_cluster_description")
def test_kaas_create_child(kaas_manager, check_sg, proxy_manager, _):
    """Deploy child cluster on top of management cluster."""

    # Collecting env data
    cluster_name = settings.KAAS_CHILD_CLUSTER_NAME
    namespace_name = settings.KAAS_NAMESPACE
    release_name = settings.KAAS_CHILD_CLUSTER_RELEASE_NAME
    external_network = settings.KAAS_EXTERNAL_NETWORK_ID
    LOG.info("Available releases: {}".format(
        kaas_manager.get_clusterrelease_names()))
    assert release_name in kaas_manager.get_clusterrelease_names()
    child_dns = settings.KAAS_CHILD_CLUSTER_DNS
    machine_image_name = settings.KAAS_CHILD_CLUSTER_MACHINE_IMAGE_NAME
    machine_flavor_name = settings.KAAS_CHILD_CLUSTER_MACHINE_FLAVOR_NAME
    machine_az_name = settings.KAAS_CHILD_CLUSTER_AZ_NAME
    slave_nodes_count = int(settings.KAAS_CHILD_CLUSTER_SLAVE_NODES_COUNT)
    master_nodes_count = int(settings.KAAS_CHILD_CLUSTER_MASTER_NODES_COUNT)
    rnd_string = utils.gen_random_string(6)
    # print(kaas_manager.api.kaas_clusterreleases.list_all())

    # Creating namespace
    LOG.info("Namespace name - %s", namespace_name)
    ns = kaas_manager.create_namespace(namespace_name)

    # Creating public key
    public_key_name = "{cluster_name}-{rnd_string}-key".format(
        cluster_name=cluster_name, rnd_string=rnd_string)
    LOG.info("Public key name - {public_key_name}".format(
        public_key_name=public_key_name))
    with open(settings.KAAS_CHILD_CLUSTER_PUBLIC_KEY_FILE) as pub_key:
        pub_key_content = pub_key.read()
    ns.create_publickey(public_key_name, pub_key_content)

    # Creating openstack credential
    credential_name = "{cluster_name}-{rnd_string}-cred".format(
        cluster_name=cluster_name, rnd_string=rnd_string)
    LOG.info("Credential name - {credential_name}".format(
        credential_name=credential_name))
    clouds_yaml_str = templates.render_template(settings.CLOUDS_YAML_PATH)
    ns.create_openstack_credential(credential_name, clouds_yaml_str)
    ns.wait_openstack_resources_available(credential_name)

    helm_releases = []
    for hrelease in settings.KAAS_CHILD_CLUSTER_EXTRA_HELM_RELEASES:
        helm_releases.append({"name": hrelease})

    if ((settings.KAAS_EXTERNAL_PROXY_ACCESS_STR or
         settings.KAAS_INSTALL_FAKE_PROXY) and
            not settings.KAAS_OFFLINE_DEPLOYMENT):
        LOG.error("Proxy variables is set but KAAS_OFFLINE_DEPLOYMENT "
                  "is False.Using Proxy on online deployments makes no sense.")
        raise RuntimeError("Proxy variables is set but KAAS_OFFLINE_DEPLOYMENT"
                           " is False")

    if settings.KAAS_OFFLINE_DEPLOYMENT:
        offline_security_group = [check_sg['name']]
        LOG.info('Offline deployment is enabled. External network will be '
                 'restricted on child cluster nodes. Using '
                 'security group: {0}'.format(
                  [check_sg['name']]))
    else:
        offline_security_group = None

    if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
        LOG.info("Found KAAS_EXTERNAL_PROXY_ACCESS_STR. Creating "
                 "proxyobject with proxy {0}".
                 format(settings.KAAS_EXTERNAL_PROXY_ACCESS_STR))
        proxy_access_str = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
        proxy_object = ns.create_proxyobject(proxy_str=proxy_access_str)
        proxy_object_name = proxy_object.data['metadata']['name']
    elif settings.KAAS_INSTALL_FAKE_PROXY:
        LOG.info("Variable KAAS_EXTERNAL_PROXY_ACCESS_STR is not set but"
                 "KAAS_INSTALL_FAKE_PROXY is enabled. Fake squid proxy will "
                 "be installed on management cluster to use it in child.")
        proxy_manager.deploy_proxy_sts()
        proxy_manager.deploy_proxy_svc()
        proxy_access_str = proxy_manager.generate_proxy_string()
        proxy_manager.check_proxy_conn()
        proxy_object = ns.create_proxyobject(proxy_str=proxy_access_str)
        proxy_object_name = proxy_object.data['metadata']['name']
    else:
        proxy_object_name = None

    cluster = ns.create_cluster(
        cluster_name,
        release_name,
        credential_name,
        external_network_id=external_network,
        provider="openstack",
        services_cidr="10.96.0.0/12",
        pods_cidr="192.168.0.0/16",
        nodes_cidr="10.10.10.0/24",
        dns=child_dns,
        extra_helm_releases=helm_releases,
        lma_enabled=settings.KAAS_CHILD_CLUSTER_DEPLOY_LMA,
        public_key_name=public_key_name,
        proxy_name=proxy_object_name)

    labels = []
    # Add os labels if CR name is not MKE
    os_control_labels = []
    os_computes_labels = []

    release_version = \
        kaas_manager.get_clusterrelease(release_name).data['spec'][
            'version'].replace('-rc', '')

    # Testing SL placement for cluster releases starting from 5.18.0 or
    # 6.18.0 if sl-node-placement-enabled feature flag enabled
    sl_placement_enabled = \
        settings.STACKLIGHT_ENABLE_HA and \
        (version.parse('5.18.0') <= version.parse(release_version) <
         version.parse('6.0.0') or version.parse('6.18.0') <=
         version.parse(release_version) < version.parse('7.0.0') or
         version.parse(release_version) >= version.parse('7.1.0'))

    if slave_nodes_count < 3 and sl_placement_enabled:
        raise Exception("Not enough worker nodes for LMA HA configuration")

    if 'mke' not in release_name:
        # for MOS child we will create 3 os control nodes
        # (default value of KAAS_CHILD_CLUSTER_SLAVE_NODES_COUNT)
        # and same number of compute nodes
        os_control_labels_names = [
            {'key': 'openstack-control-plane', 'value': 'enabled'},
            {'key': 'openstack-gateway', 'value': 'enabled'},
            {'key': 'openvswitch', 'value': 'enabled'}]
        for name in os_control_labels_names:
            os_control_labels.append(name)

        os_computes_labels_names = [
            {'key': 'openstack-compute-node', 'value': 'enabled'},
            {'key': 'openvswitch', 'value': 'enabled'}]
        for name in os_computes_labels_names:
            os_computes_labels.append(name)
        for node in range(slave_nodes_count):
            cluster.create_machine(
                node_type="node",
                node_flavor=machine_flavor_name,
                node_image=machine_image_name,
                node_az=machine_az_name,
                node_labels=os_control_labels,
                node_sg=offline_security_group,
            )
            cluster.create_machine(
                node_type="node",
                node_flavor=machine_flavor_name,
                node_image=machine_image_name,
                node_az=machine_az_name,
                node_labels=os_computes_labels,
                node_sg=offline_security_group,
            )
    else:
        labels = \
            [{'key': k, 'value': v} for k, v in
             cluster.get_allowed_node_labels().items()]

        # we define 2 SL nodes for both ha and non-ha cases
        # for non-ha nothing else is needed
        # however for ha 1 addtional node is required
        stacklight_nodes_count = 2

        for i in range(stacklight_nodes_count):
            cluster.create_machine(
                node_type="node",
                node_flavor=machine_flavor_name,
                node_image=machine_image_name,
                node_az=machine_az_name,
                node_sg=offline_security_group,
                node_labels=labels
            )
        slave_nodes_count -= stacklight_nodes_count

        for _ in range(slave_nodes_count):
            cluster.create_machine(
                node_type="node",
                node_flavor=machine_flavor_name,
                node_image=machine_image_name,
                node_az=machine_az_name)

    for node in range(master_nodes_count):
        cluster.create_machine(
            node_type="master",
            node_flavor=machine_flavor_name,
            node_image=machine_image_name,
            node_az=machine_az_name,
            node_sg=offline_security_group)

    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()

    # Check SL was not deployed if there is not enough SL labels
    if sl_placement_enabled:
        def _sl_release_exists():
            helmbundle = cluster.get_helmbundle().data
            for chart in helmbundle['spec']['releases']:
                if chart['name'] == 'stacklight':
                    LOG.info("Stacklight release is present in helmbundle")
                    return True
            return False

        LOG.info("Checking that Stacklight was not deployed")
        assert not _sl_release_exists(), \
            'StackLight HA was deployed on cluster with not ' \
            'enough SL labels on worker machines'

        LOG.info("Adding one more worker node with SL label")
        cluster.create_machine(
            node_type="node",
            node_flavor=machine_flavor_name,
            node_image=machine_image_name,
            node_az=machine_az_name,
            node_sg=offline_security_group,
            node_labels=labels)

        stacklight_nodes_count += 1

        cluster.check.check_machines_status()
        cluster.check.check_cluster_nodes()

        LOG.info("Waiting for helm releases (including Stacklight) to be "
                 "deployed")
        cluster.check.check_helmbundles(timeout=600)

    if labels:
        machines = [
            x for x in cluster.get_machines(machine_type="worker")
            if 'nodeLabels' in x.spec['providerSpec']['value'].keys() and
               labels == x.spec['providerSpec']['value']['nodeLabels']
        ]
        assert len(machines) == stacklight_nodes_count, \
            "Not all machines have expected nodeLabels"

        for machine in machines:
            waiters.wait(
                lambda: machine.check_k8s_nodes_has_labels(),
                timeout=600, interval=60)

    # Collecting artifacts
    cluster.store_k8s_artifacts()
    cluster.check.check_k8s_nodes()
    if 'mos' in release_name and cluster.openstack_lma_enabled():
        expected_fails = cluster.check.EXCLUDED_PODS
        expected_condition_fails = cluster.check.EXCLUDED_CONDITIONS
        cluster.check.check_cluster_readiness(
            exp_provider_status=False,
            expected_fails=expected_fails,
            expected_condition_fails=expected_condition_fails)
    else:
        cluster.check.check_cluster_readiness()

    cluster.check.check_helmbundles(timeout=900)

    if settings.RUN_POD_CHECKS:
        # Check/wait for correct docker service replicas in cluster
        cluster.check.check_actual_expected_docker_services()
        cluster.check.check_k8s_pods()
        cluster.check.check_actual_expected_pods(timeout=3200,
                                                 check_all_nss=True)
    cluster.provider_resources.save_artifact()

    if settings.KAAS_OFFLINE_DEPLOYMENT:
        cluster.check.check_offline_isolation(
            settings.KAAS_EXTERNAL_PROXY_ACCESS_STR)
