#    Copyright 2019 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from si_tests import settings, logger

from si_tests.managers import bootstrap_manager
from si_tests.managers import bootstrap_packet

LOG = logger.logger


class TestProvisionBmCluster():
    def test_bootstrap_kaas_bm_phase1(self, openstack_client, show_step):
        """Bootstrap KaaS_BM on OpenStack VM

        Scenario:
            1. Erase previous KaaS environment and BM node, if exists
            2. Create a stack with a node0
            3. Prepare a node0 and upload git repos
            4. Check bootstrap templates for management cluster
            5. Prepare for kaas-bm install, using ansible playbooks
        """

        bootstrap = bootstrap_manager.BootstrapManager()

        # step 001
        if settings.KEEP_ENV_BEFORE:
            LOG.info('Skip erase due KEEP_ENV_BEFORE flag is set')
        else:
            if openstack_client._current_stack:
                # Erase the stack with the seed node
                show_step(1)
                bootstrap.step_001_erase_env_before(openstack_client)
            else:
                LOG.warning("Can't erase environment due "
                            "seed node stack is absent")

        # Create a stack with a seed node
        show_step(2)
        bootstrap.step_002_provision_seed_vm(openstack_client)

        show_step(3)
        # INFO(alexz): since we change network configuration
        # on seed-node => retrigger docker network,
        # we cant move those part to phase2, aka
        # run test on seed node.
        bootstrap.step_003_prepare_bm_seed_node(openstack_client)

        # step 004
        if settings.SKIP_BOOTSTRAP_TEMPLATES_CHECK:
            LOG.warning("Skip bootstrap templates check due "
                        "SKIP_BOOTSTRAP_TEMPLATES_CHECK flag is set")
        else:
            show_step(4)
            bootstrap.check_templates(openstack_client)

        show_step(5)
        bootstrap.step_004_prepare_for_kaas_bm_cluster(openstack_client)

    def test_bootstrap_kaas_bm_phase2(self, openstack_client, show_step):
        """Bootstrap KaaS_BM on OpenStack VM

        Scenario:
            6. Install kaas-bm mgmt cluster using ansible playbooks
            7. Create default MCC users, if requested
            8. Wait for all pods to be Running and Ready
            9. Check all expected pods exists
            10. Download bootstrap artifacts
            11. Check LDAP integration if enabled
        """

        bootstrap = bootstrap_manager.BootstrapManager()
        show_step(6)
        bootstrap.step_004_deploy_kaas_bm_mgmt_cluster(openstack_client)

        if settings.CREATE_DEFAULT_MCC_USERS_VIA_CLI:
            show_step(7)
            bootstrap.step_004_create_default_mcc_users(
                openstack_client,
                custom_bootstrap_path=settings.KAAS_BOOTSTRAP_TARGET_DIR,
                baremetal=True)

        show_step(8)
        bootstrap.step_006_wait_for_pods()

        show_step(9)
        if settings.RUN_POD_CHECKS:
            bootstrap.step_006_postdeployment_checks(openstack_client,
                                                     check_svc=True)

        show_step(10)
        bootstrap.step_007_download_bootstrap_artifacts(openstack_client)

        if settings.KEYCLOAK_LDAP_ENABLED:
            show_step(11)
            bootstrap.check_keystone_ldap_integration()

    def test_bootstrap_kaas_bm_equinix_phase1(self, show_step):
        """Bootstrap seed node on equinix HS
        Scenario:
            1. Erase previous seed node by hostname,if exists.
               Create a equinix node like seed
        """
        bootstrapEq = bootstrap_packet.BootstrapManagerPacket()
        show_step(1)
        bootstrapEq.step_002_provision_seed_vm()

    def test_bootstrap_kaas_bm_equinix_phase2(self, openstack_client,
                                              show_step):
        """Bootstrap KaaS_BM on OpenStack VM

        Scenario:
            6. Install kaas-bm mgmt cluster using ansible playbooks
            7. Create default MCC users, if requested
            8. Wait for all pods to be Running and Ready
            9. Check all expected pods exists
            10. Download bootstrap artifacts
            11. Check LDAP integration if enabled
        """
        bootstrap = bootstrap_manager.BootstrapManager()
        show_step(3)
        bootstrap.step_003_prepare_bm_seed_node(openstack_client)

        # step 004
        if settings.SKIP_BOOTSTRAP_TEMPLATES_CHECK:
            LOG.warning("Skip bootstrap templates check due "
                        "SKIP_BOOTSTRAP_TEMPLATES_CHECK flag is set")
        else:
            show_step(4)
            bootstrap.check_templates(openstack_client)

        show_step(5)
        bootstrap.step_004_prepare_for_kaas_bm_cluster(openstack_client)

        show_step(6)
        bootstrap.step_004_deploy_kaas_bm_mgmt_cluster(openstack_client)

        if settings.CREATE_DEFAULT_MCC_USERS_VIA_CLI:
            show_step(7)
            bootstrap.step_004_create_default_mcc_users(
                openstack_client,
                custom_bootstrap_path=settings.KAAS_BOOTSTRAP_TARGET_DIR,
                baremetal=True)

        show_step(8)
        bootstrap.step_006_wait_for_pods()

        show_step(9)
        if settings.RUN_POD_CHECKS:
            bootstrap.step_006_postdeployment_checks(openstack_client,
                                                     check_svc=True)

        show_step(10)
        bootstrap.step_007_download_bootstrap_artifacts(openstack_client)

        if settings.KEYCLOAK_LDAP_ENABLED:
            show_step(11)
            bootstrap.check_keystone_ldap_integration()

    def test_bootstrap_kaas_bm_physical(self, openstack_client, show_step):
        """Bootstrap KaaS_BM on OpenStack VM

        Scenario:
            1. Erase previous KaaS environment and BM node, if exists
            2. Create a stack with a node0
            3. Prepare a node0 and upload git repos
            4. Check templates
            5. Install kaas-bm cluster using ansible playbooks
            6. Create default MCC users, if requested
            7. Wait for all pods to be Running and Ready
            8. Check that all expected pods exists
            9. Download bootstrap artifacts
            10. Check LDAP integration if enabled
        """

        bootstrap = bootstrap_manager.BootstrapManager()

        # step 001
        if settings.KEEP_ENV_BEFORE:
            LOG.info('Skip erase due KEEP_ENV_BEFORE flag is set')
        else:
            if openstack_client._current_stack:
                # Erase the stack with the seed node
                show_step(1)
                bootstrap.step_001_erase_env_before(openstack_client)
            else:
                LOG.warning("Can't erase environment due "
                            "seed node stack is absent")

        # Create a stack with a seed node
        show_step(2)
        bootstrap.step_002_provision_seed_vm(openstack_client)

        show_step(3)
        bootstrap.step_003_prepare_bm_seed_node(openstack_client)

        # step 004
        if settings.SKIP_BOOTSTRAP_TEMPLATES_CHECK:
            LOG.warning("Skip bootstrap templates check due "
                        "SKIP_BOOTSTRAP_TEMPLATES_CHECK flag is set")
        else:
            show_step(4)
            bootstrap.check_templates(openstack_client)

        show_step(5)
        bootstrap.step_004_deploy_kaas_bm_cluster_physical(
            openstack_client, settings.ENV_CONFIG_NAME)

        if settings.CREATE_DEFAULT_MCC_USERS_VIA_CLI:
            show_step(6)
            bootstrap.step_004_create_default_mcc_users(
                openstack_client,
                custom_bootstrap_path=settings.KAAS_BOOTSTRAP_TARGET_DIR,
                baremetal=True)

        show_step(7)
        bootstrap.step_006_wait_for_pods()

        show_step(8)
        if settings.RUN_POD_CHECKS:
            bootstrap.step_006_postdeployment_checks(openstack_client,
                                                     check_svc=True)

        show_step(9)
        bootstrap.step_007_download_bootstrap_artifacts(openstack_client)

        if settings.KEYCLOAK_LDAP_ENABLED:
            show_step(10)
            bootstrap.check_keystone_ldap_integration()
