import os.path
import re
# import pytest

from si_tests import settings
from si_tests import logger
from si_tests.utils import utils

from si_tests.managers import bootstrap_manager


LOG = logger.logger


def test_prepare_vsphere_vm_template(openstack_client, show_step):
    """Bootstrap seed node and build VM template with selected OS.

    Scenario:
        1. Erase previous KaaS environment and seed node, if exists
        2. Create a seed VM using heat template
        3. Prepare seed node
        4. Check templates
        5. Download iso to seed node
        6. Build VM template
        7. Save artifacts
        8. Remove Heat stack
    """
    bootstrap = bootstrap_manager.BootstrapManager()

    # step 001
    if settings.KEEP_ENV_BEFORE:
        LOG.info("Skip erase due KEEP_ENV_BEFORE flag is setted")
    else:
        if openstack_client._current_stack:
            # Erase the stack with the seed node
            show_step(1)
            bootstrap.step_001_erase_env_before(openstack_client)
        else:
            LOG.warning("Can't erase environment due "
                        "seed node stack is absent")

    # Create a stack with a seed node
    show_step(2)
    bootstrap.step_002_provision_seed_vm(openstack_client)

    # step 003
    show_step(3)
    bootstrap.step_003_prepare_seed_node(openstack_client)

    # step 004
    if settings.SKIP_BOOTSTRAP_TEMPLATES_CHECK:
        LOG.warning("Skip bootstrap templates check due "
                    "SKIP_BOOTSTRAP_TEMPLATES_CHECK flag is set")
    else:
        show_step(4)
        bootstrap.check_templates(openstack_client)

    remote = bootstrap.remote_seed(openstack_client)
    show_step(5)
    iso_url = ''
    if settings.KAAS_VSPHERE_TEMPLATE_OS_NAME == 'rhel':
        iso_url = settings.KAAS_VSPHERE_RHEL_ISO_URL
    elif settings.KAAS_VSPHERE_TEMPLATE_OS_NAME == 'centos':
        iso_url = settings.KAAS_VSPHERE_CENTOS_ISO_URL
    iso_file_name = os.path.basename(iso_url)
    remote.execute(
        f"curl -L {iso_url} -o /tmp/{iso_file_name}",
        verbose=True)

    show_step(6)
    run_envs = {
        "VSPHERE_PACKER_ISO_FILE": f"/tmp/{iso_file_name}",
        "VSPHERE_CLUSTER_NAME": "vCluster-BMTeam",
        "VSPHERE_PACKER_IMAGE_OS_NAME": settings.KAAS_VSPHERE_TEMPLATE_OS_NAME,
        "VSPHERE_PACKER_IMAGE_OS_VERSION":
            settings.KAAS_VSPHERE_TEMPLATE_OS_VERSION,
        "CLUSTER_NAME": settings.CLUSTER_NAME,
        "TEMPLATES_DIR":
            f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/templates/vsphere"
    }
    envs_string = utils.make_export_env_strting(run_envs)
    envs_string = envs_string + \
        f"; export VSPHERE_RO_USER={settings.KAAS_VSPHERE_RO_USER}" + \
        f"; export VSPHERE_RO_PASSWORD={settings.KAAS_VSPHERE_RO_PASSWORD}"
    res = remote.execute(
        "{envs_string}; bash -x ./{bootstrap}/bootstrap.sh "
        "vsphere_template 2>&1 | tee -a image_build_log".format(
            envs_string=envs_string,
            bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
        verbose=True)

    assert res.exit_code == 0, \
        "Building a new vSphere template has been failed"
    show_step(7)
    template_name = None
    pattern = re.compile(r".*Full template path: (.*?)\. Please update.*")

    for line in res.stdout_str.splitlines():
        if "Full template path:" in line:
            LOG.info("Found target log line - %s", line)
            match = pattern.match(line)
            assert match is not None, \
                "Can't find template name in the bootstrap.sh output"
            template_name = match.group(1)
            break

    assert template_name is not None, \
        "Can't find template name in the bootstrap.sh output"
    artifacts_dir = settings.ARTIFACTS_DIR
    template_name_filename = os.path.join(artifacts_dir, "new_template_name")
    with open(template_name_filename, "w") as f:
        f.write(template_name)

    show_step(8)
    stack_name = settings.ENV_NAME
    LOG.info("Remove heat stack {0}".format(stack_name))
    openstack_client.delete_stack()
