import pytest

import time
from si_tests import settings
from si_tests import logger
from si_tests.clients.iam import keycloak_client
from si_tests.utils import utils

LOG = logger.logger


def test_add_iamrolebinding_user_failed(kaas_manager):
    """
    1. Apply IAMRoleBinding with incorrect username.
    'iamusers.iam.mirantis.com %s not found' should be in the output.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    user_name = "{rnd_string}-user".format(rnd_string=rnd_string)
    LOG.info("IAMRoleBinding name - {iamrolebinding_name}, "
             "username - {username}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 username=user_name)
             )
    try:
        ns.create_iamrolebinding(iamrolebinding_name, "writer", user_name)
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong username")


def test_add_iamrolebinding_role_failed(kaas_manager):
    """
    2. Apply IAMRoleBinding with incorrect rolename.
    'iamroles.iam.mirantis.com %s not found' should be in the output.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    role_name = "{rnd_string}-role".format(rnd_string=rnd_string)
    LOG.info("IAMRoleBinding name - {iamrolebinding_name}, "
             "rolename - {rolename}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 rolename=role_name)
             )
    try:
        ns.create_iamrolebinding(iamrolebinding_name, role_name, "reader")
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong rolename")


def test_add_iamclusterrolebinding_cluster_failed(kaas_manager):
    """
    3. Apply IAMClusterRoleBinding with incorrect clustername.
    'clusters.cluster.k8s.io %s not found' should be in the output.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)
    cluster_name = "{rnd_string}-cluster".format(rnd_string=rnd_string)
    LOG.info("IAMClusterRoleBinding name - {iamrolebinding_name}, "
             "clustername - {clustername}".format(
                 iamrolebinding_name=iamrolebinding_name,
                 clustername=cluster_name)
             )
    try:
        ns.create_iamclusterrolebinding(iamrolebinding_name,
                                        "writer",
                                        "reader",
                                        cluster_name)
    except Exception as e:
        assert e.status == 400, ('Wrong response code. {}'.format(e))
    else:
        pytest.fail("It is possible to create iamrolebinding"
                    "with wrong cluster name")


def test_add_iamrolebinding_success(kaas_manager):
    """
    4. Apply IAMRoleBinding with correct parameters.
    5. Get IAMRoleBinding and compare it with given.
    6. Request roles list from KeyCloak to check role presence.
    7. Remove IAMRoleBinding from k8s API.
    8. Request roles list from KeyCloak to check role absence.
    """
    rnd_string = utils.gen_random_string(6)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)

    iamrolebinding_name = "{rnd_string}-iamrb".format(rnd_string=rnd_string)

    ns.create_iamrolebinding(iamrolebinding_name, "writer", "reader")

    keycloak_ip = kaas_manager.get_keycloak_ip()
    pwd = kaas_manager.get_secret_data('iam-api-secrets',
                                       'kaas', 'keycloak_password')
    kclient = keycloak_client.KeycloakAdminClient(ip=keycloak_ip,
                                                  user=settings.KEYCLOAK_USER,
                                                  password=pwd,
                                                  realm='iam')
    uid = kclient.client.get_user_id('reader')
    roles = kclient.client.get_realm_roles_of_user(uid)
    role_found = False
    target_role = 'm:kaas:{ns}@writer'.format(ns=settings.TARGET_NAMESPACE)
    for role in roles:
        if role['name'] == target_role:
            role_found = True
    if not role_found:
        pytest.fail("Role wasn't assigned to user")

    iamrb = ns.get_iamrolebinding(iamrolebinding_name)
    try:
        iamrb.delete()
        # Need to wait until controller remove role from user
        time.sleep(5)
    except Exception as e:
        LOG.info(e)
    roles = kclient.client.get_realm_roles_of_user(uid)
    LOG.info(roles)
    role_found = False
    target_role = 'm:kaas:{ns}@writer'.format(ns=settings.TARGET_NAMESPACE)
    for role in roles:
        if role['name'] == target_role:
            role_found = True
    if role_found:
        pytest.fail("Role wasn't removed from user")
