import pytest
from si_tests import settings

skip_kill_pods = [
    # we don't need to check this pod
    'image-precaching',
    # PRODX-3376
    'libvirt-libvirt-default',
    # PRODX-4218
    'csi-rbdplugin',
    # PRODX-4220
    'rook-ceph-osd-0', 'rook-ceph-osd-1', 'rook-ceph-osd-2',
    # PRODX-6277
    'octavia-health-manager-default',
    # PRODX-3376
    'nova-compute-default'
]


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
def test_ha_kill_proc_pods(target_cluster, command, pod_group, _):
    """Kill processed in every container in all cluster pods
    Precondition - all expected pods and their replicas must be presented
    The following scenario is executed for every namespace and
    expected pod entry in get_expected_pods
    Note: 'kill -- - 1' will kill all processes except pid 1

    Scenario:
        1. Iterate by each pod in a group
        2. Iterate by each container in a pod and
           kill processes in the container
        3. Wait till number of replicas will be restored
        4. Check pods and container statuses in this group
           (Running and Ready)
        5. Check all pods phases (must be Running or Completed)
           (in fixture)

    Expected result - pods are recreated, number of replicas is restored.
    """
    pod_name, pod_num, pods = pod_group
    if pod_name in skip_kill_pods:
        pytest.skip("Skip {} kill proc check for "
                    "MOSK child cluster".format(pod_name))

    target_cluster.check.check_ha_kill_proc_pods(pod_group, command)


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
@pytest.mark.skip(reason="ENGORC-7836")
def test_ha_kill_containerd_shim(target_cluster, _):
    """Kill all containerd-shim processes on every cluster node
    Precondition - all expected pods and their replicas must be presented
    The following scenario is executed for every node

    Scenario:
        1. SSH to node and run kill command for containerd-shim processes
        2. Wait
        3. Check that all pods are Running and Ready

    Expected result - pods and their containers are Running and Ready.
    Number of replicas are the same.
    """
    target_cluster.check.check_ha_kill_containerd_shim()
    target_cluster.check.check_actual_expected_pods()


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
@pytest.mark.skip(reason="ENGORC-7837")
def test_ha_stop_containerd(target_cluster, _):
    """Stop containerd service on every cluster node (one at a time)
    Precondition - all expected pods and their replicas must be presented
    The following scenario is executed for every node

    Scenario:
        1. SSH to node
        2. Get number of containerd-shim processes
        3. Stop containerd service and make sure it is stopped
        4. Get number of containerd-shim processes and
           compare with previous value (must be the same)
        5. Check that all pods are Running and Ready
        6. Start containerd service

    Expected result - containerd-shim processes are running independently
    from containerd service and all containers in the pods are
    still operational even if containerd is down.
    """
    target_cluster.check.check_ha_stop_containerd()
    target_cluster.check.check_actual_expected_pods()
