import pytest
import yaml

from si_tests.utils import waiters
from si_tests import logger
from si_tests import settings
from si_tests.utils import exceptions as timeout_exceptions
LOG = logger.logger


@pytest.mark.usefixtures('log_method_time')
def test_cleanup_resources(openstack_client):
    """ Test cleanup resources """
    resources_map = {'routers': [], 'servers': [], 'loadbalancers': [],
                     'fips': [], 'networks': [], 'volumes': [],
                     'security_groups': [], 'keypairs': [], 'stacks': []}
    not_deleted_resources = {}
    trigger_cleanup = settings.TRIGGER_RESOURCES_CLEANUP
    trigger_stacks_cleanup = settings.TRIGGER_STACKS_CLEANUP

    def wait_routers_deleted():
        try:
            LOG.info("Waiting for routers are deleted")
            waiters.wait(lambda: not [
                rt['id'] for rt in openstack_client.neutron.list_routers()[
                    'routers'] if rt['id'] in resources_map['routers']],
                         interval=15, timeout=300)
        except timeout_exceptions.TimeoutError:
            not_deleted_routers = [
                rt['id'] for rt in openstack_client.neutron.list_routers()[
                    'routers'] if rt['id'] in resources_map['routers']]
            not_deleted_resources['routers'] = not_deleted_routers
            LOG.warning("Some routers are not deleted: {}".format(
                not_deleted_routers))

    def wait_servers_deleted():
        try:
            LOG.info("Waiting for servers are deleted")
            waiters.wait(lambda: not [
                sv.id for sv in openstack_client.nova.servers.list()
                if sv.id in resources_map['servers']],
                    interval=15, timeout=300)
        except timeout_exceptions.TimeoutError:
            not_deleted_servers = [
                sv.id for sv in openstack_client.nova.servers.list()
                if sv.id in resources_map['servers']]
            not_deleted_resources['servers'] = not_deleted_servers
            LOG.warning("Some servers are not deleted: {}".format(
                not_deleted_servers))

    def wait_loadbalancers_deleted():
        try:
            LOG.info("Waiting for loadbalancers are deleted")
            waiters.wait(lambda: not [
                lbl['id'] for lbl in
                openstack_client.octavia.load_balancer_list()[
                    'loadbalancers'] if lbl['id'] in resources_map[
                    'loadbalancers']], interval=15, timeout=300)
        except timeout_exceptions.TimeoutError:
            not_deleted_lbs = \
                [lbl['id'] for lbl in
                 openstack_client.octavia.load_balancer_list()[
                     'loadbalancers'] if lbl['id'] in resources_map[
                     'loadbalancers']]
            not_deleted_resources['loadbalancers'] = not_deleted_lbs
            LOG.warning("Some loadbalancers are not deleted: {}".format(
                not_deleted_lbs))

    def wait_fips_deleted():
        try:
            LOG.info("Waiting for floating ips are are deleted")
            waiters.wait(lambda: not [
                fip['id'] for fip in
                openstack_client.neutron.list_floatingips()[
                    'floatingips'] if fip['id'] in resources_map[
                    'fips']], interval=10, timeout=60)
        except timeout_exceptions.TimeoutError:
            not_deleted_fips = \
                [fip['id'] for fip in
                 openstack_client.neutron.list_floatingips()[
                     'floatingips'] if fip['id'] in resources_map[
                     'fips']]
            not_deleted_resources['fips'] = not_deleted_fips
            LOG.warning("Some floating ips are not deleted: {}".format(
                not_deleted_fips))

    def wait_networks_deleted():
        try:
            LOG.info("Waiting for networks are are deleted")
            waiters.wait(lambda: not [
                net['id'] for net in
                openstack_client.neutron.list_networks()[
                    'networks'] if net['id'] in resources_map[
                    'networks']], interval=15, timeout=300)
        except timeout_exceptions.TimeoutError:
            not_deleted_nets = \
                [net['id'] for net in
                 openstack_client.neutron.list_networks()[
                     'networks'] if net['id'] in resources_map[
                     'networks']]
            not_deleted_resources['networks'] = not_deleted_nets
            LOG.warning("Some networks are not deleted: {}".format(
                not_deleted_nets))

    def wait_volumes_deleted():
        try:
            LOG.info("Waiting for volumes are deleted")
            waiters.wait(lambda: not [
                vol.id for vol in openstack_client.cinder.volumes.list()
                if vol.id in resources_map['volumes']],
                    interval=30, timeout=300)
        except timeout_exceptions.TimeoutError:
            not_deleted_volumes = \
                [vol.id for vol in openstack_client.cinder.volumes.list()
                 if vol.id in resources_map['volumes']]
            not_deleted_resources['volumes'] = not_deleted_volumes
            LOG.warning("Some volumes are not deleted: {}".format(
                not_deleted_volumes))

    resources_map['routers'] = [
        rt.get('id') for rt in openstack_client.get_routers(
            name_prefix='kaas')]

    resources_map['servers'] = [
        server.get('id') for server in openstack_client.get_servers(
            name_prefix='kaas')]

    resources_map['loadbalancers'] = [
        lb.get('id') for lb in openstack_client.get_loadbalancers() if lb.get(
            'provisioning_status') == 'ACTIVE']

    resources_map['networks'] = [
        net.get('id') for net in openstack_client.get_networks(
            name_prefix='kaas')]

    resources_map['keypairs'] = [
        key.get('name') for key in openstack_client.get_keypairs(
            name_prefix='bootstrap')]

    vol_name_prefixes = ['pvc-', 'kubernetes', 'kaas']
    for prefix in vol_name_prefixes:
        vols = [vol.get('id') for vol in openstack_client.get_volumes(
                name_prefix=prefix)]
        if vols:
            resources_map['volumes'].extend(vols)

    resources_map['security_groups'] = [
        sg.get('name') for sg in openstack_client.get_security_groups(
            name_prefix='kaas')]

    LOG.debug("Next resources are found for deletion "
              "(Excluding FIPs): \n{}".format(yaml.dump(
                  resources_map)))

    if trigger_stacks_cleanup:
        stacks_to_cleanup_pref = settings.STACKS_TO_CLEANUP_PREFIX
        if stacks_to_cleanup_pref:
            for prefix in stacks_to_cleanup_pref:
                stacks = [st.get('id') for st in openstack_client.get_stacks(
                    name_prefix=prefix)]
                if stacks:
                    resources_map['stacks'].extend(stacks)
        else:
            stacks = [
                st.get('id') for st in openstack_client.get_stacks() if
                not any(
                    st.get('stack_name').startswith(pref)
                    for pref in settings.FILTERED_STACKS_PREFIX)]
            if stacks:
                resources_map['stacks'].extend(stacks)

        LOG.debug("Stacks cleanup is triggered. Next stacks will be "
                  "deleted: {}".format(resources_map['stacks']))

        def wait_stacks_deleted():
            try:
                LOG.info("Waiting for stacks are deleted")
                waiters.wait(lambda: not [
                    stack.id for stack in openstack_client.heat.stacks.list()
                    if stack.id in resources_map['stacks']], timeout=300)
            except timeout_exceptions.TimeoutError:
                not_deleted_stacks = \
                    [stack.id for stack in openstack_client.heat.stacks.list()
                     if stack.id in resources_map['stacks']]
                not_deleted_resources['stacks'] = not_deleted_stacks
                LOG.warning("Some stacks are not deleted: {}".format(
                    not_deleted_stacks))

        if resources_map['stacks']:
            LOG.info("Next stack are found for deletion: {}".format(
                resources_map['stacks']))
            for s in resources_map['stacks']:
                try:
                    openstack_client.delete_stack_by_id(s)
                except Exception as e:
                    LOG.warning("Unable to delete stack {}. "
                                "Reason: {}".format(s, e))
            wait_stacks_deleted()

    if trigger_cleanup:
        # Some resources are depend from others, so will wait in right order
        if resources_map['servers']:
            for s in resources_map['servers']:
                try:
                    openstack_client.delete_server(s)
                except Exception as e:
                    LOG.warning("Unable to delete server {}. "
                                "Reason: {}".format(s, e))
            wait_servers_deleted()

        if resources_map['loadbalancers']:
            for lb in resources_map['loadbalancers']:
                try:
                    openstack_client.delete_loadbalancer(lb)
                except Exception as e:
                    LOG.warning("LB {} can not be deleted. {}".format(lb, e))
            wait_loadbalancers_deleted()

        if resources_map['routers']:
            for r in resources_map['routers']:
                try:
                    openstack_client.delete_router(r)
                except Exception as e:
                    LOG.warning("Unable to delete router {}. "
                                "Reason: {}".format(r, e))
            wait_routers_deleted()

        if resources_map['networks']:
            for n in resources_map['networks']:
                try:
                    openstack_client.delete_network(n)
                except Exception as e:
                    LOG.warning("Unable to delete network {}. "
                                "Reason: {}".format(n, e))
            wait_networks_deleted()

        if resources_map['keypairs']:
            for key in resources_map['keypairs']:
                try:
                    openstack_client.delete_keypair(key)
                except Exception as e:
                    LOG.warning("Unable to delete keypair {}. "
                                "Reason: {}".format(key, e))

        if resources_map['volumes']:
            for v in resources_map['volumes']:
                try:
                    openstack_client.delete_volume(v)
                except Exception as e:
                    LOG.warning("Unable to delete volume {}. "
                                "Reason: {}".format(v, e))
            wait_volumes_deleted()

        # Cleanup all of floating IPs which are not associated with ports
        resources_map['fips'] = [
            fip.get('id') for fip in openstack_client.get_fips()
            if fip.get('port_id') is None]
        if resources_map['fips']:
            LOG.debug("Next FIPs are found for deletion: \n{}".format(
                yaml.dump(resources_map['fips'])))
            for ip in resources_map['fips']:
                try:
                    openstack_client.delete_fip(ip)
                except Exception as e:
                    LOG.warning("Unable to delete floating IP {}. "
                                "Reason: {}".format(ip, e))
            wait_fips_deleted()

    if not_deleted_resources:
        LOG.warning("Next resources were not deleted: \n{}".format(
            yaml.dump(not_deleted_resources)))
    else:
        if trigger_stacks_cleanup:
            LOG.info("All found stacks have been deleted")
        if trigger_cleanup:
            LOG.info("All found resources have been deleted")
