import pytest
import yaml

from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters, templates, utils
from si_tests.managers import openstack_manager

LOG = logger.logger

cluster_name = settings.TARGET_CLUSTER
namespace_name = settings.TARGET_NAMESPACE
region = settings.TARGET_REGION
osdpl_name = settings.OSH_DEPLOYMENT_NAME


@pytest.mark.usefixtures('log_method_time')
def test_add_bm_compute(kaas_manager, show_step):
    """Add bm compute node.

    Scenario:
        1. Gather nodes information from yaml file
        2. Check hypervisor count and state before scale
        3. Get info about nodeforscale node
        4. Create bmh
        5. Add labels
        6. Create BareMetal machine
        7. Run cluster check
        8. Run nova cell mappint job
        9. Check hypervisor count and state after scale


    """

    # Get namespace
    LOG.info(f"Namespace name: {namespace_name}")
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)

    # Create ssh key
    public_key_name = "{cluster_name}-key".format(
        cluster_name=cluster_name)
    LOG.info("Public key name - {public_key_name}".format(
        public_key_name=public_key_name))
    ns.get_publickey(public_key_name)

    show_step(1)
    child_data = yaml.safe_load(templates.render_template(
        settings.BM_CHILD_SETTINGS_YAML_PATH))[0]
    bm_hosts_data = child_data['nodes']
    child_cluster_data = child_data.get('child_config')
    tf = child_cluster_data.get('tungstenfabric', [])
    # Getting pods with openstack client
    client_pod = cluster.k8sclient.pods.list(
        namespace="openstack",
        name_prefix='keystone-client')
    assert len(client_pod) > 0, ("No pods found with prefix "
                                 "keystone-client in namespace "
                                 "openstack")
    client_pod = client_pod[0]
    cmd_hypervisor_list = ['/bin/sh', '-c',
                           'openstack hypervisor list -f yaml']
    show_step(2)
    cmp_count = len(yaml.safe_load(client_pod.exec(
        cmd_hypervisor_list)))

    def check(count=None):
        hyper_list = yaml.safe_load(client_pod.exec(cmd_hypervisor_list))
        hyper_states = sum(h.get('State') == 'up' for h in hyper_list)
        LOG.info(f"Count hv: {hyper_states}")
        if hyper_states == count and len(hyper_list) == count:
            return True
        else:
            return False

    waiters.wait(
        lambda: check(count=cmp_count),
        timeout=300, interval=10,
        timeout_msg="Some hypervisor is not ready")
    show_step(3)
    worker_scale = []
    for node in bm_hosts_data:
        if 'nodeforscale' in node.get('si_roles', []):
            cred_name = node['name'] + '-cred'
            # Dirty hack, for not copy-paste bmh_name|cred_name across whole
            # test.
            bmh_name = utils.render_bmh_name(node['name'],
                                             cluster_name,
                                             node.get('bootUEFI', True),
                                             node['bmh_labels'],
                                             si_roles=node.get('si_roles',
                                                               False))
            _node = node.copy()
            _node.update({'bmh_name': bmh_name,
                          'cred_name': cred_name})
            worker_scale.append(_node)
            if not kaas_manager.api.secrets.present(name=cred_name,
                                                    namespace=namespace_name):
                secret_data = {
                    "username": node.get('ipmi').get('username'),
                    "password": node.get('ipmi').get('password')
                }
                ns.create_secret(name=cred_name, data=secret_data,
                                 region=region, provider="baremetal")

    assert len(worker_scale) >= 1, "BMH with si_roles " \
                                   "'nodeforscale' not found"
    show_step(4)
    new_nodes_name = []
    osdpl_node_overrides = dict()
    for node in worker_scale:
        bmh_name = node['bmh_name']
        new_nodes_name.append(bmh_name)
        node_override = node.get('osdpl_node_override', {})
        if node_override:
            osdpl_node_overrides[node['name']] = node_override
        ns.create_baremetalhost(bmh_name=bmh_name,
                                bmh_secret=node['cred_name'],
                                bmh_mac=node['networks'][0]['mac'],
                                bmh_ipmi=node['ipmi']['ipmi_ip'],
                                hardwareProfile=node.get('hardwareProfile',
                                                         False),
                                labels=node['bmh_labels'],
                                bootUEFI=node.get('bootUEFI', True))
    LOG.info(f"New BMH nodes:\n{new_nodes_name}")

    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='ready',
                                    retries=40,
                                    interval=60)
    show_step(5)
    labels = node.get('node_labels', {})
    if tf:
        add_labels = {'openstack-compute-node': 'enabled',
                      'tfvrouter': 'enabled'}
    else:
        add_labels = {'openstack-compute-node': 'enabled',
                      'openvswitch': 'enabled'}
    labels.update(add_labels)
    show_step(6)
    for node in worker_scale:
        custom_bmhp = False
        if node.get('bmh_profile'):
            custom_bmhp = {
                'namespace': namespace_name,
                'name': node['bmh_profile']
            }
        cluster.create_baremetal_machine(
            genname=node['bmh_name'],
            node_pubkey_name=public_key_name,
            matchlabels={'kaas.mirantis.com/'
                         'baremetalhost-id':
                             node['bmh_labels']
                             ['kaas.mirantis.com/baremetalhost-id']},
            baremetalhostprofile=custom_bmhp,
            l2TemplateSelector=node.get('l2TemplateSelector', dict()),
            labels=node['machine_labels'],
            node_labels=labels
        )

    # Wait Baremetal hosts be provisioned
    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='provisioned',
                                    retries=30,
                                    interval=90)

    show_step(7)
    # Waiting for cluster,pods,hb are Ready after machines were added
    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()
    cluster.check.check_k8s_nodes()
    cluster.check.check_cluster_readiness()
    cluster.check.check_helmbundles()
    cluster.check.check_k8s_pods()

    show_step(8)
    cell_cron_job = cluster.k8sclient.cronjobs.list(
        name_prefix="nova-cell", namespace="openstack")[0]
    job = cell_cron_job.read()
    job_template = job.spec.job_template
    job_template.metadata.name = "nova-cell-setup-" + \
                                 utils.gen_random_string(3)
    cluster.k8sclient.jobs.create(body=job_template,
                                  namespace="openstack")

    if settings.OPENSTACK_DEPLOY_DPDK:
        aggregates_cmd = ['/bin/sh', '-c',
                          'openstack aggregate list -f yaml']
        zone = settings.OPENSTACK_DPDK_AVAILABILITY_ZONE_NAME
        if not tf:
            # Automation of these steps
            # https://docs.mirantis.com/mos/latest/deployment-guide/deploy-openstack/advanced-config/enable-dpdk.html # noqa: E501
            child_kubeconfig_name, child_kubeconfig = \
                cluster.get_kubeconfig_from_secret()
            with open('child_conf', 'w') as f:
                f.write(child_kubeconfig)
            os_manager = openstack_manager.OpenStackManager(
                kubeconfig='child_conf')
            for node_name, node_data in osdpl_node_overrides.items():
                machine = [
                    machine for machine in cluster.get_machines()
                    if node_name in machine.name][0]
                LOG.info("Install DPDK driver on new compute")
                machine.run_cmd(
                    "sudo apt install -y "
                    f"{settings.OPENSTACK_DPDK_DRIVER_PACKAGE}",
                    timeout=600)
                k8s_node = machine.get_k8s_node()
                osdpl = os_manager.get_openstackdeployment(osdpl_name)
                osdpl_update = {'spec': {'nodes': {
                    'kubernetes.io/hostname::{}'.format(
                        k8s_node.name): node_data}}}
                osdpl.patch(osdpl_update)
                os_manager.wait_all_osdpl_children_status()
                os_manager.wait_openstackdeployment_health_status()
                os_manager.wait_all_os_pods(timeout=900, interval=90)
                # Adding new compute to aggregate advanced-compute
                # if it exists
                aggregates = yaml.safe_load(client_pod.exec(
                    aggregates_cmd))
                if aggregates:
                    aggregate = [
                        a for a in aggregates if zone in a[
                            'Availability Zone']]
                    if aggregate:
                        # If several aggregates exist with the same AZ
                        # it is not matter where new host will be added
                        # So we will take first aggregate
                        aggregate = aggregate[0]
                        host_name = k8s_node.name
                        aggr_name = aggregate.get('Name')
                        aggr_id = aggregate.get('ID')
                        add_host_cmd = [
                            '/bin/sh', '-c', 'openstack aggregate add host '
                            '{} {}'.format(aggr_id, host_name)]
                        LOG.info("Host {} will be added to "
                                 "aggregate {}".format(host_name, aggr_name))
                        res = client_pod.exec(add_host_cmd)
                        LOG.info("Adding host result: \n{}".format(res))
                    else:
                        LOG.warning(
                            "There are no aggregates with {} AZ. "
                            "Existing aggregates: {}".format(
                                zone, [a['Name'] for a in aggregates]))
                else:
                    LOG.warning("There are no aggregates. Compute will "
                                "be in default (nova) availability zone")
        else:
            # Will be implemented
            pass
        cluster.check.check_k8s_pods()
        cluster.check.check_cluster_readiness()
    show_step(9)
    new_compute_count = cmp_count + len(worker_scale)
    waiters.wait(
        lambda: check(count=new_compute_count),
        timeout=300, interval=10,
        timeout_msg="new hypervisor not ready")
