import pytest
import os
import re
import yaml

import kubernetes

from si_tests.deployments import openstack_deploy
from si_tests.managers import openstack_manager
from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters
from si_tests.managers.kaas_manager import Manager
from si_tests.utils.utils import Providers

LOG = logger.logger


def get_upgrade_path(supported_crs, update_versions, path=None):
    path = path or []
    if update_versions:
        for update_version in update_versions:
            new_data = supported_crs[update_version]
            new_path = path.copy()
            new_path.append(new_data['name'])
            new_update_versions = [
                ver['version']
                for ver in new_data.get('availableUpgrades', [])]
            for path in get_upgrade_path(supported_crs,
                                         new_update_versions,
                                         new_path):
                yield path
    else:
        yield path


def generete_update_release_names():
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    child_cluster = ns.get_cluster(settings.TARGET_CLUSTER)

    target_release_name = settings.KAAS_CHILD_CLUSTER_UPDATE_RELEASE_NAME
    if not target_release_name:
        LOG.info("KAAS_CHILD_CLUSTER_UPDATE_RELEASE_NAME wasn't set, trying "
                 "to acquire available clusterrelease version for upgrade")
        cr_available = child_cluster.available_clusterrelease_version
        assert cr_available is not None, "Version for upgrade wasn't found"
        LOG.info("Cluster release version {0} will be used for upgrade"
                 .format(cr_available))
        target_release_name = cr_available

    cr_current = child_cluster.clusterrelease_version
    if cr_current == target_release_name:
        LOG.info(f"\n\nTarget release name is the same as the current release:"
                 f" {target_release_name}\n")
        return target_release_name

    supported_crs = kaas_manager.get_supported_clusterreleases()
    LOG.debug(f"Supported releases:\n{supported_crs}")
    assert len([data for _, data in supported_crs.items()
                if data['name'] == cr_current]) == 1, (
        f"More than one clusterrelease with name {cr_current} found in "
        f"kaasrelease object spec.supportedClusterReleases:\n"
        f"{yaml.dump(supported_crs)}"
    )

    update_versions = [
        [
            ver['version'] for ver in data.get('availableUpgrades', [])
        ]
        for _, data in supported_crs.items() if data['name'] == cr_current
    ][0]

    if not update_versions:
        raise Exception(f"Clusterrelease {cr_current} don't have any "
                        f"'availableUpgrades' to reach the specified "
                        f"version {target_release_name}")

    paths = list(get_upgrade_path(supported_crs, update_versions))
    paths_str = "\n".join([str(p) for p in paths])
    LOG.info(f"All available upgrade paths for {cr_current}:\n{paths_str}")

    for path in paths:
        if target_release_name in path:
            break
    LOG.info(f"Select the following upgrade path to reach the release "
             f"{target_release_name}:\n{path}")
    for release_name in path:
        yield release_name
        if release_name == target_release_name:
            break


update_release_names = list(generete_update_release_names())
is_update_test_failed = False


@pytest.fixture(scope='function', params=update_release_names,
                ids=[f"RELEASE={x}" for x in update_release_names])
def update_release_name(request):
    global is_update_test_failed
    # Check if the previous update steps failed
    if is_update_test_failed:
        msg = (f"Skip updating clusterrelease to {request.param} because "
               f"previous update step failed")
        LOG.info(msg)
        pytest.skip(msg)

    yield request.param

    # Check the result of the current step
    test_passed = (hasattr(request.node, 'rep_call') and
                   request.node.rep_call.passed)
    if not test_passed:
        is_update_test_failed = True


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
@pytest.mark.usefixtures("store_updated_child_cluster_description")
@pytest.mark.usefixtures("introspect_PRODX_9696")
def test_update_child_clusterrelease(kaas_manager, update_release_name, _):
    """Update child cluster release"""

    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE

    ns = kaas_manager.get_namespace(namespace_name)
    child_cluster = ns.get_cluster(cluster_name)
    cr_before = child_cluster.clusterrelease_version

    if update_release_name == cr_before:
        msg = (f"Requested {update_release_name} is the same as current "
               f"clusterrelease version {cr_before}, skipping update")
        pytest.skip(msg)

    LOG.info("Collect child cluster fixed resources before update")
    child_client = ns.get_cluster(settings.TARGET_CLUSTER)
    lcm_type = child_cluster.data['status'].get(
        'providerStatus', {}).get('releaseRefs', {}).get(
        'current', {}).get('lcmType', '')
    LOG.info("LCM Type is {0}".format(lcm_type))
    cluster_resources_before = child_client.describe_fixed_resources()
    LOG.debug("Child cluster resources before update:\n{0}"
              .format(yaml.dump(cluster_resources_before)))
    ucp_tag_in_cr_before = set(
        [x['params']['ucp_tag'] for x in kaas_manager.get_clusterrelease(
            cr_before).data['spec']['machineTypes']['control']
         if 'ucp_tag' in x['params'].keys()])

    LOG.info("Cluster release before update {0}".format(cr_before))
    LOG.info('ucp_tag before update {0}'.format(ucp_tag_in_cr_before))
    LOG.info(("=" * 80) + "\n\nUpdating child cluster {0} "
             "to the clusterrelease {1}\n\n"
             .format(cluster_name, update_release_name))
    # TODO (tleontovich) Delete after 2.11 as
    #  label will be added during cluster creation
    # Check if we need labels
    lma_label = {"key": "stacklight", "value": "enabled"}
    not_labeled = [x for x in child_cluster.get_machines(
        machine_type="worker") if lma_label not in
                   [{'key': y['key'], 'value': y['value']}
                    for y in x.nodeLabels]]
    lma_labeled = [x for x in child_cluster.get_machines(
        machine_type="worker") if lma_label in
                   [{'key': y['key'], 'value': y['value']}
                    for y in x.nodeLabels]]
    if child_cluster.sl_ha_enabled() and len(lma_labeled) < 3:
        LOG.info(f"Labeled nodes {lma_labeled}")
        LOG.info(f"Non labeled nodes {not_labeled}")
        for machine in not_labeled:
            LOG.info(f"Updating machine {machine} with {lma_label}")
            machine.add_labels([lma_label])
            waiters.wait(lambda: machine.check_k8s_nodes_has_labels(),
                         timeout=300, interval=10)
            assert lma_label in machine.nodeLabels, \
                f"Failed to update {machine} with {lma_label}"

    child_cluster.update_cluster(update_release_name)
    child_cluster.check.check_cluster_release(update_release_name)
    child_cluster.check.check_update_finished()
    # Nodes will not have updating status if there is no changes
    # in ucp_tag or CR versions
    ucp_tag_after = set(
        [x['params']['ucp_tag']
         for x in kaas_manager.get_clusterrelease(
            update_release_name).data['spec']['machineTypes']['control']
         if 'ucp_tag' in x['params'].keys()]
    )
    child_cluster.check.check_cluster_nodes()
    if ('mos' in update_release_name and
            child_cluster.openstack_lma_enabled()):
        expected_fails = child_cluster.check.EXCLUDED_PODS
        expected_condition_fails = child_cluster.check.EXCLUDED_CONDITIONS
        child_cluster.check.check_cluster_readiness(
            exp_provider_status=False,
            expected_fails=expected_fails,
            expected_condition_fails=expected_condition_fails)
    else:
        child_cluster.check.check_cluster_readiness()
    child_cluster.check.check_helmbundles()
    child_cluster.check.check_k8s_nodes()
    # Check/wait for correct docker service replicas in cluster
    # We have renamed agent if ucp changes version
    if update_release_name != cr_before and \
            ucp_tag_in_cr_before != ucp_tag_after:

        ucp_worker_agent_name = child_cluster.check.get_ucp_worker_agent_name()
        child_cluster.check.check_actual_expected_docker_services(
            changed_after_upd={'ucp-worker-agent-x': ucp_worker_agent_name})
    else:
        child_cluster.check.check_actual_expected_docker_services()
        assert ucp_tag_in_cr_before == ucp_tag_after,\
            "UCP tag was updated, when shouldn't"

    child_cluster.check.check_k8s_pods()
    child_cluster.check.check_actual_expected_pods(timeout=3200)

    if child_cluster.provider in Providers.with_ceph():
        child_cluster.check.check_ceph_pvc()

    # check k8s version
    failed = {}
    for node in child_cluster.k8sclient.nodes.list_all():
        node_info = node.read().status.node_info
        if node_info.kubelet_version != node_info.kube_proxy_version:
            failed[node.name] = "kubelet_version {0} doesn't match " \
                                "kube_proxy_version {1} version".format(
                node_info.kubelet_version,
                node_info.kube_proxy_version)

    assert failed == {}, "k8s versions mismatch " \
                         "Details: {}".format(failed)

    child_cluster.store_k8s_artifacts()
    child_cluster.provider_resources.save_artifact()

    # Check for changed/missing resources
    child_cluster.check.check_k8s_resources_after_update(
        child_client, cluster_resources_before)
    LOG.info("\n*** Child cluster <{0}> "
             "have been upgraded to the clusterrelease {1}"
             .format(cluster_name, update_release_name))

    if re.search(r"mos.?-.*", child_cluster.clusterrelease_version):
        LOG.info("Child cluster name contains MOS pattern. "
                 "coredns ConfigMap is going to be adjusted.")
        child_kubeconfig_name, child_kubeconfig = \
            child_cluster.get_kubeconfig_from_secret()
        child_kubeconfig_path = os.path.join(
            settings.ARTIFACTS_DIR, child_kubeconfig_name)
        os_manager = openstack_manager.OpenStackManager(
            kubeconfig=child_kubeconfig_path)
        timeouts = settings.OPENSTACK_DEPLOY_TIMEOUT
        try:
            os_manager.get_openstackdeployment(
                name=settings.OSH_DEPLOYMENT_NAME, read=True)
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                LOG.info("There is no MOS apps deployed.")
        else:
            LOG.info("Configuring coredns in kube-system to work "
                     "with it.just.works domain")
            openstack_deploy.configure_dns_kubesys(
                os_manager,
                timeouts.get("3", 180)
            )
            os_manager.wait_osdpl_status(timeout=3600)
            LOG.info("Wait osdpl health status=Ready")
            os_manager.wait_openstackdeployment_health_status(
                timeout=1800)
            LOG.info("Wait os jobs to success and pods to become Ready")
            os_manager.wait_os_resources(timeout=1800)
