#    Copyright 2019 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import pytest
import json
from packaging import version

from si_tests import logger
from si_tests import settings
from si_tests.managers import bootstrap_manager
from si_tests.utils.utils import Providers
from si_tests.managers.kaas_manager import Manager

LOG = logger.logger


class TestUpdateMgmCluster():

    @pytest.mark.usefixtures("store_updated_mgmt_cluster_description")
    @pytest.mark.usefixtures("introspect_PRODX_9696")
    def test_update_mgmt_cluster(self, kaas_manager: Manager):
        """Check mgmt is updated

        Scenario:
            1. Check release
        """
        def sanitize_lcmmachines(lcmmachines):
            for lcm in lcmmachines:
                if 'self_link' in lcm.get('metadata', {}).keys():
                    del lcm['metadata']['self_link']
                if 'resource_version' in lcm.get('metadata', {}).keys():
                    del lcm['metadata']['resource_version']
                if 'generation' in lcm.get('metadata', {}).keys():
                    del lcm['metadata']['generation']
                if 'mcrVersion' in lcm.get('spec', {}).keys():
                    del lcm['spec']['mcrVersion']
                if 'mcrVersion' in lcm.get('status', {}).keys():
                    del lcm['status']['mcrVersion']
                if 'observedGeneration' in lcm.get('status', {}).keys():
                    del lcm['status']['observedGeneration']
                for lcmm in lcm.get('metadata', {}).get('managed_fields', []):
                    if 'manager' in lcmm.keys():
                        del lcmm['manager']
                    if 'time' in lcmm.keys():
                        del lcmm['time']

        cr_before = kaas_manager.get_mgmt_cluster().clusterrelease_version
        LOG.info("Management cluster clusterrelease version is "
                 "{}".format(cr_before))

        ucp_tag_in_cr_before = set(
            [x['params']['ucp_tag'] for x in kaas_manager.get_clusterrelease(
                cr_before).data['spec']['machineTypes']['control']
             if 'ucp_tag' in x['params'].keys()])
        LOG.info("Current management cluster ucp version is "
                 "{}".format(ucp_tag_in_cr_before))

        kaas_manager.wait_kaasrelease(
            releases_count=2, timeout=settings.KAAS_UPDATE_MGMT_WAIT_TIMEOUT)

        avail_kaas_releases = kaas_manager.get_kaasrelease_names()
        LOG.info('Available KaaS releases are {}'.format(avail_kaas_releases))

        assert len(avail_kaas_releases) > 1, (
            "There are not new kaas releases, current list: {}".format(
                avail_kaas_releases))

        kaas_r_sorted = str(max(
            [version.parse(x) for x in avail_kaas_releases]))
        LOG.info('Sorted release {}'.format(kaas_r_sorted))
        # Lets wait for pods statuses before check label
        cluster_name = settings.TARGET_CLUSTER
        cluster_ns = settings.TARGET_NAMESPACE
        ns = kaas_manager.get_namespace(cluster_ns)
        cluster = ns.get_cluster(cluster_name)
        cluster.check.check_k8s_pods()

        lcmms = \
            kaas_manager.api.kaas_lcmmachines.list_raw().to_dict()
        all_managed_lcmm = [lm for lm in lcmms['items']
                            if lm['metadata']['namespace'] != "default"]
        r_clusters = kaas_manager.get_regional_clusters()
        for r_cluster in r_clusters:
            lcmms = \
                r_cluster.k8sclient.kaas_lcmmachines.list_raw().to_dict()
            all_managed_lcmm.extend(
                [lm for lm in lcmms['items']
                 if lm['metadata']['namespace'] != r_cluster.namespace]
            )
        sanitize_lcmmachines(all_managed_lcmm)

        kr_data = kaas_manager.get_kaasrelease(kaas_r_sorted).data
        LOG.debug("Release_details {}".format(kr_data))
        kaas_manager.wait_kaasrelease_label(kaas_r_sorted, timeout=9000)
        formated_kr = kr_data['spec']['version'].replace('.', '-')
        assert formated_kr == kaas_r_sorted.replace('kaas-', '')
        mgmt_cluster = kaas_manager.get_mgmt_cluster().spec['providerSpec']
        assert kaas_r_sorted == mgmt_cluster['value']['kaas']['release']
        LOG.info('MGMT spec {}'.format(mgmt_cluster['value']['release']))

        cluster_releases = kaas_manager.get_clusterrelease_names()
        spec_cr = kr_data['spec']['clusterRelease']
        LOG.info('Cluster version in new KaaSRelease: {}'.format(spec_cr))
        assert spec_cr in cluster_releases
        assert spec_cr == mgmt_cluster['value']['release'], (
            "Cluster version in KaasRelease and Mgmt "
            "Cluster differs, that's wrong")

        cluster.check.check_cluster_release(spec_cr)
        # TODO: we need to check regional child cluster updates too
        cluster.check.check_update_finished(check_managed_clusters=True)

        ucp_tag_after = set(
            [x['params']['ucp_tag']
             for x in kaas_manager.get_clusterrelease(
                spec_cr).data['spec']['machineTypes']['control']
             if 'ucp_tag' in x['params'].keys()])
        if ucp_tag_in_cr_before != ucp_tag_after:
            LOG.info("Management cluster new ucp version will be "
                     "{}".format(ucp_tag_after))

        cluster.check.check_cluster_nodes()
        cluster.check.check_cluster_readiness(timeout=3600)
        cluster.check.check_helmbundles()
        cluster.check.check_k8s_nodes()
        cluster.check.check_persistent_volumes_mounts()
        cluster.check.check_k8s_pods()
        if cluster.provider in Providers.baremetal.value:
            cluster.check.check_ceph_pvc()

        # Check/wait for correct docker service replicas in cluster
        cluster.check.check_actual_expected_docker_services()
        cluster.check.check_actual_expected_pods(timeout=600)

        lcmms = \
            kaas_manager.api.kaas_lcmmachines.list_raw().to_dict()
        all_managed_lcmm_new = [lm for lm in lcmms['items']
                                if lm['metadata']['namespace'] != "default"]

        for r_cluster in r_clusters:
            # get regional clusters and check CR there
            r_cr = r_cluster.spec['providerSpec']['value']['release']
            LOG.info("Current regional cr is {}".format(r_cr))
            assert spec_cr in r_cr, ("Cluster version in regional cluster "
                                     "differs from latest kaas release, "
                                     "that's wrong")
            r_cluster.check.check_cluster_release(spec_cr)

            r_cluster.check.check_cluster_nodes()
            r_cluster.check.check_cluster_readiness()
            r_cluster.check.check_helmbundles()
            r_cluster.check.check_k8s_nodes()
            r_cluster.check.check_persistent_volumes_mounts()
            r_cluster.check.check_k8s_pods()

            # Check/wait for correct docker service replicas in cluster
            r_cluster.check.check_actual_expected_docker_services()
            r_cluster.check.check_actual_expected_pods(timeout=600)
            lcmms = \
                r_cluster.k8sclient.kaas_lcmmachines.list_raw().to_dict()
            all_managed_lcmm_new.extend(
                [lm for lm in lcmms['items']
                 if lm['metadata']['namespace'] != r_cluster.namespace]
            )

        LOG.info("Check child clusters")
        child_clusters = kaas_manager.get_child_clusters()
        for child in child_clusters:
            LOG.info(f" - check {child.name} cluster")
            child.check.check_cluster_readiness()

        sanitize_lcmmachines(all_managed_lcmm_new)

        LOG.debug(f"All managed lcmmachines before update: "
                  f"{json.dumps(all_managed_lcmm, indent=4, default=str)}")
        LOG.debug(f"All managed lcmmachines after update: "
                  f"{json.dumps(all_managed_lcmm_new, indent=4, default=str)}")

        assert all_managed_lcmm_new == all_managed_lcmm,\
            "Error: lcmmachines of managed cluster have changed"

        if settings.FEATURE_FLAGS.enabled("tarball-update"):
            try:
                LOG.info('Updating bootstrap version')
                bootstrap = bootstrap_manager.BootstrapManager()
                bootstrap.update_bootstrap_version()
            except Exception as e:
                LOG.warning(f"Unable to update bootstrap version: "
                            f"{e}")
